//
//	test the inter-chip gaps of EPN
//
//	REFERENCES:
//		[1] : "The behavior of the chip boundaries in pn-EPIC", M. Popp
//
//	UL, Thu Sep 30 11:46:13 MET DST 1999
//
#include <cassert>
#include <cmath>
#include <fstream>
#include <iostream>
#include "CalServer.h"
#include "Xmm.h"
#include "GeometryDataServer.h"
#include "common.h"

using namespace std;

namespace {

struct Positions {
	CalReal32Vector	x, y;
};

Positions pixelGaps(const CalInt16Vector &xi, const CalInt16Vector &yi,
				    const CalInt16Vector &ccd)
{
	unsigned N = ccd.size();
	assert(N > 1);

	CalReal32Vector xrv(0, N),
					yrv(0, N);

	real32 xr = 0.f,
		   yr = 0.f;
	for (unsigned n=0; n<N; ++n) {
		calServer.state()->ccd()->set(ccd[n]);
		GeometryDataServer *ct = calServer.getAtom(ct);
		CalInt16Vector x0(&xi[n], 1),
					   y0(&yi[n], 1);
		XmmPixelList pl = ct->pixCoord1List(x0, y0);
		XmmCartesianPointList cpl = pl.toChipCoord().toCamCoord1();

		xrv[n] = 1000. * (cpl.x()[0]-xr);
		yrv[n] = 1000. * (cpl.y()[0]-yr);
		cout<<"ccd="<<ccd[n]<<" "<<x0[0]<<"/"<<y0[0]
			<<" = "<<cpl.x()[0]<<"/"<<cpl.y()[0]
			<<" - diff: "<<(cpl.x()[0]-xr)*1000.<<"/"
						 <<(cpl.y()[0]-yr)*1000.<<endl;
		xr = cpl.x()[0];
		yr = cpl.y()[0];
	}
	cout<<endl;

	Positions p; p.x = xrv; p.y = yrv;
	return p;
}

bool isClose(real32 a, real32 b)
{
	if (fabs(b) < 5.e-3)
		//	absolute
		return fabs(a) < 5.e-3;
	else
		return fabs(fabs(a)/fabs(b)-1.) < 5.e-3;
}

}

int main()
{
	calServer.setInstrument(Xmm::EPN);
	calServer.state()->randomize()->set(false);
	calServer.state()->temperature()->set(293.);
	calServer.state()->temperatureCamera()->set(293);

	//
	//	CCD 1 -> 10 in Y - cross a quadrant boundary Q0->Q3
	//
	int16 x1[] = {   1,   1,   1,  64,  64,  64 },
		  y1[] = { 198, 199, 200, 200, 199, 198 },
		ccd1[] = {   1,   1,   1,  10,  10,  10 };
	float dx1[] = { -182, 0, 0, 0, 0, 0 },
		  dy1[] = { 446, -150, -175, -241, -175, -150 };

	unsigned N = NUM_OF(x1);
	CalInt16Vector x1v(x1, N),
				   y1v(y1, N),
				   ccd1v(ccd1, N);
	Positions p1 = pixelGaps(x1v, y1v, ccd1v);
	for (unsigned n=0; n<N; ++n)
		assert(isClose(p1.x[n], dx1[n]) && isClose(p1.y[n], dy1[n]));

	//
	//	CCD 1 -> 2 in X	- cross chip boundary in quadrant Q0
	//
	int16 x2[] = { 62, 63, 64, 1, 2, 3 },
		  y2[] = {  1,  1,  1, 1, 1, 1 },
		ccd2[] = {  1,  1,  1, 2, 2, 2 };
	float dx2[] = { -9332, -150, -150, -344, -150, -150 },
		  dy2[] = { 29996, 0, 0, 0, 0, 0 };

	N = NUM_OF(x2);
	CalInt16Vector x2v(x2, N),
				   y2v(y2, N),
				   ccd2v(ccd2, N);
	Positions p2 = pixelGaps(x2v, y2v, ccd2v);
	for (unsigned n=0; n<N; ++n)
		assert(isClose(p2.x[n], dx2[n]) && isClose(p2.y[n], dy2[n]));

	//
	//	CCD 1 -> 4 in X - cross a quadrant boundary Q0 -> Q1
	//
	int16 x3[] = {   3,   2,   1,  64,  63,  62 },
		  y3[] = { 200, 200, 200, 200, 200, 200 },
		ccd3[] = {   1,   1,   1,   4,   4,   4 };
	float dx3[] = { -482, 150, 150, 364, 150, 150 },
		  dy3[] = { 121, 0, 0, 0, 0, 0 };

	CalInt16Vector x3v(x3, NUM_OF(x3)),
				   y3v(y3, NUM_OF(y3)),
				   ccd3v(ccd3, NUM_OF(ccd3));

	Positions p3 = pixelGaps(x3v, y3v, ccd3v);
	for (unsigned n=0; n<N; ++n)
		assert(isClose(p3.x[n], dx3[n]) && isClose(p3.y[n], dy3[n]));


	//
	//	check GeometryDataServer::onCcd()
	//
	GeometryDataServer *ct = calServer.getAtom(ct);
	double xcc2 = 0.,
		   ycc2 = 0.,
		   deltaX = 150.e-3/10.,	//	5 steps per pixel
		   deltaY = -1.e-3;
	int maxXsteps = 3*64*10+200;
	for (int n=-maxXsteps; n<maxXsteps; ++n) {
		double xp = xcc2 + n*deltaX,
			   yp = ycc2 + 0*deltaY;
		cout<<xp<<'/'<<yp<<" is on CCD: "<<ct->onCcd(xp, yp)<<endl;
	}
	cout<<endl;

	return 0;
}
