!
!   caltest_emos.f90  --- f90 program to test the CAL (EMOS aspects only)
!
!   DESCRIPTION:
!       This program tests the correctness of the EMOS-related CAL part by
!       performing a sequence of subroutine calls followed by
!       checks of the output. If any of these checks fails, the
!       program shall exit with a non-zero return value. If all tests
!       succeed 0 will be returned to the calling environment.
!
program caltest_emos

use cal             ! use the Calibration Access Layer
use dal
use errorhandling

    implicit none
    logical                        :: isClose
    integer                        :: i, j, k, isum(2), nx, ny
    integer(int16), dimension(:), pointer :: &
                                   rawX, rawY, badPixType, yextent
    integer(int8), dimension(:), pointer  :: ccd
    integer(int16), dimension(:), pointer    :: offX, offY
    real(single), dimension(:,:), pointer  :: darkFrame
    integer(int8), dimension(:,:,:), pointer :: imagPatterns, timingPatterns
    integer(int16)                           :: eduThreshold
    logical, dimension(:), pointer :: uplinked
    integer(int16), dimension(0:6) :: &
                                  rawXtest = (/1, 23, 34, 56, 118, 567, 300/),&
                                  rawYtest = (/1, 234, 43, 56, 98, 576, 447/),&
                                  Ein = (/123, 456, 789, 999, 364, 888, 1000/)
    integer(int8), dimension(0:6)  :: &
                                  pattern = (/0, 1, 4, 5, 8, 9, 31/)
    real(double), dimension(0:5)  :: &
                                   timeTags = (/0._double, 0._double, &
                                                0._double, 0._double, &
                                                0._double, 0._double/)
    integer(int16), dimension(:), pointer :: PHAint, PIint
    real(kind=single), dimension(:), pointer :: PHAreal, PIreal, EinReal
    real(double), dimension(:), pointer :: xmilli, ymilli, zmilli, &
                                           thetaOut, phiOut
    integer(int16), dimension(:,:), pointer :: ibadpixmap, &
                                               fovmap

    type(BoresightT)                       :: boresight
    real(double), dimension(:,:), pointer  :: matrix, matrix2, cc2sacMat
    real(single), dimension(:,:), pointer  :: PSFmap
    type(PsfT)                      :: psf
    type(PsfWindowT)                :: window
    real(double)                    :: focalLength, plateScaleX, area, vign
    character(len=*), parameter     :: myBadPixFile = &
                                      './emos1_badpix.ccf'
    type(CalInfo)                   :: badPixFileInfo
    integer(int32)                  :: instrumentOut, ccdChipIdOut, &
                                       ccdNodeIdOut, accLevelOut, modeNr
    real(single)                    :: sum
    real(double), dimension(:), pointer :: mmZ, mmY, mmX, theta, phi
    real(double), dimension(0:0)    ::    theta0, phi0
    type(DataSetT)                  :: set
    type(BlockT)                    :: inBlock

    write(*, *) 'setting state from data set'
    call CAL_setState('emos.fits')
    call CAL_getState(ccdModeId = modeNr)
    write(*, *) 'mode=', modeNr

    set = dataSet('emos.fits', READ)
    write(*, *) 'setting state from set'
    call CAL_setState(set);

    inBlock = block(set, 0, READ)
    write(*, *) 'setting state from block'
    call CAL_setState(inBlock)
    call release(set)

    call CAL_getState(instrument=instrumentOut, ccdChipId = ccdChipIdOut, &
                      ccdNodeId=ccdNodeIdOut, accuracyLevel = accLevelOut)

    if (instrumentOut /= EMOS2) &
        call error('caltest_emos', "instrumentOut /= EMOS2")
    if (ccdChipIdOut /= EMOS_CCD_2) &
        call error('caltest_emos', "ccdChipIdOut /= EMOS_CCD_2")
    if (ccdNodeIdOut /= 1) &
        call error('caltest_emos', "ccdNodeIdOut /= 1")


    call CAL_setState(EMOS1, ccdChipId=EMOS_CCD_CENTRAL, ccdNodeId=0, &
                      accuracyLevel = ACCURACY_LOW)
    call CAL_getState(instrument=instrumentOut, ccdChipId = ccdChipIdOut, &
                      ccdNodeId=ccdNodeIdOut, accuracyLevel = accLevelOut)
    if (instrumentOut /= EMOS1) &
        call error('caltest_emos', 'instrumentOut /= EMOS1')
    if (ccdChipIdOut /= EMOS_CCD_CENTRAL) &
        call error('caltest_emos', "ccdChipIdOut /= EMOS_CCD_CENTRAL")
    if (ccdNodeIdOut /= 0) &
        call error('caltest_emos', "ccdNodeIdOut /= 0")
    if (accLevelOut /= ACCURACY_LOW) &
        call error('caltest_emos', "accLevelOut /= ACCURACY_LOW")

    call CAL_ccfInfo('BADPIX', badPixFileInfo)
    call CAL_setCalibrationFile(myBadPixFile)
    call CAL_setCalibrationFile('myemoslincoord.ccf')

    !
    !   get bad pixel table / check a few
    !
    if (CAL_getBadPixelTableCode() /= INVALID_BAD_PIXEL_TABLE_CODE .or. &
        CAL_getAduconvCode() /= INVALID_BAD_PIXEL_TABLE_CODE) &
        call error('caltest_emos', &
                    "unexpected bad-pixel-table code")

    call CAL_setState(ccdChipId = EMOS_CCD_FIRST, ccdNodeId = 0)
    call CAL_getBadPixelList(rawX, rawY, badPixType, yextent, uplinked)
    if (size(rawX) /= 50) call error('caltest_emos', "size(rawX) /= 100")
    if (rawX(0) /= 9 .or. rawY(0) /= 247 .or. &
        yextent(0) /= 3 .or. badPixType(0) /= DEAD .or. &
        uplinked(0)) &
        call error('caltest_emos', &
                   "something wrong with bad pixel 0 - node 0")

    if (rawX(49) /= 581 .or. rawY(49) /= 62 .or. &
        yextent(49) /= 6 .or. badPixType(49) /= HOT .or. &
        uplinked(49)) &
        call error('caltest_emos', "something wrong with bad pixel 49")
    call CAL_releaseMemory(rawX, rawY)
    call CAL_releaseMemory(badPixType, yextent)
    call CAL_releaseMemory(uplinked)

    call CAL_setState(ccdNodeId = 1)
    call CAL_getBadPixelList(rawX, rawY, badPixType, yextent, uplinked)
    if (size(rawX) /= 50) &
         call error("size(rawX) /= 100 (node1)")
    if (rawX(5) /= 94 .or. rawY(5) /= 173 .or. &
        yextent(5) /= 2 .or. badPixType(5) /= HOT .or. &
        uplinked(5)) &
            call error('caltest_emos', &
                       "something wrong with bad pixel 0 - node1")

    !   bad pixel map (ccd #0/node #1)
    call CAL_getBadPixelMap(ibadpixmap)
    do i=0, size(rawX)-1
        j = ibadpixmap(rawX(i)-1, rawY(i)-1)
        if ((mod(j, 10) /= badPixType(i))) then
            write(*, *) 'defect type of pixel', rawX(i), '/', rawY(i), &
                        ' is not right'
            call error('caltest_emos', "problem with bad pixel map")
        endif
        if (((j/10 .eq. 1) .and. .not.uplinked(i)) .or. &
            ((j/10 .eq. 0) .and. uplinked(i))) then
            write(*, *) 'uplink status of pixel', rawX(i), '/', rawY(i), &
                        ' is not right'
            call error('caltest_emos', &
                       "problem with uplink status of bad pixel")
        endif
    enddo
    call CAL_setState(ccdNodeId = 0)
    call CAL_releaseMemory(rawX, rawY)
    call CAL_releaseMemory(badPixType, yextent)
    call CAL_releaseMemory(uplinked)
    call CAL_releaseMemory(ibadpixmap)

    do i=0, 1
        write(*, *) 'Getting FOV submap [298, 303]x[1, 5] for node ', i
        call CAL_setState(ccdChipId = 1, ccdNodeId = i)
        call CAL_getFOVmap(fovmap, 298_int16, 1_int16, 6_int16, 5_int16)
        isum(1) = 0
        do j=0, size(fovmap, 1)/2-1
            do k=0, size(fovmap, 2)-1
                isum(1) = isum(1) + fovmap(j, k)
            enddo
            write(*, *) (fovmap(j, k), k=0, size(fovmap, 2)-1)
        enddo
        isum(2) = 0
        do j=size(fovmap, 1)/2, size(fovmap, 1)-1
            do k=0, size(fovmap, 2)-1
                isum(2) = isum(2) + fovmap(j, k)
            enddo
            write(*, *) (fovmap(j, k), k=0, size(fovmap, 2)-1)
        enddo
        if (isum(i+1) /= 15 .or. isum(1-i+1) /= 0) &
            call error('caltest_emos', "problem with FOV maps")

        call CAL_releaseMemory(fovmap)
    enddo
    call CAL_setState(ccdNodeId = 0)


    !
    !   check get boresight matrix - sum of elements should be 3
    !
    boresight = CAL_getBoresight(0.d0)
    matrix => CAL_toDirectionCosineMatrix(boresight)
    write(*, *) 'Boresight misalignment martrix:'
    do i=0, 2
        write(*, *) (matrix(i, j), j=0, 2)
    enddo
    sum = abs(matrix(0, 0)) + abs(matrix(0, 1)) + abs(matrix(0, 2)) + &
          abs(matrix(1, 0)) + abs(matrix(1, 1)) + abs(matrix(1, 2)) + &
          abs(matrix(2, 0)) + abs(matrix(2, 1)) + abs(matrix(2, 2))
    write(*, *) sum
    if (.not. isClose(sum, 3.0044713)) &
        call error('caltest_emos', "boresight matrix is not right")

    matrix2 => CAL_toDirectionCosineMatrix(CAL_toEulerAngles(boresight))
    do i=0, 2
        do j=0, 2
            if (.not. isClose(real(matrix(i, j)), real(matrix2(i, j)))) &
                call error('caltest_emos', "boresight matrices differ")
        enddo
    enddo

    call CAL_getCamCoord2ToSacCoord(0._double, cc2sacMat)
    do i=0, 2
        write(*, *) (cc2sacMat(i, j), j=0, 2)
    enddo
    sum = abs(cc2sacMat(0, 0))+abs(cc2sacMat(0, 1))+abs(cc2sacMat(0, 2)) + &
          abs(cc2sacMat(1, 0))+abs(cc2sacMat(1, 1))+abs(cc2sacMat(1, 2)) + &
          abs(cc2sacMat(2, 0))+abs(cc2sacMat(2, 1))+abs(cc2sacMat(2, 2))
    if (.not. isClose(sum, 3.)) &
        call error('caltest_emos', "CC2->SACCOORD matrix is not right")

    !
    !   misc data
    !
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLength)
    call CAL_getMiscellaneousData('PLATE_SCALE_X', plateScaleX);
    if (.not. isClose(real(focalLength, single), 7494.)) &
        call error('caltest_emos', "focal length is wrong")
    if (.not. isClose(real(plateScaleX, single), 1.10080)) &
        call error('caltest_emos', "plate scale is wrong")

    !
    !   pixel to mm
    !
    call CAL_setState(randomize = .false.)
    call CAL_rawXY2mm(rawXtest, rawYtest, xmilli, ymilli, zmilli)
    write(*, *) 'SACCOORD'
    write(*, *) xmilli, ymilli, zmilli
    call CAL_sacCoordToCamCoord2(xmilli, ymilli, zmilli, xmilli,ymilli,zmilli)
    write(*, *) 'CAMCOORD2'
    write(*, *) xmilli, ymilli, zmilli
    call CAL_camCoord2ToCamCoord1(xmilli, ymilli, zmilli, xmilli,ymilli,zmilli)
    write(*, *) 'CAMCOORD1'
    write(*, *) xmilli, ymilli, zmilli
    call CAL_camCoord1ToChipCoord(xmilli, ymilli, zmilli, rawX, rawY, ccd)
    do i=0, size(rawXtest)-1
       write(*, *) i, rawX(i), rawY(i), ccd(i), rawXtest(i), rawYtest(i)
       if (ccd(i) /= 1) &
            call error('caltest_emos', "CCD nr incorrect");
       if (rawX(i) /= rawXtest(i) .or. &
           rawY(i) /= rawYtest(i)) &
       call error('caltest_emos', "XY<->mm not correct")
    enddo

    write(*, *) 'CAMCOORD2 -> TELCOORD'
    call CAL_camCoord2ToTelCoord(xmilli, ymilli, zmilli, thetaOut, phiOut)
    do i=0, size(xmilli)-1
        write(*, *) xmilli(i), ymilli(i), zmilli(i), thetaOut(i), phiOut(i)
    enddo

    call CAL_releaseMemory(xmilli, ymilli, zmilli)
    call CAL_releaseMemory(rawX, rawY)

    !
    !   theta/phi <--> mm
    !

    phi0(0) = 1.57   ! radians
    theta0(0) = 0.   ! on-axis case
    call CAL_telCoordToCamCoord2(theta0, phi0, mmX, mmY, mmZ)
    call CAL_camCoord2ToCamCoord1(mmX, mmY, mmZ, mmX, mmY, mmZ)
    write(*, *) mmX(0), mmY(0), mmZ(0)
    call CAL_camCoord1ToChipCoord(mmX, mmY, mmZ, rawX, rawY)
    write(*, *) rawX(0), int(CAL_getMiscellaneousDataValue('OPTICS_X'))
    write(*, *) rawY(0), int(CAL_getMiscellaneousDataValue('OPTICS_X'))
    write(*, *) CAL_getMiscellaneousDataValue('LINCOORD::LINCOORD%ARC_MM_Y')
    if (rawX(0) /= int(CAL_getMiscellaneousDataValue('OPTICS_X')) .or. &
        rawY(0) /= int(CAL_getMiscellaneousDataValue('OPTICS_Y'))) &
        call error('caltest_emos', 'something wrong with OPTICS_X/Y')

    phi0(0) = 1.9    ! radians
    theta0(0) = 20.3 ! arcsecs
    call CAL_telCoordToCamCoord2(theta0, phi0, mmX, mmY, mmZ)
    call CAL_camCoord2ToTelCoord(mmX, mmY, mmZ, theta, phi)
    write(*, *) theta0(0), theta(0)
    write(*, *) phi0(0), phi(0)
    if (.not.isClose(real(theta(0), single), real(theta0(0), single)) .or. &
        .not.isClose(real(phi(0), single), real(phi0(0), single))) &
        call error('caltest_emos', 'problem with thetaPhi <-> mm conversion')

    !
    !   CTI correction
    !
    call CAL_setState(ccdChipId = EMOS_CCD_CENTRAL, ccdNodeId = 0)
    rawXtest(0) = EMOS_MIN_X
    rawXtest(1) = EMOS_MAX_X
    rawXtest(2) = EMOS_MAX_X
    rawXtest(3) = EMOS_MIN_X
    rawXtest(4) = EMOS_MIN_X
    rawYtest(0) = EMOS_MIN_Y
    rawYtest(1) = EMOS_MIN_Y
    rawYtest(2) = EMOS_MAX_Y
    rawYtest(3) = EMOS_MAX_Y
    rawYtest(4) = EMOS_MIN_Y
    call CAL_integerToReal(Ein, EinReal)
    call CAL_mosCtiCorrect(timeTags, rawXtest, rawYtest, EinReal, PHAreal)
    call CAL_realToInteger(PHAreal, PHAint)
    write(*, *) Ein
    write(*, *) PHAint
    if (PHAint(0) /= 123 .or. &
        PHAint(1) /= 459 .or. &
        PHAint(2) /= 793 .or. &
        PHAint(3) /= 999 .or. &
        PHAint(4) /= 364 .or. &
        PHAint(5) /= 892 .or. &
        PHAint(6) /= 1002) &
        call error('caltest_emos', "CTI correction does not work")

    !
    !   gain correction
    !
    call CAL_mosGainCorrect(timeTags, pattern, PHAreal, PIreal)
    call CAL_realToInteger(PIreal, PIint)
    write(*, *) PHAint
    write(*, *) PIint
    if (PIint(0) /= 404 .or. &
        PIint(1) /= 1505 .or. &
        PIint(2) /= 2601 .or. &
        PIint(3) /= 3207 .or. &
        PIint(4) /= 1159 .or. &
        PIint(5) /= 2876 .or. &
        PIint(6) /= 3234) &
            call error('caltest_emos', "gain correction does not work")
    call CAL_releaseMemory(EinReal, PHAreal, PHAreal)
    call CAL_releaseMemory(PHAint, PIint)

    !
    !   get PSF map and check a window
    !
    psf = CAL_getPSF(1500._double, 0._double, 0._double);
    window%halfX = 1._double
    window%halfY = 1._double
    call CAL_psfImage(psf, window, PSFmap)

    nx = size(PSFmap, 1)/2
    ny = size(PSFmap, 2)/2
    write(*, *) PSFmap(nx, ny)
    write(*, *) PSFmap(nx+1, ny+1)
    write(*, *) PSFmap(nx+2, ny+2)
    write(*, *) PSFmap(nx+3, ny+3)
    write(*, *) PSFmap(nx+4, ny+4)

    if (.not. isClose(PSFmap(nx, ny), 7.3609924) .or. &
        .not. isClose(PSFmap(nx+1, ny+1), 6.6444502) .or. &
        .not. isClose(PSFmap(nx+2, ny+2), 5.1652832) .or. &
        .not. isClose(PSFmap(nx+3, ny+3), 3.4581327) .or. &
        .not. isClose(PSFmap(nx+4, ny+4), 1.9938920)) &
                             call error('caltest_emos', "PSF is wrong")
    call CAL_releaseMemory(PSFmap)

    !
    !   get the event pattern library
    !
    write(*, *) 'Imaging mode event patterns'
    call CAL_getEventPatterns(imagPatterns, eduThreshold)
    do i=0, size(imagPatterns, 1)-1
        write(*, '(a10,I3)') 'Pattern', i
        do j=0, size(imagPatterns, 2)-1
            write (*, '(5I2)') (imagPatterns(i, j, k), &
                                        k=0, size(imagPatterns, 3)-1)
        enddo
    enddo
    if (size(imagPatterns, 1) /= 32 .or. size(imagPatterns, 2) /= 5 .or. &
        size(imagPatterns, 3) /= 5) &
        call error('caltest_emos', 'wrong size of pattern array')
    if (imagPatterns(27, 0, 0) /= 0 .or. &
        imagPatterns(27, 0, 4) /= 0 .or. &
        imagPatterns(27, 1, 1) /= 2 .or. &
        imagPatterns(27, 2, 2) /= 1 .or. &
        imagPatterns(27, 3, 3) /= 2 .or. &
        imagPatterns(27, 4, 4) /= 0 .or. &
        imagPatterns(27, 3, 1) /= 1 .or. &
        imagPatterns(27, 4, 0) /= 2 .or. &
        imagPatterns(27, 4, 3) /= 0) call error('caltest_emos', &
                                                'event image patterns')

    !
    !   mode: Compressed Timing
    !
    write(*, *) 'Compressed-timing mode event patterns'
    call CAL_setState(ccdModeId = FAST_COMPRESSED)
    call CAL_getEventPatterns(timingPatterns, eduThreshold)
    do i=0, size(timingPatterns, 1)-1
        write(*, '(a10,I3)') 'Pattern', i
        do j=0, size(timingPatterns, 2)-1
            write (*, '(5I2)') (timingPatterns(i, j, k), &
                                        k=0, size(timingPatterns, 3)-1)
        enddo
    enddo
    if (timingPatterns(2, 0, 0) /= 0 .or. &
        timingPatterns(2, 0, 4) /= 0 .or. &
        timingPatterns(2, 2, 0) /= 0 .or. &
        timingPatterns(2, 2, 1) /= 2 .or. &
        timingPatterns(2, 2, 2) /= 1 .or. &
        timingPatterns(2, 2, 3) /= 1 .or. &
        timingPatterns(2, 2, 4) /= 1 .or. &
        timingPatterns(2, 4, 0) /= 0 .or. &
        timingPatterns(2, 4, 3) /= 0) call error('caltest_emos', &
                                                'event timing patterns')

    !
    !   offset vectors and darkframes
    !
    do modeNr=PRIME_FULL_WINDOW, PRIME_PARTIAL_W2
        call CAL_setState(ccdModeId = modeNr)
        call CAL_getMOSoffsets(offX, offY)
        write(*, *) 'MOS offset vectors in mode ', modeId(modeNr)
        write(*, *) 'offX(', size(offX), '):'
        write(*, *) (offX(i), i=0, 9), ' ...'
        write(*, *) 'offY(', size(offY), '):'
        write(*, *) (offY(i), i=0, 9), ' ...'
        call CAL_releaseMemory(offX, offY)
    enddo
    call CAL_setState(ccdChipId=7, ccdModeId = PRIME_FULL_WINDOW)
    call CAL_getMOSdarkFrameMap(darkFrame)
    write(*, *) 'dark frame map - extent: ', size(darkFrame, 1), &
                                             size(darkFrame, 2)
!    do i=0, 9
!        write(*, *) (darkFrame(j, i), j=0, 9)
!    enddo
    do i=427, 431
        write(*, *) (darkFrame(j, i), j=391, 395)
    enddo
    

    call CAL_releaseMemory(darkFrame)


    !
    !   effective area + vignetting
    !
    area = CAL_getEffectiveArea(1500._double, 0._double, 0._double)
    write (*, *) 'on-axis effective area at 1500 eV: ', area
    if (.not. isClose(665.16754432175162, real(area, single))) &
        call error('caltest_emos', "effective area is wrong")
    vign = CAL_getVignettingFactor(1500._double, 500._double, 0._double)
    write (*, *) 'vignetting at 1500 eV - 500arcsec off-axis: ', vign
    if (.not. isClose(0.6870442041437571, real(vign, single))) &
        call error('caltest_emos', "vignetting is wrong")

    call message(AppMsg, Verbose, "all tests passed")
end program

 
logical function isClose(value, trueval)
use types

    real(single), intent(in)    :: value, trueval
    real(single)                :: absval, abstrueval

    isClose = .true.
    if (value * trueval .lt. 0.) isClose = .false.
    absval = abs(value)
     abstrueval = abs(trueval)
    if (absval .lt. .9999 * abstrueval .or. &
        absval .gt. 1.001 * abstrueval) isClose = .false.
end function isClose
