//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Boresight.cc --- implementation of Boresight class
//
//	DESCRIPTION:
//		This implements all methods declared in the Boresight class
//
//	CHANGE HISTORY:
//		UL, 1999-08-17: 1.0.0: - first version from CalAtomsCommon
//		UL, 2001-03-02: 1.1.0: - added withSacCoordOrientation()
//							   - added projectOntoTelCoord()
//
//	AUTHOR:
//		U. Lammers (UL), Tue Aug 17 11:54:35 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/Boresight.cc,v 1.11 2001/10/29 16:57:55 ulammers Exp $
//

#include <cmath>
#include "CalOalUtility.h"
#include "CalServer.h"
#include "Boresight.h"
#include "GeometryDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( Boresight )--------------------------------------------------
//
Boresight::Boresight(const EulerAngles &bsAngles,
					 CalAtomCollection<Boresight> *parent)
	: CalChild<CalAtomCollection<Boresight> >(parent), _angles(bsAngles),
		_mIsValidForTime(-1.), _anglesIsValidForTime(-1.),
		_saccoordDefined(false)
{
	PROC

	_m = _angles.toMatrix();

	ENDPROC
} // Boresight::Boresight()


const EulerAngles &Boresight::angles(double time)
{
	PROC

	_anglesIsValidForTime = time;

	RETURN _angles;
} // Boresight::angles()


void Boresight::angles(const EulerAngles &eulerAngles)
{
	PROC

	_angles = eulerAngles;
	_m = _angles.toMatrix();

	ENDPROC
} // Boresight::angles()


const AttitudeMatrix &Boresight::matrix(double time) const
{
	PROC

	_mIsValidForTime = time;

	RETURN _m;
} // Boresight::matrix()


const Boresight &Boresight::withSacCoordOrientation(
										const SpacecraftAttitude &att) const
{
	PROC

	_saccoordMatrix = att.toAttitudeMatrix();
	_saccoordDefined = true;

	RETURN *this;
} // Boresight::withSacCoordOrientation()


XmmTelescopePointList Boresight::projectOntoTelCoord(
										const EquatorialDirection &dir) const
{
	PROC

	vector<EquatorialDirection> v;
	v.push_back(dir);

	RETURN projectOntoTelCoord(v);
} // projectOntoTelCoord()


XmmTelescopePointList Boresight::projectOntoTelCoord(
								const vector<EquatorialDirection> &dl) const
{
	PROC

	if (!_saccoordDefined)
		errstr<<"Define first orientation of SACCORD"
			  <<Sas::error("SacCoordOrientationUndefined");

	AttitudeMatrix BSA = matrix(0.) * _saccoordMatrix;

	unsigned N = dl.size();
	CalReal64Vector	theta(0, N), phi(0, N);
	for (unsigned n=0; n<N; ++n) {
		CartesianVector srcVec = dl[n].toUnitCartesian(),
						v = BSA * srcVec;
		theta[n] = acos(v.x())*RAD_TO_DEG*3600.,
		phi[n] = fmod(3./2.*M_PI + atan2(v.z(), v.y()), M_TWO_PI);
	}

	GeometryDataServer *gds = calServer.getAtom(gds);

	RETURN gds->telCoordList(theta, phi);
} // Boresight::projectOntoTelCoord()


CalAtomCollection<Boresight>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	Boresight *null = 0;
	_collection = new vector<Boresight *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::BoreSight);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet,
							e->queryBlockName(Ccf::toString(Ccf::BoreSight)));
	const Table *table = ccftab.block();
	ColumnData *cdex = table->column("EULER_PHI")->data(),
			   *cdey = table->column("EULER_THETA")->data(),
			   *cdez = table->column("EULER_PSI")->data(),
			   *cddesc = table->column("INSTRUMENT_ID")->data();
	real64	*ex = cdex->real64Data(),
			*ey = cdey->real64Data(),
			*ez = cdez->real64Data();

	EulerAngles ea(0., 0., 0., EulerAngles::Euler321);
	//	look for the row corresponding to current instrument
	string instrString = Xmm::toString(instrument->id());
	bool read = false;
	for (unsigned n=0; n<table->rows(); ++n)
		if (stringColumnRow(cddesc, n) == instrString) {
			//	row corresponding to current instrument found
			ea.phi(ex[n]);
			ea.theta(ey[n]);
			ea.psi(ez[n]);
			read = true;
			break;
		}

	if (!read)
		errstr<<"The CCF boresight misalignment file "
			  <<dataSet->qualifiedName()
			  <<" does not contain an entry for "
			  <<Xmm::toString(instrument->id())
			  <<Sas::error("NoSuchEntry");

	Boresight *&bsm = (*_collection)[0];
	if (!bsm)
		bsm = new Boresight(ea, this);

	delete cdex;
	delete cdey;
	delete cdez;
	delete cddesc;

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<Boresight>::CalAtomCollection()



Boresight *CalAtomCollection<Boresight>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[0];
} // CalAtomCollection<Boresight>::select()
