//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalAtomsMisc.cc --- implementation of miscallaneous CalAtomsCommon classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1997-07-25: 1.0.0: - skeleton
//		UL, 1997-??-??: 1.1.0: - extended
//		UL, 1998-08-28: 1.2.0: - EnergyCorrector added
//		UL, 1999-03-05: 1.3.0: - changes enforced by lowmem Dal
//		UL, 1999-05-11: 1.4.0: - channel range of Ebounds/Ebins is now
//								 0-4095 for EMOS+EPN
//							   - changed fatal -> error and changed all
//								 error tags
//		UL, 2000-11-09: 1.5.0: - changes to cope with vector columns in
//								 CCF::MODEPARAM; this fixes SPR 1827:
//								 Does not recognize PrimeFullWindow for MOS 1
//		UL, 2001-02-19: 1.6.0: - moved BadPixel related stuff into separate
//								 file
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/CalAtomsMisc.cc,v 1.2 2001/10/29 16:57:55 ulammers Exp $
//

#include "CalAtomsMisc.h"
#include "CalOalUtility.h"
#include "CalServer.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

//
//	----------------------( mode parameters )-----------------------------
//
CalAtomCollection<ModeParameters>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	PROC
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
	_nEntries	= _nModes;
	ModeParameters *null = 0;
	_collection = new vector<ModeParameters *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::ModeParam);
	DataSet *dataSet = e->open();
	Table *table = dataSet->table(
					e->queryBlockName(Ccf::toString(Ccf::ModeParam)));

	ColumnData *cdmodeId = table->column("MODE_ID")->data(),
			   *cdintTime = table->column("INTEGRATION_T")->data(),
			   *cdshiftTime = table->column("SHIFT_T")->data(),
			   *cdframeTime = table->column("FRAME_T")->data(),
			   *cdcycleTime = table->column("CYCLE_T")->data(),
			   *cdtimeRes = table->column("TIME_RES")->data(),
			   *cddutyCycle = table->column("DUTY_CYCLE")->data(),
			   *cdsmearedFrac = table->column("SMEARED_FRAC")->data(),
			   *cdpntSrcPileUp = table->column("PTSRCE_PILEUP")->data(),
			   *cdlimitingFlux = table->column("LIMITING_FLUX")->data(),
			   *cdwx0 = table->column("WINDOWX0")->data(),
			   *cdwy0 = table->column("WINDOWY0")->data(),
			   *cdwdx = table->column("WINDOWDX")->data(),
			   *cdwdy = table->column("WINDOWDY")->data();
	int8	*modeId = cdmodeId->int8Data();
	real32	*intTime = cdintTime->real32Data(),
			*shiftTime = cdshiftTime->real32Data(),
			*frameTime = cdframeTime->real32Data(),
			*cycleTime = cdcycleTime->real32Data(),
			*timeRes = cdtimeRes->real32Data(),
			*dutyCycle = cddutyCycle->real32Data(),
			*smearedFrac = cdsmearedFrac->real32Data(),
			*pntSrcPileUp = cdpntSrcPileUp->real32Data(),
			*limitingFlux = cdlimitingFlux->real32Data();
	int16	*wx0 = cdwx0->int16Data(),
			*wy0 = cdwy0->int16Data(),
			*wdx = cdwdx->int16Data(),
			*wdy = cdwdy->int16Data();

	short underX = table->attribute("UNDER_X")->asInt(),
		  overX = table->attribute("OVER_X")->asInt(),
		  underY = table->attribute("UNDER_Y")->asInt(),
		  overY = table->attribute("OVER_Y")->asInt();
	for (unsigned n=0; n<table->rows(); ++n) {
		if (modeId[n] > INT(_nModes))
			errstr<<"Illegal mode number in row #"
				  <<n<<" of CCF file "
				  <<Xmm::toString(instrument->id())<<"/"
				  <<Ccf::toString(Ccf::ModeParam)
			  <<Sas::error("IllegalValues");

		unsigned pos = index(modeId[n]);

		if (!(*_collection)[pos])
			(*_collection)[pos] = new ModeParameters;

		(*_collection)[n] = new ModeParameters;

		#define VALUE(a)	a[cd##a->position(n, 0)]

		(*_collection)[n]->_integrationTime = VALUE(intTime);
		(*_collection)[n]->_shiftTime = VALUE(shiftTime);
		(*_collection)[n]->_frameTime = VALUE(frameTime);
		(*_collection)[n]->_cycleTime = VALUE(cycleTime);
		(*_collection)[n]->_timeResolution = VALUE(timeRes);
		(*_collection)[n]->_dutyCycle = VALUE(dutyCycle);
		(*_collection)[n]->_smearedFraction = VALUE(smearedFrac);
		(*_collection)[n]->_pointSourcePileUp = VALUE(pntSrcPileUp);
		(*_collection)[n]->_limitingFlux = VALUE(limitingFlux);
		(*_collection)[n]->_windowX0 = VALUE(wx0);
		(*_collection)[n]->_windowY0 = VALUE(wy0);
		(*_collection)[n]->_windowDX = VALUE(wdx);
		(*_collection)[n]->_windowDY = VALUE(wdy);
		(*_collection)[n]->_underX = underX;
		(*_collection)[n]->_overX = overX;
		(*_collection)[n]->_underY = underY;
		(*_collection)[n]->_overY = overY;

		#undef VALUE
	}
	delete cdmodeId;
	delete cdintTime;
	delete cdshiftTime;
	delete cdframeTime;
	delete cdcycleTime;
	delete cdtimeRes;
	delete cddutyCycle;
	delete cdsmearedFrac;
	delete cdpntSrcPileUp;
	delete cdlimitingFlux;
	delete cdwx0;
	delete cdwy0;
	delete cdwdx;
	delete cdwdy;
	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<ModeParameters>::CalAtomCollection


ModeParameters *CalAtomCollection<ModeParameters>::select(
					Ccd,
					Node,
					unsigned modeOffset) const
{
	PROC

	unsigned n = index(modeOffset);

	RETURN (*_collection)[n];
} // CalAtomCollection<ModeParameters>::select


//
//	----------------------( Background )-----------------------------------
//
CalAtomCollection<Background>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	PROC
	_filter		= filter;
	_accuracy	= accuracy;

	_nCCDs	= instrument->nCcds();
	_nNodes	= instrument->nNodes();
	_nModes	= instrument->nModes();
	assert(_nCCDs && _nNodes && _nModes);
	_nEntries	= 1;
	Background	*null = 0;
	_collection = new vector<Background *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::Background);
	DataSet *dataSet = e->open();

	//
	//	read particle background parameters
	//
	CcfConstituentTable ccftab1(dataSet, e->queryBlockName("PARTICLE"));
	const Table *t1 = ccftab1.block();
	ColumnData *cdpar = t1->column("PAR")->data();
	real32	*par = cdpar->real32Data();

	if (!(*_collection)[0])
		(*_collection)[0] = new Background;

	(*_collection)[0]->_particleP0 = par[0];
	(*_collection)[0]->_particleP1 = par[1];

	//
	//	read diffuse X-ray components
	//
	CcfConstituentTable ccftab2(dataSet, e->queryBlockName("X_RAY_DIFFUSE"));
	const Table *t2 = ccftab2.block();
	ColumnData *d = t2->column("SPECTRUM")->data();
	unsigned nChanns = d->elements();

	for (unsigned n=0; n<nChanns; ++n) {
		unsigned long pos = d->position(0, n);
		(*_collection)[0]->_diffuse.pushChannel(d->real32Data()[pos]);
	}

	//
	//	read diffuse X-ray modulation part
	//
	CcfConstituentTable ccftab3(dataSet,
								e->queryBlockName("X_RAY_DIFFUSE_MODULATION"));
	const Table *t3 = ccftab3.block();
	ColumnData *cdt3 = t3->column("PAR")->data();
	par = cdt3->real32Data();
	(*_collection)[0]->_diffModP0 = par[0];
	(*_collection)[0]->_diffModP1 = par[1];
	(*_collection)[0]->_diffModP2 = par[2];

	//	tidy up
	delete cdt3;
	delete d;
	delete cdpar;
	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<Background>::CalAtomCollection


Background *CalAtomCollection<Background>::select(Ccd, Node, unsigned)
const
{
	PROC

	RETURN (*_collection)[0];
} // CalAtomCollection<Background>::select()


BackgroundSpectrum Background::particle(double time = 0, double r = 0.)
const
{
	PROC

	errstr<<"Background::particle: dependence on time ("<<time<<")"
		  <<" and r ("<<r<<") not yet implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	BackgroundSpectrum particleBackground;
	const vector<unsigned> &diffChanns = _diffuse.channels();
	unsigned nChanns = _diffuse.size();
	for (unsigned n=0; n<nChanns; ++n)
		particleBackground.pushChannel(_particleP0 +
									   _particleP1 * DBL(diffChanns[n]));

	RETURN particleBackground;
} // Background::particle()


BackgroundSpectrum Background::diffuseX(double time = 0, double r = 0.) const
{
	PROC

	errstr<<"Background::particle: dependence on time ("<<time<<")"
		  <<" and r ("<<r<<") not yet implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	BackgroundSpectrum diffuseBackground(_diffuse);
	//
	//	modify
	//
	double mod = _diffModP0 + _diffModP1 * r + _diffModP2 * r * r;
	unsigned nChanns = diffuseBackground.size();
	for (unsigned n=0; n<nChanns; ++n)
		diffuseBackground[n] *= mod;

	RETURN diffuseBackground;
} // Background::diffuseX
