//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalServer.cc	--- implementation of CalServer class methods
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1998-04-22: 1.0.0: - first version
//		UL, 1998-11-18: 2.0.0: - overhaul
//		UL, 1999-02-10: 2.1.0: - use templated member functions
//		UL, 1999-03-29: 3.0.0: - overhaul in line with CalServer.h-3.0.0
//		UL, 1999-04-12: 3.1.0: - allow ERM as valid instrument
//		UL, 1999-04-22: 3.1.1: - ugly change: allow different units of FRMTIME
//								 attribute in recognition of EMOS mode
//		UL, 1999-05-12: 3.2.2: - changed fatal -> error and changed all
//								 error/warning tags
//		UL, 1999-09-10: 3.3.0: - added AbsorptionDataServer
//		UL, 2000-01-05: 3.4.0: - retrieval of atoms produces now proper error
//								 message in case of missing CCF data
//		UL, 2000-05-24: 3.5.0: - for figuring out the camera mode read the
//								 CCF mode parameters and pass data to
//								 EMOS/EPNcam::toMode()
//		UL, 2000-09-28: 2.6.0: - setState(): also set the filter variable
//								 (from "FILTER" attribute)
//		UL, 2000-10-20: 2.7.0: - fixed SPR 1667: MOS RFS mode: Outer CCDs are
//								 not in RFS (see corresponding code comment in
//								 CalServer::setState())
//		UL, 2000-11-10: 2.7.1: - some minor changes to eliminate warnings
//								 from optimizer
//		UL, 2001-07-18: 2.7.2: - also allow FILTER="UNDEFINED" in definition
//								 of state
//		UL, 2001-11-26: 2.8.0: - if SUBMODE attribute is present determine
//								 camera mode from value and skip mode
//								 recognition logic
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/CalServer.cc,v 1.60 2001/11/30 16:24:47 ulammers Exp $
//

#include "CalServer.h"
#include "RealCcf.h"
#include "DalUtils.h"
#include "Random.h"

using namespace std;

//
//	---( CalServer )-------------------------------------------------------
//
CalServer::CalServer(Xmm::Instrument instrument) :
	_instrument(0), _state(0), _ccf(0), _atomServer(0),
	_oldRandMin(-.5), _oldRandMax(.5)
{
	PROC

	setInstrument(instrument);
	setRandom();

	ENDPROC
} // CalServer::CalServer()


void CalServer::setInstrument(Xmm::Instrument instrument)
{
	PROC

	if (_instrument && _instrument->id() == instrument)
		//	nothing to change
		ALTRETURN;

	init();

	errstr<<"Setting CAL state to instrument "<<instrument<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	switch (instrument) {
		case Xmm::EMOS1: _instrument	= new EMOScam1; break;
		case Xmm::EMOS2: _instrument	= new EMOScam2; break;
		case Xmm::EPN:	 _instrument	= new EPNcam; break;
		case Xmm::ERM:	 _instrument	= new ERMcam; break;
		case Xmm::RGS1:	 _instrument	= new RGScam1; break;
		case Xmm::RGS2:	 _instrument	= new RGScam2; break;
		case Xmm::OM:	 _instrument	= new OMcam; break;
		case Xmm::SC:	 _instrument	= new XmmInstrument; break;
		default		   : errstr<<"illegal instrument `"
							   <<Xmm::toString(instrument)<<"'"
							   <<Sas::error("IllegalInstrument");
	}
	defState();

	ENDPROC
} // CalServer::setInstrument()


CalState &CalServer::getState() const
{
	PROC

	RETURN *_state;
} // CalServer::getState();


void CalServer::setState(const CalState &calState)
{
	PROC

	delete _state;
	_state = new CalState(calState);

	ENDPROC
} // CalServer::setState()


void CalServer::pushState()
{
	PROC

	_stateStack.push_back(_state);
	defState();

	ENDPROC
} // CalServer::pushState()


void CalServer::popState()
{
	PROC

	if (_stateStack.size()) {
		delete _state;
		_state = _stateStack[_stateStack.size() - 1];
		_stateStack.pop_back();
	} else
		errstr<<"Cannot pop state stack because it is empty - state remains "
				"unchanged"
			  <<Sas::error("OperationNotPermitted");

	ENDPROC
} // CalServer::pushState()


CalState *CalServer::state() const
{
	PROC
	RETURN _state;
} // CalServer::state()


Ccf *CalServer::setCcf(const string &cifName = "") const
{
	PROC

	delete _ccf;
	delete _atomServer;
	_ccf = new RealCcf(cifName, this);
	_atomServer = new CalAtomCollectionServer(this);

	RETURN _ccf;
} // CalServer::setCcf()


Ccf *CalServer::ccf() const
{
	PROC

	checkCcf();

	RETURN _ccf;
} // CalServer::ccf();


void CalServer::setRandom(double min, double max)
{
	PROC

	errstr<<"Setting interval for random number generator to ["
		  <<min<<", "<<max<<"]\n"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	_oldRandMin	= randGen.min();
	_oldRandMax = randGen.max();
	randGen.min(min);
	randGen.max(max);	

	ENDPROC
} // CalServer::setRandom()


void CalServer::setRandom()
{
	PROC

	randGen.min(_oldRandMin);
	randGen.max(_oldRandMax);

	ENDPROC
} // CalServer::setRandom()


CalAtom *CalServer::getAtom(Ccf::CalDataTypes)
{
	PROC

	errstr<<"Not implemented."<<Sas::warning("NotYetImplemeneted");

	checkCcf();
	static BadPixelDataServer *map = getAtom(map);

	RETURN map;
} // CalServer::getAtom()


template<class AtomType>
AtomType *CalServer::getAtom(AtomType *p, bool generateError) const
{
	PROC

	checkCcf();

	if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
		errstr<<"Retrieving CAL atom "<<Xmm::toString(_instrument->id())<<"/"
			  <<Ccf::toString(p->id())<<" corresponding to "
			  <<"CCD  = "<<(_state->ccd()->isActive() ?
							toString(_state->ccd()->get()) : "*inactive*")
						 <<endl
			  <<"node = "<<(_state->node()->isActive() ?
						Xmm::toString(_state->node()->get()) : "*inactive*")
						 <<endl
			  <<"mode = "<<(_state->mode()->isActive() ?
						Xmm::toString(_state->mode()->get()) : "*inactive*")
						 <<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

	CalAtomCollection<AtomType> *ptr =
								_atomServer->getCollection(_instrument, p);

	AtomType *atom = ptr->select(
		_state->ccd()->isActive() ? _state->ccd()->get() : 0,
		_state->node()->isActive() ? _state->node()->get() : Xmm::PRIMARY_NODE,
		_state->mode()->isActive() ?
							_instrument->modeIndex(_state->mode()->get()) : 0);

	if (!atom && generateError)
		errstr<<"The CCF constituent "<<Xmm::toString(_instrument->id())<<"/"
			  <<Ccf::toString(p->id())<<" does not contain an entry "
				"corresponding to "
			  <<"CCD="<<(_state->ccd()->isActive() ?
							toString(_state->ccd()->get()) : "*inactive*")
			  <<"/node="<<(_state->node()->isActive() ?
						Xmm::toString(_state->node()->get()) : "*inactive*")
			  <<"/mode="<<(_state->mode()->isActive() ?
						Xmm::toString(_state->mode()->get()) : "*inactive*")
						 <<endl
			  <<Sas::error("NoSuchEntry");

//	if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose) {
//		atom->dump(errstr);
//		errstr<<endl<<Sas::message(Msg::Lib, Msg::Verbose);
//	}

	RETURN atom;
} // CalServer::getAom()

#define getAtomImp(a)	template a *CalServer::getAtom(a *) const;

getAtomImp(AbsorptionDataServer)
getAtomImp(Background)
getAtomImp(BadPixelDataServer)
getAtomImp(Boresight)
getAtomImp(CanonicalCrossPsf)
getAtomImp(CanonicalRedist)
getAtomImp(CcdQuantumDataServer)
getAtomImp(CtiCorrector)
getAtomImp(DarkFrameDataServer)
getAtomImp(EffectiveAreaDataServer)
getAtomImp(EncircledEnergyDataServer)
getAtomImp(EnergyCorrector)
getAtomImp(FilterDataServer)
getAtomImp(GeometryDataServer)
getAtomImp(GratingDataServer)
getAtomImp(HKvalRangesServer)
getAtomImp(MiscDataServer)
getAtomImp(ModeParameters)
getAtomImp(OmColorTransformator)
getAtomImp(OmFlatFieldServer)
getAtomImp(OmFluxConverter)
getAtomImp(OmGeoLinearizer)
getAtomImp(OmPsf)
getAtomImp(PatternLibServer)
getAtomImp(PsfDataServer)
getAtomImp(TimeCorrector)

#undef getAtomImp

template<class AtomType>
void CalServer::releaseAtom(AtomType *p) const
{
	PROC

	checkCcf();
	_atomServer->releaseCollection(_instrument, p);

	ENDPROC
} // CalServer::releaseAtom()

#define relAtomImp(a)	template void CalServer::releaseAtom(a *) const;

relAtomImp(AbsorptionDataServer)
relAtomImp(Background)
relAtomImp(BadPixelDataServer)
relAtomImp(Boresight)
relAtomImp(CanonicalCrossPsf)
relAtomImp(CanonicalRedist)
relAtomImp(CcdQuantumDataServer)
relAtomImp(GeometryDataServer)
relAtomImp(CtiCorrector)
relAtomImp(DarkFrameDataServer)
relAtomImp(EffectiveAreaDataServer)
relAtomImp(EncircledEnergyDataServer)
relAtomImp(EnergyCorrector)
relAtomImp(FilterDataServer)
relAtomImp(GratingDataServer)
relAtomImp(HKvalRangesServer)
relAtomImp(MiscDataServer)
relAtomImp(ModeParameters)
relAtomImp(OmColorTransformator)
relAtomImp(OmFlatFieldServer)
relAtomImp(OmFluxConverter)
relAtomImp(OmGeoLinearizer)
relAtomImp(OmPsf)
relAtomImp(PatternLibServer)
relAtomImp(PsfDataServer)
relAtomImp(TimeCorrector)

#undef relAtomImp


CalAtomCollectionServer *CalServer::atomServer() const
{
	PROC

	checkCcf();	

	RETURN _atomServer;
} // CalServer::atomServer()


void CalServer::print(ostream &os) const
{
	PROC

	os<<"   instrument             : "<<UNT(_instrument->id())<<
				"["<<Xmm::toString(_instrument->id())<<"]\n"<<
		*_state<<endl;

	ENDPROC
} // CalServer::print()


CalInt16Vector CalServer::int16Vector(const int16 *ptr, unsigned n)
{
	PROC
	RETURN CalInt16Vector(const_cast<int16 *>(ptr), n, this);
} // CalServer::int16Vector()


CalReal32Vector CalServer::real32Vector(const real32 *ptr, unsigned n)
{
	PROC
	RETURN CalReal32Vector(const_cast<real32 *>(ptr), n, this);
} // CalServer::real32Vector()


CalServer::~CalServer()
{
	PROC

	errstr<<"CCF constituents accessed during the livetime of the CAL:\n"
		  <<ccfAccessLog()<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	init();
	delete _atomServer;
	delete _ccf;
	//	restore RNG to what it was
	randGen.min(_oldRandMin);
	randGen.max(_oldRandMax);

	ENDPROC
} // CalServer::~CalServer()


void CalServer::defState()
{
	PROC
	switch (_instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2: _state			= new CalStateEmos(this);
						 break;
		case Xmm::EPN:	 _state			= new CalStateEpn(this);
						 break;
		case Xmm::RGS1:
		case Xmm::RGS2:	 _state			= new CalStateRgs(this);
						 break;
		case Xmm::OM:	 _state			= new CalStateOm(this);
						 break;
		case Xmm::ERM:
		case Xmm::SC:	 _state			= new CalState(this);
						 break;
						
		default		   : errstr<<"illegal instrument `"
							   <<Xmm::toString(_instrument->id())<<"'"
							   <<Sas::error("IllegalInstrument");
	}
} // CalServer::defState()


void CalServer::init()
{
	PROC

	delete _instrument;	_instrument = 0;
	delete _state; _state = 0;
	while (_stateStack.size()) {
		delete _stateStack[_stateStack.size()-1];
		_stateStack.pop_back();
	}

	ENDPROC
} // CalServer::init()


void CalServer::checkCcf() const
{
	PROC

	if (!_ccf)
		//	no Ccf defined yet - instantiate one first
		setCcf();

	ENDPROC
} // CalServer::checkCcf()


vector<CoreModeParameters> CalServer::coreModeParameters()
{
	PROC

	pushState();
	vector<CoreModeParameters> v;
	const vector<Mode> &modes = _instrument->modes();
	for (unsigned n=0; n<modes.size(); ++n) {
		_state->mode()->set(modes[n]);
		ModeParameters *modeParams = getAtom(modeParams, false);
		if (!modeParams)
			break;
		CoreModeParameters cmp;
		cmp.windowX0 = modeParams->windowX0();
		cmp.windowY0 = modeParams->windowY0();
		cmp.windowDX = modeParams->windowDX();
		cmp.windowDY = modeParams->windowDY();
		cmp.frameTime = UNT(modeParams->frameTime());
		v.push_back(cmp);
	}
	popState();

	RETURN v;
} // CalServer::coreModeParameters()

namespace {
ostream &sciwarn(const string &listOfAttribs,
				 const string &statevar, bool silent)
{
	PROC

	if (!silent)
		errstr<<"In defining the Cal state from a data set or block "
				"(one or more of) the mandatory attribute(s) "
			  <<listOfAttribs
			  <<" was/were not found - from this/these attribute(s) the "
			  <<"actual "<<statevar
			  <<" is derived which is now not possible - will "
				"assume a default value for the "<<statevar<<" - please check "
				"the diagnostic output below (requires verbosity >= Verbose) "
				"for the effective Cal state settings. Please note: "
				"In case of a possibly incorrect "<<statevar<<" the retrieval "
				"of "<<statevar<<"-dependent calibration quantities via the "
				"Cal is likely to yield INADEQUATE data which ultimately might"
				" lead to INCORRECT scientific results"
			  <<Sas::warning("MandatoryAttributeNotFound");

	RETURN errstr;
} // sciwarn()
}

void CalServer::setState(const vector<const Attributable *> &av, bool silent,
						 bool ignoreFilter)
{
	PROC

	//
	//	see if we have an XMM data set
	//
	try {
		string s(DalUtils::attribute(av, "TELESCOP")->asString());
		if (s != "XMM")
			errstr<<"The data set used to define the Cal state does not seem "
					"to be an XMM set - the value of the attribute "
					"`TELESCOP\' is: "<<s
				  <<Sas::error("IllegalAttributeValue");
	} catch(...) {
		errstr<<"No attribute named `TELESCOP\' found"
			  <<Sas::error("MandatoryAttributeNotFound");
	}

	//
	//	determine the instrument
	//
	Xmm::Instrument instr = Xmm::EMOS1;
	try {
		instr = Xmm::toInstrument(DalUtils::attribute(av,
													"INSTRUME")->asString());
	} catch(...) {
		errstr<<"The data set used to define the Cal state does not contain "
				"an attribute named `INSTRUME\' - this is needed to determine "
				"the current instrument."
			  <<Sas::error("MandatoryAttributeNotFound");
	}
	setInstrument(instr);

	//
	//	at this point, the state variables have their default values
	//
	CalState *state = this->state();

	//
	//	check if there is SUBMODE
	//
	bool modeSet = false;
	try {
		state->mode()->set(Xmm::toMode(
							DalUtils::attribute(av, "SUBMODE")->asString()));
		modeSet = true;
	} catch(...) {}

	//	instrument-specific state settings
	switch(instr) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
			try {
				state->ccd()->set(DalUtils::attribute(av, "CCDID")->asInt());
			} catch(...) {
				sciwarn("`CCDID\'", "CCD", silent);
			}
			try {
				state->node()->
					set(Xmm::toNode(DalUtils::attribute(av,
														"CCDNODE")->asInt()));
			} catch(...) {
				sciwarn("`CCDNODE'", "node", silent);
			}

			if (!modeSet) {
				pair<Mode, bool> mode = EMOScam::toMode(av,
												coreModeParameters(),
												state->ccd()->get(),
												silent);
				if (state->ccd()->get() > 1 &&
					mode.first == Xmm::PRIME_PARTIAL_RFS)
					//
					//	For camera-RFS mode the inner CCD is operated
					//	in Xmm::PRIME_PARTIAL_RFS and the outer ones
					//	in Xmm::PRIME_FULL_WINDOW mode; however, the
					//	ODS for all CCDs wrongly puts the RFS frame
					//	time in the event list, hence, the mode detection
					//	logic gets fooled and wrongly yields
					//	Xmm::PRIME_PARTIAL_RFS for the outer CCDs
					//
					mode.first = Xmm::PRIME_FULL_WINDOW;
				state->mode()->set(mode.first, !mode.second);
			}
			break;
		case Xmm::EPN:
			try {
				state->ccd()->set(EPNcam::toChipNr(
								DalUtils::attribute(av, "CCDID")->asInt(),
								DalUtils::attribute(av, "QUADRANT")->asInt()));
			} catch(...) {
				sciwarn("`CCDID\', `QUADRANT\'", "CCD", silent);
			}

			if (!modeSet) {
				pair<Mode, bool> mode = EPNcam::toMode(av,
													   coreModeParameters(),
													   silent);
				state->mode()->set(mode.first, !mode.second);
			}
			break;
		case Xmm::RGS1:
		case Xmm::RGS2:
			try {
				state->ccd()->set(DalUtils::attribute(av, "CCDID")->asInt());
			} catch (...) {
				sciwarn("`CCDID\'", "CCD", silent);
			}

			try {
				state->ocb()->set(DalUtils::attribute(av, "CCDOCB")->asInt());
			} catch (...) {
				sciwarn("`CCDOCB\'", "OCB", silent);
			}

			if (!modeSet) {
				pair<Mode, bool> mode = RGScam::toMode(av,
													   state->ocb()->get(),
													   silent);
				state->mode()->set(mode.first, !mode.second);
			}
			break;
		case Xmm::OM:
			state->ccd()->set(1);	//	lovely, there is only one CCD
			if (!modeSet) {
				pair<Mode, bool> mode = OMcam::toMode(av, silent);
				state->mode()->set(mode.first, !mode.second);
			}
			break;
		default:
			errstr<<"Initialization from `"<<Xmm::toString(instr)<<"' not "
					"supported"<<Sas::error("IllegalInstrument");
			break;
	}

	//
	//	set filter
	//
	if (state->filter()->isActive()) {
		if (!ignoreFilter)
			try {
				string filter = DalUtils::attribute(av, "FILTER")->asString();
				if (filter.empty() || uppercase(filter) == "UNKNOWN"
								   || uppercase(filter) == "UNDEFINED") {
					errstr<<"Value of FILTER attribute is `"<<filter
						  <<"' - cannot determine active filter"
						  <<Sas::warning("FilterUnknown");
					state->filter()->uncertain();
				} else
					state->filter()->set(Xmm::toFilter(filter));
			} catch(...) {
				if (!silent)
					sciwarn("`FILTER\'", "filter", silent);
				state->filter()->uncertain();
			}
		else
			state->filter()->uncertain();
	}

	//
	//	observation start time
	//
	try {
		state->observationStartTime()->set(STime(
							DalUtils::attribute(av, "DATE-OBS")->asString()));
	} catch(...) {
		sciwarn("`DATE-OBS\'", "observation start date", silent);
	}

	//
	//	all done, output distagnostic messges
	//
	errstr<<"Cal state set to:\n"
		  <<"   instrument             : "<<instr<<endl
		  <<*state<<Sas::message(Msg::Lib, Msg::Verbose);

	ENDPROC;
} // CalServer::setState()


void CalServer::setState(const string &dataSet, bool silent, bool ignoreFilter)
{
	PROC

	DataSet *set = dataSetServer->open(dataSet, DataSet::Read);
	setState(*set, silent, ignoreFilter);
	dataSetServer->close(set);

	ENDPROC
} // CalServer::setState()


void CalServer::setState(const Block &block, bool silent, bool ignoreFilter)
{
	PROC

	vector<const Attributable *> av;
	av.push_back(&block);
	setState(av, silent, ignoreFilter);

	ENDPROC
} // CalServer::setState()


void CalServer::setState(const DataSet &dataSet, bool silent,
						 bool ignoreFilter)
{
	PROC

	vector<const Attributable *> av;
	av.push_back(&dataSet);
	av.push_back(dataSet.block(0));
	setState(av, silent, ignoreFilter);

	ENDPROC
} // CalServer::setState()


CalServer &operator<<(CalServer &server, const string &dataSet)
{
	PROC

	server.setState(dataSet);
	RETURN server;
} // operator<<


CalServer &operator<<(CalServer &server, const Block &block)
{
	PROC

	server.setState(block);

	RETURN server;
} // CalServer::operator<<


CalServer &operator<<(CalServer &server, const DataSet &dataSet)
{
	PROC

	server.setState(dataSet);

	RETURN server;
} // CalServer::operator<<


CalServer &operator<<(CalServer &server,
					  const vector<const Attributable *> av)
{
	PROC

	server.setState(av);
	RETURN server;
} // operator<<


ostream &operator<<(ostream &os, CalServer &server)
{
	PROC

	server.print(os);

	RETURN os;
} // operator<<


CalServer &calServerFunc()
{
	static CalServer *calServerInstance = new CalServer;
	return *calServerInstance;
} // calServerFunc()
