//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					  (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Ccf.cc	--- interface between the CAL and the CIF in the CCF
//
//	DESCRIPTION:
//		This defines the interface between the CAL and the CIF in the CCF
//
//	CHANGE HISTORY:
//		UL, 1997-10-20: 1.0.0: - skeleton
//		UL, 1998-??-??: 1.1.0: - working version
//		UL, 1999-02-10: 1.1.1: - infer scope of CIF constituent from contents
//								 of "SCOPE" attribute (was: "INSTRUME")
//		UL, 1999-05-12: 1.1.2: - changed fatal -> error and changed all
//								 error/warning tags
//		UL, 1999-10-11: 1.1.3: - no longer check value of CATEGORY keyword
//								 against "XMMCCF"; print value as message
//								 if SAS_VERBOSITY >= VERBOSE
//		UL, 2000-02-03: 1.2.0: - make presence of EXTSEQU/EXTSEQID in CCF
//								 override constituents optional; if they
//								 are not found: EXTSEQU = 'ANY' and
//								 EXTSEQUID = <list of block names>
//		UL, 2000-04-27: 1.3.0: - be robust against VALDATE/DATE possibly
//								 containing invalid dates
//		UL, 2000-05-23: 1.4.0: - SUBDATE is now optional
//		UL, 2000-10-03: 1.5.0: - various changes related to support for
//								 CCF constituent search path (SAS_CCFPATH)
//		UL, 2000-12-18: 1.6.0: - added class CcfConstituentBlock
//		UL, 2001-03-12: 1.7.0: - fixed SPR 1949: Wrong verbosity level
//		UL, 2001-03-15: 1.8.0: - changed error->fatal to signal that
//								 further execution does not make sense
//		UL, 2001-09-27: 1.9.0: - bug fix: if a CCF constituent was specified
//								 with path component (in CIF or as replacement
//								 on command line) it was nevertheless searched
//								 along SAS_CCFPATH
//
//	AUTHOR:
//		U. Lammers, Tue Oct 14 14:45:35 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Indexing of the XMM Current Calibration File, U. Lammers,
//			 15/07/97
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/Ccf.cc,v 1.27 2001/10/29 16:57:56 ulammers Exp $
//

#include <fstream>
#include <iostream>
#include <string>
#include <sys/types.h>
#include <dirent.h>

#include "CalServer.h"
#include "common.h"
#include "stringutil.h"
#include "CalOalUtility.h"
#include "StringIntMap.h"
#include "Ccf.h"
#include "Dal.h"
#include "Tracer.h"
#include "FileLocator.h"
#include "errstr.h"

using namespace std;

//
//	CIF entries
//

//
//	assignment op/copy constructor
//
CifEntry &CifEntry::operator=(const CifEntry &entry)
{
	PROC

	if (this == &entry)
		ALTRETURNV(*this);

	_instrument = entry._instrument;
	_calDataType = entry._calDataType;
	_issue = entry._issue;
	_fileName = entry._fileName;
	_validityDate = entry._validityDate;
	_submissionDate = entry._submissionDate;
	_extensionKey = entry._extensionKey;
	_extensionSequence = entry._extensionSequence;

	RETURN *this;
} // CifEntry::operator=()	


CifEntry::CifEntry(const CifEntry &entry) : CalChild<Ccf>(entry.parent())
{
	PROC

	operator=(entry);

	ENDPROC
} // CifEntry::CifEntry()

//
//	construct from calibration file name
//
CifEntry::CifEntry(const string calibrationFileName, const Ccf *ccf)
	: CalChild<Ccf>(ccf)
{
	PROC

	if (!ccf)
		parent(calServer.ccf());

	//	open the data set
	_fileName			= calibrationFileName;
	DataSet *set = open(DataSetServer::Low, false);

	errstr<<"CATEGORY of CCF constituent `"<<set->qualifiedName()<<"': "
		  <<set->attribute("CATEGORY")->asString()
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	_instrument			= Xmm::toInstrument(set->attribute("SCOPE")->
							asString());
	_calDataType		= Ccf::toTypeId(set->attribute("TYPEID")->
							asString());
	_issue				= set->attribute("ISSUE")->asInt();
	string validityDate = set->attribute("VALDATE")->asString();
	if (STime::isValidTime(validityDate))
		_validityDate		= STime(validityDate);

	if (set->hasAttribute("SUBDATE")) {
		string submissionDate(set->attribute("SUBDATE")->asString());
		if (STime::isValidTime(submissionDate))
			_submissionDate	= STime(submissionDate);
	} else
		errstr<<"CCF constituent `"<<set->qualifiedName()<<"' has "
				"no SUBDATE"<<Sas::warning("NoSUBDATEKeyword");

	if (set->hasAttribute("EXTSEQU") && set->hasAttribute("EXTSEQID")) {	
		_extensionKey		= set->attribute("EXTSEQU")->asString();
		_extensionSequence	= set->attribute("EXTSEQID")->asString();
	} else {
		errstr<<"CCF constituent `"<<set->qualifiedName()<<"' does not "
				"contain EXTSEQU/EXTSEQID keywords"
			  <<Sas::warning("MissingKeywords");
		_extensionKey = "ANY";
		unsigned N = set->blocks();
		for (unsigned n=0; n<N-1; ++n)
			_extensionSequence += set->block(n)->name() + ' ';
		_extensionSequence += set->block(N-1)->name();
	}

	//	close data set
	dataSetServer->close(set);

	ENDPROC
} // CifEntry::CifEntry


bool CifEntry::hasBlock(const string &truncatedBlockName) const
{
	PROC
	StringIntMap<int> _map(_extensionSequence);
	RETURN _map[truncatedBlockName] >= 0; 
} // CifEntry::hasBlock


//	query extension number
string CifEntry::queryBlockName(const string &truncatedBlockName,
								unsigned *blockNr = 0) const
{
	PROC

	string calBlockName;
	StringIntMap<int> _map(_extensionSequence);
	int calBlockId = _map[truncatedBlockName];

	if (calBlockId >= 0)
		//
		//	this extension exists
		//
		if (truncatedBlockName.find("X-", 0, 2) != string::npos ||
			_extensionKey == "ANY")
			//
			//	name starts with "X-" or the key is "ANY", i.e.,
			//	
			calBlockName = truncatedBlockName;
		else
			//
			//	put the extension sequence key in front
			//
			calBlockName = _extensionKey + "-" + truncatedBlockName;
	else
		errstr<<"unknown block `"<<truncatedBlockName
			  <<"'"<<Sas::fatal("UnknownBlock");

	if (blockNr)
		*blockNr = calBlockId;

	RETURN calBlockName;
} // CifEntry::queryBlockName()


bool operator<(const CifEntry &e1, const CifEntry &e2)
{
	PROC
	RETURN Xmm::toString(e1.instrument())+':'+Ccf::toString(e1.calDataType()) <
		   Xmm::toString(e2.instrument())+':'+Ccf::toString(e2.calDataType());
} // operator<


//
//	open
//
DataSet *CifEntry::open(DataSetServer::MemoryModel memModel, bool logAccess)
const
{
	PROC

	if (!parent())
		errstr<<"CifEntries can only be opened after they have been "
				"registered with a CIF"
			  <<Sas::fatal("CifEntryNotRegisteredWithCif");

	string ccfPath = parent()->ccfSearchPath();
	if (ccfPath.empty()) {
		char *ccfPathPtr = getenv(Ccf::SAS_CCFPATH);
		if (ccfPathPtr)
			ccfPath = string(ccfPathPtr);
		if (!ccfPath.empty())
			ccfPath += ':';

		ccfPath += parent()->cifDirName();
		ccfPath += ":.";
	}
	FileLocator fileLocator(ccfPath);

	_fileName = fileName();
	if (!_fileName.empty() && _fileName[0] != '/' && _fileName[0] != '.')
		//	just file name - look along SAS_CCFPATH for location
		_fileName = fileLocator.locate(_fileName);
	DataSet *set = dataSetServer->open(_fileName, Dal::Read, memModel);

	if (logAccess)
		CcfAccessLog::instance().insert(*this);

	RETURN set;
} // CifEntry::open()


//	destructor
CifEntry::~CifEntry() { PROC ENDPROC }


// print CifEntry to a stream
ostream &operator<<(ostream &os, const CifEntry &e)
{
	PROC
	os<<"CifEntry{"<<Xmm::toString(e._instrument)<<", "
	  <<Ccf::toString(e._calDataType)<<", "
	  <<e._issue<<", "<<e._fileName<<", "
	  <<e._validityDate<<"}";

	RETURN os;
} // operator<<


//
//	CCF class implementation
//
static string __tstr[] = { TYPE_STR, "" };
StringIntMap<Ccf::CalDataTypes> Ccf::_calDataTypes(__tstr);


Ccf::Ccf(const string cifFileName, const CalServer *server)
	: CalChild<CalServer>(server)
{
	PROC

	string cifName = cifFileName;
	if (cifName.empty()) {
		//	no file name has been specified - inquire environment
		//	for location of CIF to use
		char *cifPtr;
		if ((cifPtr = getenv(Ccf::SAS_CCF)))
			cifName = string(cifPtr);
		else
			errstr<<"No file name for Calibration Index File specified and "
					"environment variable `"
				  <<Ccf::SAS_CCF<<"' is not set either - "
					"cannot continue."
				  <<Sas::fatal("CannotLocateCIF");
	}

	_cifDirName = "./";
	DIR *dir = opendir(cifName.c_str());
	if (dir) {
		closedir(dir);
		_cifDirName = cifName + '/';
		cifName = _cifDirName + Ccf::SAS_DEFAULT_CIFNAME;
	} else {
		ifstream f(cifName.c_str());
		if (!f)
			errstr<<"`"<<cifName<<"' is neither the name of a directory nor "
					"the name of a CIF"
				  <<Sas::fatal("NoCifSpecified");
		else {
			string::size_type pos =  cifName.rfind('/');
			if (pos != string::npos)
				_cifDirName = cifName.substr(0, pos);
		}
	}
	_cifFileName = cifName;

	ENDPROC
} // Ccf::Ccf()


Ccf::CalDataTypes Ccf::toTypeId(const string &typeString)
{
	PROC
	RETURN _calDataTypes[typeString];
} // Ccf::toTypeId


const string &Ccf::toString(CalDataTypes dataTypeId)
{
	PROC
	RETURN _calDataTypes[dataTypeId];
} // Ccf::toString


const string &Ccf::toString(unsigned dataTypeId)
{
	PROC
	RETURN toString(static_cast<CalDataTypes>(dataTypeId));
} // Ccf::toString
	

Ccf::~Ccf()
{
	PROC
	ENDPROC
}


void Ccf::print(ostream &os) const
{
	PROC
	os<<"Ccf{"<<_cifFileName<<"}";
	ENDPROC
} // Ccf::print()


// print CCF to a stream
ostream &operator<<(ostream &os, const Ccf *ccf)
{
	PROC
	ccf->print(os);
	RETURN os;
} // operator<<


//	CcfAccessLog
CcfAccessLog CcfAccessLog::_instance;


ostream &operator<<(ostream &os, const CcfAccessLog &log)
{
	PROC

	for (set<CifEntry>::const_iterator i = log.begin(); i != log.end(); ++i)
		os<<*i<<endl;

	RETURN os;
} // operator<<


#define CODE																\
	_algoid = 0;															\
	if (_block->hasAttribute("ALGOID"))										\
		_algoid = _block->attribute("ALGOID")->asInt();						\
																			\
	errstr<<"Accessing table `"<<_block->qualifiedName()<<"' with ALGOID="	\
		  <<_algoid<<Sas::message(Msg::Lib, Msg::Verbose);					\
																			\
	if (_algoid > acceptAlgoIdsLessEqThan)									\
		errstr<<"The table/array `"<<_block->qualifiedName()				\
			  <<"' specifies ALGOID="<<_algoid<<" but this version of the"	\
			  <<" CAL merely supports 0<=ALGOID<="<<acceptAlgoIdsLessEqThan	\
			  <<" - you may want to upgrade to a more recent version of the"\
				" CAL that knows about the reported ALGOID."				\
			  <<Sas::fatal("UnsupportedAlgoId");


CcfConstituentBlock<Table>::CcfConstituentBlock(const DataSet *set,
											const string &name,
											unsigned acceptAlgoIdsLessEqThan)
	: _algoid(0), _block(0)
{
	PROC

	if (set->hasBlock(name)) {
		_block = set->table(name);
		CODE
	}

	ENDPROC
} // CcfConstituentBlock<Table>::CcfConstituentBlock()


CcfConstituentBlock<Table>::CcfConstituentBlock(const DataSet *set,
											unsigned nr,
											unsigned acceptAlgoIdsLessEqThan)
{
	PROC

	if (nr < set->blocks()) {
		_block = set->table(nr);
		CODE
	}

	ENDPROC
} // CcfConstituentBlock<Table>::CcfConstituentBlock()


CcfConstituentBlock<Array>::CcfConstituentBlock(const DataSet *set,
											const string &name,
											unsigned acceptAlgoIdsLessEqThan)
{
	PROC

	if (set->hasBlock(name)) {
		_block = set->array(name);
		CODE
	}

	ENDPROC
} // CcfConstituentBlock<Array>::CcfConstituentBlock()


CcfConstituentBlock<Array>::CcfConstituentBlock(const DataSet *set,
											unsigned nr,
											unsigned acceptAlgoIdsLessEqThan)
{
	PROC

	if (nr < set->blocks()) {
		_block = set->array(nr);
		CODE
	}
	ENDPROC
} // CcfConstituentBlock<Array>::CcfConstituentBlock()

#undef CODE

