//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EnergyCorrector.cc --- implementation of EnergyCorrector class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-05-25: 1.0.0: - first version (extracted from
//								 CalAtomsCommon.cc)
//		UL, 1999-06-08: 1.1.0: - minor bug fixes
//		UL, 1999-06-15: 1.2.0: - added implementation of EpnEnergyCorrector
//							     class methods
//      UL, 1999-08-13: 1.3.0: - added piTokeV()
//		UL, 1999-09-03: 1.4.0: - changed ser->grade in
//								   RgsEnergyCorrector::offsetCorrect and
//								 updated implementation
//		UL, 1999-11-24: 1.5.0: - brought in line with calpnalgo: pass
//								 additional parameter vector to gain init
//								 routine
//		UL, 2000-01-05: 1.6.0: - removed check that cross-tests pattern
//								 IDs in energy combinator coefficient
//								 columns against those in pattern lib
//								 (fixes SPR SPR 1052 : CAL_gainCorrect does
//								  not work in Timing mode for EMOS)
//		UL, 2000-02-01: 1.7.0: - made EPN-specific gainCorrect() in generic
//								 corrector use generic gainCorrect()
//		UL, 2000-02-29: 1.8.0: - brought RGS offset correction in line with
//								 [2]
//		UL, 2000-03-01: 1.9.0: - RgsEnergyCorrector::offsetCorrect: bug
//								 fix + more sensible handling of illegal
//								 event shapes
//		UL, 2000-04-04: 1.10.0:- E_RAND_MIN/E_RAND_MAX -> RAND_MIN/RAND_MAX
//		UL, 2000-04-13: 1.11.0:- check error code from PN gain correction
//		UL, 2000-05-10: 1.12.0:- overhaul in the process of implementing new
//								 MOS gain correction
//		UL, 2000-05-22: 1.13.0:- fixed problem with generic/EMOS-specific
//								 gain correction routine
//							   - fall back upon CCF if given RGS offset
//								 override table is empty
//		UL, 2000-07-31: 1.14.0:- EPN gain correction takes now real-valued
//								 input amplitude
//		UL, 2000-11-08: 1.15.0:- EmosEnergyCorrector::gainCorrect() in modes
//								 FastUncompressed/FastCompressed for CCD 1
//								 takes additional multiplicative factor
//								 (FAST_G in CCF::AduConv) into accoun
//		UL, 2000-11-14: 1.15.1:- bug fix in
//									RgsEnergyCorrector::RgsEnergyCorrector
//								 [fixes SPR 1834: Task aborts with Bus Error
//		UL, 2000-11-21: 1.15.2:- PN gain correction: better message in case
//								 of calpnalgo failure
//		UL, 2001-04-03: 1.16.0:- several changes related to ALGOID=1
//								 RGS offset correction; offset values
//								 are now time dependent:
//									offset = offsetConst + t*offsetSlope
//								 t = obsStartDate - refTime
//								 refTime comes from ADUCONV
//		UL, 2001-10-30: 1.17.0:- added implementation of new RGS HTR offset
//								 correction
//		UL, 2001-11-21: 1.18.0:- fixed problem in HTR-specific offset
//								 correction: was testing against wrong
//								 camera mode
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/EnergyCorrector.cc,v 1.33 2001/12/18 17:10:03 ulammers Exp $
//

#include <iomanip>
#include <algorithm>
#include <cmath>

#include "errstr.h"
#include "common.h"
#include "EnergyCorrector.h"
#include "CalServer.h"
#include "CalOalUtility.h"
#include "BadPixelDataServer.h"	//	INVALID_BAD_PIXEL_TABLE_CODE
#include "CalPnAlgo.h"

using namespace std;

//
//	---( EnergyCorrector )---------------------------------------------------
//
CalAtomCollection<EnergyCorrector>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	EnergyCorrector *null = 0;
	_collection = new vector<EnergyCorrector *>(_nEntries, null);

	//
	//	read the principal data from the CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::ADUConv);
	DataSet *dataSet = e->open();
	CcfConstituentTable aduConvTab(dataSet,
								   e->queryBlockName("OFFSET_GAIN"),
								   1);	//	we know about ALGOID: 0+1
															
	//
	//	instantiate high-accuracy/instrument-specific gain corrector
	//
	switch (instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2: {
			CcfConstituentTable ccftab(dataSet, "ENERGY_COMBINE");
			const Table *eCombine = ccftab.block();
			for (unsigned n=0; n<_nEntries; ++n)
				(*_collection)[n] =
					new EmosEnergyCorrector(n+1, aduConvTab, eCombine, this);
			}
			break;
		case Xmm::EPN: {
			CcfConstituentTable ccftab1(dataSet, "GAIN-HIGH"),
								ccftab2(dataSet, "REEMISSION");
			const Table *gainHigh = ccftab1.block(),
					    *reEmission = ccftab2.block();
			for (unsigned n=0; n<_nEntries; ++n)
				(*_collection)[n] =
					new EpnEnergyCorrector(n+1, aduConvTab, gainHigh,
										   reEmission, this);
			}
			break;
		case Xmm::RGS1:
		case Xmm::RGS2: {
			CifEntry *e = ccf->entry(instrument, Ccf::DarkFrame);
			DataSet *darkSet = e->open(),
					*diagSet = 0;
			CcfConstituentTable ccftab(darkSet, "DARKFRAME");
			const Table *darkFrame = ccftab.block(),
						*offsetOverride = 0;
			if (ccf->hasEntry(instrument, Ccf::DiagnosticOffset)) {
				CifEntry *e = ccf->entry(instrument, Ccf::DiagnosticOffset);
				diagSet = e->open();
				const Table *offsetTab = diagSet->table("RGSOFFSETS");
				if (offsetTab->rows())
					offsetOverride = offsetTab;
				else
					errstr<<"Specified offset override table `"
						  <<offsetTab->qualifiedName()<<"' is empty (0 rows) "
							"- offset correction will fall back upon numbers "
							"in the CCF"
						  <<Sas::warning("EmptyOffsetOverrideTable");
			}
			for (unsigned n=0; n<_nEntries; ++n) {
				(*_collection)[n] =
					new RgsEnergyCorrector(n+1, aduConvTab, darkFrame,
										   offsetOverride, this);
			}
			if (diagSet)
				dataSetServer->close(diagSet);
			dataSetServer->close(darkSet);
			}
			break;
		default:
			break;
	}

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<EnergyCorrector>::CalAtomCollection()


EnergyCorrector *CalAtomCollection<EnergyCorrector>::select(
					Ccd ccd,
					Node node,
					unsigned) const
{
	PROC

	_ccd = ccd;
	_node = node;
	unsigned n = ccd - 1;

	RETURN (*_collection)[n];
} // CalAtomCollection<EnergyCorrector>::select


EnergyCorrector::EnergyCorrector(Ccd ccd,
								 const CcfConstituentTable &aduConvTab,
								 CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent)
{
	PROC

	const Table *aduConv = aduConvTab.block();
	_bptCode = CalOalUtility::optionalAttribute(aduConv, "BPT_CODE",
						 BadPixelDataServer::INVALID_BAD_PIXEL_TABLE_CODE);

	//
    //  read PI->eV conversion values
    //
	_eVperPi = aduConv->parent()->attribute("KEV_P_PI")->asReal()*1000.;
	_eV0	 = aduConv->parent()->attribute("KEV0")->asReal()*1000.;

	ColumnData *cdccdId	= aduConv->column("CCD_ID")->data(),
			   *cdgain	= aduConv->column("GAIN")->data(),
			   *cdoffset0 = 0, *cdoffset1 = 0;

	_offsetRefTime = Xmm::missionReferenceTime;
	if (aduConvTab.algoid() == 0)
		cdoffset0 = aduConv->column("OFFSET")->data();
	else {
		cdoffset0 = aduConv->column("OFFSET_CONST")->data();
		cdoffset1 = aduConv->column("OFFSET_SLOPE")->data();
		_offsetRefTime = STime(aduConv->attribute("REFTIME")->asString());
	}

	int8	*ccdId		= cdccdId->int8Data();
	real32	*gain		= cdgain->real32Data(),
			*offset0	= cdoffset0->real32Data(),
			*offset1	= cdoffset1 ? cdoffset1->real32Data() : 0;

	for (unsigned n=0; n<aduConv->rows(); ++n) {
		unsigned nn = cdgain->elements();
		if (ccdId[n] <= 0 || ccdId[n] > parent->ccds() ||
			nn > parent->nodes())
			errstr<<"Illegal chip and/or node number in row #"
				  <<n<<" of table `"<<aduConv->qualifiedName()<<'\''
			  <<Sas::error("IllegalChipNodeId");

		if (ccdId[n] == ccd) {
			for (unsigned j=0; j<nn; ++j) {
				unsigned p = cdgain->position(n, j);
				//
				//	check for illegal gain values
				//
				if (gain[p] <= 0.)
					errstr<<"gain for CCD #"<<ccdId[n]<<"/node #"<<j
						  <<" is less than or equal to zero"
						  <<Sas::error("IllegalGainValue");

				_offsetConst.push_back(offset0[p]);
				_offsetSlope.push_back(offset1 ? offset1[p] : 0.);
				_gain.push_back(gain[p]);
			}
			break;
		}
	}

	//
	//	check that we have read enough data
	//
	if (_gain.size() != parent->nodes() ||
		_offsetConst.size() != parent->nodes())
		errstr<<"Expected to read "<<parent->nodes()<<" values for the "
				"offset/gain from table `"
			  <<aduConv->qualifiedName()<<"' but actually found: "
			  <<_offsetConst.size()<<"/"<<_gain.size()
			  <<Sas::error("wrongNumberOfElements");

	//	tidy up
	delete cdccdId;
	delete cdgain;
	delete cdoffset0;
	delete cdoffset1;

	ENDPROC
} // EnergyCorrector::EnergyCorrector()


void EnergyCorrector::dump(ostream &os)
{
	PROC

	os<<"EnergyCorrector{\n"
	  <<"offset: ";
	for (unsigned n=0; n<_offsetConst.size(); ++n)
		os<<_offsetConst[n]<<'/'<<_offsetSlope[n]<<' ';
	os<<"gain  : ";
	for (unsigned n=0; n<_gain.size(); ++n)
		os<<_gain[n]<<" ";

	os<<"}";

	ENDPROC
} // EnergyCorrector::dump()


float EnergyCorrector::offsetValue(Node node) const
{
	PROC

	unsigned n = UNT(node);
	if (n>=_offsetConst.size())
		errstr<<"Unknown readout node "<<INT(node)
			  <<Sas::error("UnknownReadoutNode");

	double deltaT = calServer.state()->observationStartTime()->get() -
						_offsetRefTime;

	RETURN _offsetConst[n] + deltaT * _offsetSlope[n];
} // EnergyCorrector::offsetValue()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &,
											  const CalInt8Vector &nodes,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	PROC

	errstr<<"Detailed time dependence of offset correction currently not "
			"implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	unsigned nNodes = _offsetSlope.size();
	float offsetVec[nNodes];
	for (unsigned n=0; n<nNodes; offsetVec[n++] = offsetValue(Xmm::Node(n)));

	float offset;
	for (unsigned n=0; n<eIn.size(); ++n) {
		if (UNT(nodes[n]) < nNodes)
			offset = offsetVec[nodes[n]];
		else {
			errstr<<"No offset value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			offset = 0.f;
		}

		eOut[n] = eIn[n] - offset;
	}

	RETURN eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &time,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	PROC

	unsigned N=eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex);

	offsetCorrect(time, nodes, eIn, eOut);

	RETURN eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::offsetCorrect(const CalReal64Vector &,
											    const CalInt16Vector &,
												const CalInt8Vector &,
												const CalInt16Vector &,
											    CalReal32Vector &eOut) const
{
	PROC

	errstr<<"Generic energy corrector cannot perform RGS HTR-specific "
			"offset correction"<<Sas::error("UnsupportedOperation");

	RETURN eOut;
} // EnergyCorrector::offsetCorrect()


CalReal32Vector &EnergyCorrector::gainCorrect(const CalReal64Vector &time,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	PROC

	unsigned N = eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex);

	gainCorrect(time, nodes, eIn, eOut);
				
	RETURN eOut;
} // EnergyCorrector::gainCorrect()



CalReal32Vector &EnergyCorrector::gainCorrect(const CalReal64Vector &,
											  const CalInt8Vector &nodes,
											  const CalReal32Vector &eIn,
											  CalReal32Vector &eOut) const
{
	PROC

	errstr<<"Time dependence of gain correction not implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	char nNodes = _gain.size();
	float gain;

	for (unsigned n=0; n<eIn.size(); ++n) {
		if (nodes[n] < nNodes)
			gain = _gain[nodes[n]];
		else {
			errstr<<"No gain value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			gain = 1.f;
		}

		eOut[n] = eIn[n];
		if (gain != 0.)
			eOut[n] /= gain;
		else
			errstr<<"Gain is 0 for node "<<INT(nodes[n])<<" on event number "
				  <<n<<" - energy unmodified"
				  <<Sas::warning("ZeroGain");
	}

	RETURN eOut;
} // EnergyCorrector::gainCorrect()


CalReal32Vector &EnergyCorrector::piToeV(const CalReal32Vector &eIn,
										 CalReal32Vector &eOut) const
{
	PROC

	for (unsigned n=0; n<eIn.size(); ++n)
		eOut[n] = _eV0 + eIn[n] * _eVperPi;

	RETURN eOut;
} // EnergyCorrector::piToeV()


//
//	--------------------------( EMOS )--------------------------------------
//
vector<float> EmosEnergyCorrector::_e1Coef,
			  EmosEnergyCorrector::_e2Coef,
			  EmosEnergyCorrector::_e3Coef,
			  EmosEnergyCorrector::_e4Coef;
bool		  EmosEnergyCorrector::_coefRead = false;


EmosEnergyCorrector::EmosEnergyCorrector(Ccd ccd,
										 const CcfConstituentTable &aduConvTab,
										 const Table *eCombine,
								 CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	PROC

	const Table *aduConv = aduConvTab.block();

	//
	//	read offset/gain coefficients
	//
	ColumnData *cdCcdId		= aduConv->column("CCD_ID")->data(),
			   *cdOffset0	= aduConv->column("OFFSET")->data(),
			   *cdOffset1	= aduConv->column("OFFSET1")->data(),
			   *cdOffset2	= aduConv->column("OFFSET2")->data(),
			   *cdOffset3	= aduConv->column("OFFSET3")->data(),
			   *cdGain0		= aduConv->column("GAIN")->data(),
			   *cdGain1		= aduConv->column("GAIN1")->data(),
			   *cdGain2		= aduConv->column("GAIN2")->data(),
			   *cdGain3		= aduConv->column("GAIN3")->data(),
			   *cdPoly0		= aduConv->column("POLY_TERM")->data(),
			   *cdPoly1		= aduConv->column("POLY_TERM1")->data(),
			   *cdPoly2		= aduConv->column("POLY_TERM2")->data(),
			   *cdPoly3		= aduConv->column("POLY_TERM3")->data();

	int8	*ccdId = cdCcdId->int8Data();
	float	*offset0 = cdOffset0->real32Data(),
			*offset1 = cdOffset1->real32Data(),
			*offset2 = cdOffset2->real32Data(),
			*offset3 = cdOffset3->real32Data(),
			*gain0 = cdGain0->real32Data(),
			*gain1 = cdGain1->real32Data(),
			*gain2 = cdGain2->real32Data(),
			*gain3 = cdGain3->real32Data(),
			*poly0 = cdPoly0->real32Data(),
			*poly1 = cdPoly1->real32Data(),
			*poly2 = cdPoly2->real32Data(),
			*poly3 = cdPoly3->real32Data();

	unsigned nodes = parent->nodes();
	if (cdOffset0->elements() != nodes || cdOffset1->elements() != nodes ||
		cdOffset2->elements() != nodes || cdOffset3->elements() != nodes ||
		cdGain0->elements() != nodes || cdGain1->elements() != nodes ||
		cdGain2->elements() != nodes || cdGain3->elements() != nodes ||
		cdPoly0->elements() != nodes || cdPoly1->elements() != nodes ||
		cdPoly2->elements() != nodes || cdPoly3->elements() != nodes)
		errstr<<"Size of vector column in `"<<aduConv->qualifiedName()
			  <<"' does not coincide with number of nodes ("<<nodes<<")"
			  <<Sas::error("CcfDataIntegrityError");

	unsigned n;
	for (n=0; n<aduConv->rows(); ++n)
		if (ccdId[n] == ccd) {
			for (unsigned j=0; j<nodes; ++j) {
				unsigned p = cdOffset0->position(n, j);

				_offsetGain[0][0][j] = offset0[p];
				_offsetGain[0][1][j] = offset1[p];
				_offsetGain[0][2][j] = offset2[p];
				_offsetGain[0][3][j] = offset3[p];

				_offsetGain[1][0][j] = gain0[p];
				_offsetGain[1][1][j] = gain1[p];
				_offsetGain[1][2][j] = gain2[p];
				_offsetGain[1][3][j] = gain3[p];

				_offsetGain[2][0][j] = poly0[p];
				_offsetGain[2][1][j] = poly1[p];
				_offsetGain[2][2][j] = poly2[p];
				_offsetGain[2][3][j] = poly3[p];
			}
			break;
		}

	if (n>aduConv->rows())
		errstr<<"Could not find entry for CCD #"<<ccd<<"in `"
			  <<aduConv->qualifiedName()<<"'"
			  <<Sas::error("CcfDataIntegrityError");

	delete cdCcdId;
	delete cdOffset0;
	delete cdOffset1;
	delete cdOffset2;
	delete cdOffset3;
	delete cdGain0;
	delete cdGain1;
	delete cdGain2;
	delete cdGain3;
	delete cdPoly0;
	delete cdPoly1;
	delete cdPoly2;
	delete cdPoly3;

	//
	//	read FAST_G attribute - see email D. Lumb 2000-11-08
	//
	_fastG = 1.;
	if (aduConv->hasAttribute("FAST_G"))
		_fastG = aduConv->attribute("FAST_G")->asReal();	

	//
	//	read energy combination coefficients
	//
	if (!_coefRead) {
		_coefRead = true;

		ColumnData *cdPattern = eCombine->column("PATTERN_ID")->data(),
				   *cdCoef = eCombine->column("ENERGY_PARAM")->data();
		int8 *pattern = cdPattern->int8Data();
		real32 *coef = cdCoef->real32Data();

		if (cdCoef->elements() != 4)
			errstr<<"Expecting 4 elements in column `"
				  <<cdPattern->parent()->qualifiedName()<<""
				  <<Sas::error("CcfDataIntegrityError");

		unsigned N = eCombine->rows();
		_e1Coef.resize(N);
		_e2Coef.resize(N);
		_e3Coef.resize(N);
		_e4Coef.resize(N);
		for (unsigned n=0; n<N; ++n) {
			unsigned patt = pattern[n];
			if (patt >= _e1Coef.size())
				errstr<<"Unexpected pattern ID "<<patt<<" in column `"
					  <<cdPattern->parent()->qualifiedName()<<"'"
					  <<Sas::error("UnknownPattern");
			_e1Coef[patt] = coef[cdCoef->position(n, 0)];
			_e2Coef[patt] = coef[cdCoef->position(n, 1)];
			_e3Coef[patt] = coef[cdCoef->position(n, 2)];
			_e4Coef[patt] = coef[cdCoef->position(n, 3)];
		}

		delete cdPattern;
		delete cdCoef;
	}

	ENDPROC
} // EmosEnergyCorrector::EmosEnergyCorrector()


//
//	see email D. Lumb, dated 2000-04-13
//
CalReal32Vector &EmosEnergyCorrector::gainCorrect(const CalReal64Vector &,
												  const CalReal32Vector &eIn,
												  const CalInt8Vector &pattern,
												  CalReal32Vector &eOut) const
{
	PROC

	//
	//	fudge: for CCD1 in modes FastUncompressed/FastCompressed additional
	//	multiplicative factor has to be taken into account
	//
	float effFastG = 1.;
	Mode mode = calServer.state()->mode()->get();
	if (calServer.state()->ccd()->get() == 1 &&
		(mode == Xmm::FAST_UNCOMPRESSED || mode == Xmm::FAST_COMPRESSED))
		effFastG = _fastG;

	int nodeIndx = parent()->node();
	unsigned N = eIn.size();
	for (unsigned n=0; n<N; ++n) {
		unsigned patternIndx = (pattern[n]+3)/4;
		if (patternIndx > 3)
			patternIndx = 3;
		//
		//	pattern ID		patternIndx
		//	0				0
		//	1,2,3,4			1
		//	5,6,7,8			2
		//	9,10,11,12		3
		//	13,...			3	- treat all pattern >12 as pattern [9, 12]
		//

		eOut[n] = _offsetGain[0][patternIndx][nodeIndx] +
				  _offsetGain[1][patternIndx][nodeIndx] * eIn[n] +
				  _offsetGain[2][patternIndx][nodeIndx] * eIn[n] * eIn[n];
		eOut[n] *= effFastG;
	}

	RETURN eOut;	
} // EmosEnergyCorrector::gainCorrect()


//
//	--------------------------( EPN )--------------------------------------
//
bool EpnEnergyCorrector::_gainParamsSet = false;
EpnEnergyCorrector::EpnGainParameters EpnEnergyCorrector::_gainParams;
vector<float> EpnEnergyCorrector::_gainParamsSuppl;
double	EpnEnergyCorrector::_Acoef[2],
		EpnEnergyCorrector::_Bcoef[2],
		EpnEnergyCorrector::_adu_eV,
		EpnEnergyCorrector::_rawy0;

EpnEnergyCorrector::EpnEnergyCorrector(Ccd ccd,
									   const CcfConstituentTable &aduConvTab,
									   const Table *gainHigh,
									   const Table *reEmission,
							   CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	PROC

	if (!_gainParamsSet) {
		_gainParamsSet = true;

		if (gainHigh->rows() != N_PARAMETERS)
			errstr<<"Expected to find "<<N_PARAMETERS<<" gain parameters "
					"in PN high-accuracy gain table `"
				  <<gainHigh->qualifiedName()<<"' but there are only "
				  <<gainHigh->rows()
				  <<Sas::error("CcfDataIntegrityError");

		const Column *gainCol = gainHigh->column("GAIN_PAR");
		ColumnData *gainParCd = gainCol->data();
		unsigned nElements = gainParCd->elements();
		if (nElements != xmmPayLoad.epn.nNodes() * xmmPayLoad.epn.nCcds())
			errstr<<"Expected to find "<<xmmPayLoad.epn.nNodes() *
										 xmmPayLoad.epn.nCcds()
				  <<" elements in column "<<gainCol->qualifiedName()
				  <<" but there are "<<nElements
				  <<Sas::error("CcfDataIntegrityError");

		real32 *gainPar = gainParCd->real32Data();
	
		//
		//	allocate memory
		//
		for (unsigned n=0; n<N_PARAMETERS; ++n)
			_gainParams.push_back(new float [nElements]);

		for (unsigned n=0; n<N_PARAMETERS; ++n)
			for (unsigned pos=0; pos<nElements; ++pos)
				_gainParams[n][pos] = gainPar[gainParCd->position(n, pos)];

		delete gainParCd;

		//
		//	read supplental gain parameters
		//
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR1")->asReal());
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR2")->asReal());
		_gainParamsSuppl.push_back(gainHigh->attribute("FINEPAR3")->asReal());

		//
		//	read re-emission threshold parameters
		//
		_adu_eV = reEmission->attribute("ADU_P_EV")->asReal();
		_rawy0 = reEmission->attribute("RAWY0")->asReal();
		ColumnData *cdA = reEmission->column("A")->data(),
				   *cdB = reEmission->column("B")->data();
		real32 *A = cdA->real32Data(),
			   *B = cdB->real32Data();
		if (cdA->elements() != 2 || cdB->elements() != 2)
			errstr<<"Expected vector column A/B to be of size 2 in `"
				  <<reEmission->qualifiedName()<<"'"
				  <<Sas::error("CcfDataIntegrityError");
		_Acoef[0] = A[0]; _Acoef[1] = A[1];
		_Bcoef[0] = B[0]; _Bcoef[1] = B[1];
		delete cdA;
		delete cdB;
	}

	ENDPROC
} // EpnEnergyCorrector::EpnEnergyCorrector()


CalReal32Vector &EpnEnergyCorrector::gainCorrect(const CalReal64Vector &time,
											   const CalInt16Vector &rawX,
											   const CalInt16Vector &rawY,
											   const CalReal32Vector &eIn,
											   CalReal32Vector &eOut) const
{
	PROC

	if (calServer.state()->accuracyLevel()->get() == LOW) {
		//
		//	low-accurcy gain correction
		//
		EnergyCorrector ec = *this;
		unsigned N = rawX.size();
		CalInt8Vector nodeVec(0, N);
		for (unsigned n=0; n<N; ++n)
			nodeVec[n] = rawX[n]-1;
		ec.gainCorrect(time, nodeVec, eIn, eOut);
		ALTRETURNV(eOut);
	}

	static bool first = true;

	if (first) {
		first = false;
		//
		//	first time called - need to initialize the PN gain corrector
		//	algorithm
		//
		unsigned N = xmmPayLoad.epn.nNodes() * xmmPayLoad.epn.nCcds();
		float *pp0 = _gainParams[0],
			  *pp1 = _gainParams[1],
			  *pp2 = _gainParams[2],
			  *pp3 = _gainParams[3],
			  *aa = _gainParams[6],
			  *bb = _gainParams[8];
		int M = _gainParamsSuppl.size();
		float suppl[M];
		for (int m=0; m<M; ++m)
			suppl[m] = _gainParamsSuppl[m];

		CALPNALGO_initGain(N, pp0, pp1, pp2, pp3, aa, bb, M, suppl);
	}

	bool mediumAccuracy = calServer.state()->accuracyLevel()->get() == MEDIUM;
	unsigned N = eIn.size();
	Ccd	ccd = parent()->ccd();
	unsigned quadId = EPNcam::toQuadrantId(ccd),
			 ccdId = EPNcam::toCcdId(ccd);
	for (unsigned n=0; n<N; ++n) {
		float	ampr = eIn[n];
		int 	x = INT(rawX[n]) - 1,
				y = INT(rawY[n]) - 1,
				ierr = 0;
		float gainCorrectedAmp, gainCtiCorrectedAmp;

		CALPNALGO_gainCorrect(quadId, ccdId, x, y, ampr, gainCorrectedAmp,
							  gainCtiCorrectedAmp, ierr);

		if (!ierr)
			eOut[n] = mediumAccuracy ? gainCtiCorrectedAmp : gainCorrectedAmp;
		else {
			errstr<<"Gain correction algorithm for event #"<<n
				  <<" (CCD/rawX = "<<ccd<<"["<<quadId<<"/"<<ccdId<<"]/"
				  <<rawX[n]<<") failed - error code: "<<ierr
				  <<" no correction applied - reason for failure: ";
			switch (ierr) {
				case 3: errstr<<"negative amplitude"; break;
				default: errstr<<"unknown"; break;
			}
			errstr<<Sas::warning("gainCorrectionFailed");
			eOut[n] = eIn[n];
		}
	}

	RETURN eOut;
} // EpnEnergyCorrector::gainCorrect()


double EpnEnergyCorrector::reEmissionThreshold(double Eadu, int16 rawY) const
{
	PROC

	double EkeV = Eadu * 1.e-3 * _adu_eV;

	RETURN (_Acoef[0] + _Acoef[1] * EkeV) +
		   (_Bcoef[0] + _Bcoef[1] * EkeV) * (DBL(rawY)-_rawy0);
} // EpnEnergyCorrector::reEmissionThreshold()


#if 0
//	old obsolete re-emission thresholds
inline static double a(double T)
{
	double T2 = T * T;
	return 8207.9 - 203.87 * T + 1.8854 * T2 - 7.6916e-3 * T2*T +
		   1.1693e-5 * T2*T2;
}

inline static double A(double E, double T)
{
	double E2 = E*E;
	return a(T) + 6.768 * E - 2.496 * E2 + .381 * E2*E - .0177 * E2*E2;
}

inline static double b(double T)
{
	return -.0488 + 3.69e-4 * T;
}

inline static double B(double E, double T)
{
	return .045 + b(T) * E;
}

	double	E = *energy * 1.e-3 * 5.289;	//	E must be in keV
	double	row = DBL(*rawY-1);

	//
	//	get temperature
	//
	double T = calServer.state()->temperature()->get();

	RETURN A(E, T) + B(E, T) * (row - 100.);
#endif


EpnEnergyCorrector::~EpnEnergyCorrector()
{
	PROC

	//	free up memory
	for (unsigned n=0; n<_gainParams.size(); ++n) {
		delete _gainParams[n]; _gainParams[n] = 0;
	}
	_gainParams.resize(0);
	_gainParamsSuppl.resize(0);
	_gainParamsSet = false;

	ENDPROC
} // EpnEnergyCorrector::~EpnEnergyCorrector()


//
//	---------------------( RGS )--------------------------------------------
//
RgsEnergyCorrector::RgsEnergyCorrector(
			Ccd ccd,
			const CcfConstituentTable &aduConvTab,
			const Table *darkFrameTable,
			const Table *offsetOverrideTable,
			CalAtomCollection<EnergyCorrector> *parent)
	: CalChild<CalAtomCollection<EnergyCorrector> >(parent),
	  EnergyCorrector(ccd, aduConvTab, parent)
{
	PROC

	const Table *aduConv = aduConvTab.block();

	//
	//	read rejection/acceptance thresholds and charge pixel
	//	ratios
	//
	_qSplitR[0] = 1.f;	//	by definition
	_qSplitR[1] = aduConv->attribute("QSPLITR2")->asReal();
	_qSplitR[2] = aduConv->attribute("QSPLITR3")->asReal();
	_qSplitR[3] = aduConv->attribute("QSPLITR4")->asReal();

	ColumnData *cdRejThresh	= aduConv->column("REJTHRESH")->data(),
			   *cdAccThresh = aduConv->column("ACCTHRESH")->data();
	real32	*rejThresh = cdRejThresh->real32Data(),
			*accThresh = cdAccThresh->real32Data();

	unsigned n = 0, N = aduConv->rows();
	for (unsigned n=0; n<N; ++n) {
		if (n+1 == ccd) {
			unsigned nn = cdRejThresh->elements();
			for (unsigned j=0; j<nn; ++j) {
				unsigned p = cdRejThresh->position(n, j);
				//
				//	check for negative threshold values
				//
				if (rejThresh[p] < 0. || accThresh[p] < 0.)
					errstr<<"rejection/acceptance threshold for CCD #"
						  <<ccd<<"/node #"<<j
						  <<" is negative"
						  <<Sas::error("CcfDataIntegrityError");
				_rejectionThresholds.push_back(rejThresh[p]);
				_acceptanceThresholds.push_back(accThresh[p]);
			}
			break;
		}
	}
	if (n>=N)
		errstr<<"Did not find rejection/acceptance tresholds for RGS CCD #"
			  <<ccd<<" in `"<<aduConv->qualifiedName()<<'\''
			  <<Sas::error("CcfDataIntegrityError");

	//
	//	try to read RAND_MIN/RAND_MAX from attributes
	//
	_eRandMin = 0.;
	_eRandMax = 1.;	//	default values
	string rmin("RAND_MIN"),
		   rmax("RAND_MAX");
	if (aduConv->hasAttribute(rmin) &&
		aduConv->hasAttribute(rmax)) {
		_eRandMin = aduConv->attribute(rmin)->asReal();
		_eRandMax = aduConv->attribute(rmax)->asReal();
	} else
		if (aduConv->parent()->hasAttribute(rmin) &&
			aduConv->parent()->hasAttribute(rmax)) {
			_eRandMin = aduConv->parent()->attribute(rmin)->asReal();
			_eRandMax = aduConv->parent()->attribute(rmax)->asReal();
		} else
			errstr<<"Did not find "<<rmin<<'/'<<rmax<<" attributes in table "
				  <<aduConv->qualifiedName()<<"' or associated set - "
					"randomization will use default value range ["<<
					_eRandMin<<", "<<_eRandMax<<"]"
				  <<Sas::warning("AttributesNotFound");

	//
	//	read dark frame parameters
	//
	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	_a0.resize(nNodes);
	_a1.resize(nNodes);
	_a2.resize(nNodes);
	_a3.resize(nNodes);
	_b0.resize(nNodes);
	_b1.resize(nNodes);
	_b2.resize(nNodes);
	_b3.resize(nNodes);
	ColumnData *cda0 = darkFrameTable->column("A_0")->data(),
			   *cda1 = darkFrameTable->column("A_1")->data(),
			   *cda2 = darkFrameTable->column("A_2")->data(),
			   *cda3 = darkFrameTable->column("A_3")->data(),
			   *cdb0 = darkFrameTable->column("B_0")->data(),
			   *cdb1 = darkFrameTable->column("B_1")->data(),
			   *cdb2 = darkFrameTable->column("B_2")->data(),
			   *cdb3 = darkFrameTable->column("B_3")->data(),
			   *cdccd = darkFrameTable->column("CCD_ID")->data(),
			   *cdnode= darkFrameTable->column("NODE_ID")->data();
	real32	*a0 = cda0->real32Data(),
			*a1 = cda1->real32Data(),
			*a2 = cda2->real32Data(),
			*a3 = cda3->real32Data(),
			*b0 = cdb0->real32Data(),
			*b1 = cdb1->real32Data(),
			*b2 = cdb2->real32Data(),
			*b3 = cdb3->real32Data();
	int8	*ccdid = cdccd->int8Data(),
			*nodeid = cdnode->int8Data();

	n = 0;
	N = darkFrameTable->rows();
	bool found = false;
	for (; n<N; ++n)
		if (ccdid[n] == ccd) {
			found = true;
			unsigned node = nodeid[n];
			if (node < nNodes) {
				_a0[node] = a0[n]; _a1[node] = a1[n];
				_a2[node] = a2[n]; _a3[node] = a3[n];
				_b0[node] = b0[n]; _b1[node] = b1[n];
				_b2[node] = b2[n]; _b3[node] = b3[n];
			} else
				errstr<<"Illegal node ID ("<<node<<") seen in `"
					  <<darkFrameTable->qualifiedName()
					  <<Sas::error("CcfDataIntegrityError");
		}

	if (!found)
		errstr<<"Did not find dark frame parameters for RGS CCD #"
			  <<ccd<<" in `"<<darkFrameTable->qualifiedName()<<"'"
			  <<Sas::error("CcfDataIntegrityError");

	delete cda0;
	delete cda1;
	delete cda2;
	delete cda3;
	delete cdb0;
	delete cdb1;
	delete cdb2;
	delete cdb3;
	delete cdccd;
	delete cdnode;

	//
	//	optionally read diagnostic offset data from dynamic CCF constituent
	//	created by rgsoffsetcalc
	//
	_hasDiagnosticOffsetData = offsetOverrideTable;
	if (!_hasDiagnosticOffsetData) {
		flatOffsetCurves();	//	define flat offset curves

		ALTRETURN;
	}

	ColumnData *cdtime = offsetOverrideTable->column("TIME")->data(),
			   *cdmean= offsetOverrideTable->column("MEAN")->data();
    cdccd = offsetOverrideTable->column("CCD_ID")->data(),
	cdnode= offsetOverrideTable->column("NODE_ID")->data();

	real64	*time = cdtime->real64Data();
	ccdid = cdccd->int8Data(),
	nodeid = cdnode->int8Data();
	real32	*mean = cdmean->real32Data();

	vector<double> t[nNodes],
				   offset[nNodes];
	double prevTime[nNodes];
	for(n=0; n<nNodes; prevTime[n++] = -1.);
	n = 0;
	N = offsetOverrideTable->rows();
	for (; n<N; ++n)
		if (UNT(ccdid[n]) == ccd) {
			unsigned node = nodeid[n];
			if (node < nNodes) {
				if (time[n] != prevTime[node]) {
					prevTime[node] = time[n];
					t[node].push_back(time[n]);
					offset[node].push_back(mean[n]);
				} else
					errstr<<"Time in row "<<n<<" of `"
						  <<offsetOverrideTable->qualifiedName()
						  <<"' for CCD "<<ccd<<" node "<<node<<" identical "
							"to previous time - ignoring entry"
						  <<Sas::warning("WrongTime");
			} else
				errstr<<"Illegal node ID ("<<node<<") seen in `"
					  <<darkFrameTable->qualifiedName()
					  <<Sas::error("CcfDataIntegrityError");
		}

	delete cdtime;
	delete cdmean;

	if (!t[0].size() && !t[1].size()) {
		errstr<<"Did not find entry for CCD #"<<ccd<<" in `"
			  <<offsetOverrideTable->qualifiedName()<<"' - offset correction "
				"for this CCD will use CCF-based values"
			  <<Sas::warning("NoDynamicOffsetData");
		unsigned nNodes = xmmPayLoad.rgs1.nNodes();
		for (unsigned node=0; node<nNodes; flatOffsetCurve(node++));

		ALTRETURN;
	}

	for (unsigned node=0; node<nNodes; ++node) {
		if (t[node].size()) {
			//
			//	insert duplicate first/last offset value for three
			//	early/late times times such that the offset curve has a
			//	plateau at the beginning/end
			//
			t[node].insert(t[node].begin(), 100.);
			t[node].insert(t[node].begin(), 10.);
			t[node].insert(t[node].begin(), 0.);
			for (n=0; n<3; ++n)
				offset[node].insert(offset[node].begin(), offset[node][0]);

			double missionDuration = Xmm::missionEndTime -
											Xmm::missionReferenceTime;
			t[node].push_back(missionDuration-1000.);
			t[node].push_back(missionDuration-100.);
			t[node].push_back(missionDuration);
			for (n=0; n<3; ++n)
				offset[node].push_back(offset[node][offset[node].size()-1]);
			_offsetCurves[node] =
				new Slatec::CubicHermiteInterpolant<double>(t[node],
															offset[node]);
		} else {
			//	no data for node <node>
			errstr<<"No dynamic offset data for node "<<node<<" have been "
					"found - using flat offset curve"
				  <<Sas::warning("NoDynamicOffsetData");
			flatOffsetCurve(node);
		}
	}

	ENDPROC
} // RgsEnergyCorrector::RgsEnergyCorrector()


void RgsEnergyCorrector::flatOffsetCurve(unsigned node)
{
	PROC

	if (node >= xmmPayLoad.rgs1.nNodes())
		errstr<<"Illegal node ID "<<node<<Sas::error("IllegalNodeId");

	//
	//	define flat offset curve
	//
	vector<double>	timeVec(2),
					offsetVec(2);
	timeVec[0] = 0.;
	timeVec[1] = 1000.;
	offsetVec[0] = offsetVec[1] = offsetValue(Xmm::Node(node));
	_offsetCurves[node] =
			new Slatec::CubicHermiteInterpolant<double>(timeVec, offsetVec);

	ENDPROC
} // RgsEnergyCorrector::flatOffsetCurve()


void RgsEnergyCorrector::flatOffsetCurves()
{
	PROC

	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	for (unsigned n=0; n<nNodes; flatOffsetCurve(n++));

	ENDPROC
} // RgsEnergyCorrector::flatOffsetCurves()



CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt16Vector &rawY,
												 const CalInt8Vector &shape,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	PROC

	unsigned N = eIn.size();
	CalInt8Vector nodes(0, N);
	int nodeIndex = parent()->node();
	for (unsigned n=0; n<N; nodes[n++] = nodeIndex);

	offsetCorrect(time, rawX, rawY, shape, nodes, eIn, eOut);

	RETURN eOut;
} // RgsEnergyCorrector::offsetCorrect()


CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt16Vector &rawY,
												 const CalInt8Vector &shape,
												 const CalInt8Vector &nodes,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	PROC

	randomize(eIn, eOut);
	float offset;
	char nNodes = _offsetConst.size();
	for (unsigned n=0; n<eIn.size(); ++n) {
		if (nodes[n] < nNodes)
			offset = (*_offsetCurves[nodes[n]])(time[n]);
		else {
			errstr<<"No offset value for node "<<INT(nodes[n])
				  <<" event #"<<n<<" - no correction applied"
				  <<Sas::warning("NoSuchValue");
			offset = 0.f;
		}

		eOut[n] -= shapeToGrade(shape[n]) * (offset +
					   darkCurrent(rawX[n], rawY[n], nodes[n]));
	}

	RETURN eOut;
} // RgsEnergyCorrector::offsetCorrect()


//
//	HTR mode offset correction
//
CalReal32Vector &RgsEnergyCorrector::offsetCorrect(
												 const CalReal64Vector &time,
												 const CalInt16Vector &rawX,
												 const CalInt8Vector &nodes,
												 const CalInt16Vector &eIn,
												 CalReal32Vector &eOut) const
{
	PROC

	CalState *state = calServer.state();

	//	verify we have HTR mode
	Mode mode = state->mode()->get();
	if (mode != Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD &&
		mode != Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD)
		errstr<<"Called RGS HTR-specific offset correction in mode `"
			  <<mode<<"'"<<Sas::error("ImproperMode");

	//	set OCB to 3 + construct dummy rawY/shape vectors
	//	set rawY to middle of readout window in OCB 3; shape is 0
	unsigned oldocb = state->ocb()->get();
	state->ocb()->set(3);

	CalInt16Vector rawY(0, rawX.size());
	int16 v = static_cast<int16>(calServer.instrument()->maxY()/3/2);
	rawY = v;
	CalInt8Vector shape(0, rawX.size());
	shape = 0;

	//	call spectroscopy offset correction
	CalReal32Vector &e =
					offsetCorrect(time, rawX, rawY, shape, nodes, eIn, eOut);
	state->ocb()->set(oldocb);

	RETURN e;
} // RgsEnergyCorrector::offsetCorrect()


float RgsEnergyCorrector::darkCurrent(int16 rawX, int16 rawY, unsigned node)
const
{
	PROC

	double x = DBL(rawX-1),
		   y = DBL(rawY-1),
		   dc = _a0[node]*exp(_a1[node]*x) +
				_a2[node]*x +
				.5*(_a3[node] + _b3[node]) +
				_b0[node]*exp(_b1[node]*y) +
				_b2[node]*y;	//	see [2], p. 47

	RETURN dc;
} // RgsEnergyCorrector::darkCurrent()


double RgsEnergyCorrector::rejectionThreshold(unsigned grade) const
{
	PROC
	RETURN rejectionThreshold(grade, calServer.state()->node()->get());
} // RgsEnergyCorrector::rejectionThreshold()


double RgsEnergyCorrector::rejectionThreshold(unsigned grade,
											  Node node) const
{
	PROC

	if (grade < 1 || grade > NGRADES)
		errstr<<"Passed grade value ("<<grade<<") invalid - must be in "
				"[1, "<<NGRADES<<"]"<<Sas::error("InvalidGradeValue");
	if (node < 0 || node > INT(xmmPayLoad.rgs1.nNodes()))
		errstr<<"Passed node value ("<<node<<") invalid - must be in "
				"[0, "<<xmmPayLoad.rgs1.nNodes()<<"]"
			  <<Sas::error("InvalidNode");

	RETURN _rejectionThresholds[node] / _qSplitR[grade-1];
} // RgsEnergyCorrector::rejectionThreshold()


double RgsEnergyCorrector::acceptanceThreshold(unsigned grade) const
{
	PROC
	RETURN acceptanceThreshold(grade, calServer.state()->node()->get());
} // RgsEnergyCorrector::acceptanceThreshold()


double RgsEnergyCorrector::acceptanceThreshold(unsigned grade,
											   Node node) const
{
	PROC

	if (grade < 1 || grade > NGRADES)
		errstr<<"Passed grade value ("<<grade<<") invalid - must be in "
				"[1, "<<NGRADES<<"]"<<Sas::error("InvalidGradeValue");
	if (node < 0 || node > INT(xmmPayLoad.rgs1.nNodes()))
		errstr<<"Passed node value ("<<node<<") invalid - must be in "
				"[0, "<<xmmPayLoad.rgs1.nNodes()<<"]"
			  <<Sas::error("InvalidNode");

	RETURN _acceptanceThresholds[node] / _qSplitR[grade-1];
} // RgsEnergyCorrector::acceptanceThreshold()


vector<float> RgsEnergyCorrector::qSplitRatios() const
{
	PROC

	vector<float> v;
	for (unsigned n=0; n<NGRADES; v.push_back(_qSplitR[n++]));
	RETURN v;
} // RgsEnergyCorrector::qSplitRatios()



unsigned short RgsEnergyCorrector::shapeToGrade(unsigned short shape) const
{
	PROC

	//
	//  SHAPE      GRADE
	//    0          1
	//    1          4
	//    2          2
	//    3          2
	//    4          3
	//    5          3
	//    6          3
	//    7          3
	//	 >7          2 + warning
	//
	static unsigned short trans[] = { 1, 4, 2, 2, 3, 3, 3, 3 };

	if (shape <NUM_OF(trans))
		ALTRETURNV(trans[shape]);

	errstr<<"Shape code "<<shape<<" has no associated grade value - shape "
			"codes must be in interval (0, "<<NUM_OF(trans)-1
		  <<"] - returning with grade 2"
		  <<Sas::warning("IllegalNullShape");

	RETURN 2;
} // RgsEnergyCorrector::shapeToGrade()


void RgsEnergyCorrector::randomize(const CalInt16Vector &eIn,
								   CalReal32Vector &eOut) const
{
	PROC

	calServer.setRandom(_eRandMin, _eRandMax);
	eIn.toReal(eOut);
	calServer.setRandom();

	ENDPROC
} // RgsEnergyCorrector::randomize()


RgsEnergyCorrector::~RgsEnergyCorrector()
{
	PROC

	unsigned nNodes = xmmPayLoad.rgs1.nNodes();
	for (unsigned n=0; n<nNodes; ++n)
		delete _offsetCurves[n];

	ENDPROC
} // RgsEnergyCorrector::~RgsEnergyCorrector()
