//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	GeometryDataServer.cc --- implementation of coordinate transformations
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-07-05: 1.0.0: - first version from CalAtomsCommon
//		UL, 1999-08-05: 1.1.0: - bug fix in toChipCoord: did not take OCB
//								 into account
//		UL, 1999-08-09: 1.2.0: - another bug fix in toChipCoord()
//							   - added CHIPCOORD->PIXCOORD1 conversion
//		UL, 1999-09-30: 1.3.0: - bug fix in xDoubleCoord()/yDoubleCoord()
//		UL, 1999-11-11: 1.4.0: - bug fix in toRowCoord(); was not in line
//								 with latest changes in LinCoord (M/G - point
//								 no longer given in SACCOORD but in telescope
//								 frame)
//		UL, 1999-11-30: 1.5.0: - toRowCoord() now takes rotation of RGA w.r.t
//								 S/C frame into account
//		UL, 2000-02-16: 1.5.1: - fixed compilation problem under Linux
//		UL, 2000-02-17: 1.6.0: - changed definition of cross-dispersion
//								 of ROWCOORD: this is now measured w.r.t
//								 the center of the mirror (not the G-point!)
//		UL, 2000-06-14: 1.7.0: - bug in toRowCoord() caused RGA misalignment
//								 correction to go in wrong direction:
//								 positive RGA +Y axis angles lead errorneously
//								 to greater BETA valus
//		UL, 2000-06-30: 1.8.0: - small fix to bring in line with selectlib-4.35
//		UL, 2001-03-08: 1.9.0: - bug fix in definition of CCD bounding boxes
//							   - implemented
//									XmmCartesianPointList::toChipCoord()
//									XmmCartesianPointList::onCcd()
//		UL, 2001-03-29: 1.10.0:- bug fix: construction of chip bounding
//								 boxes did not take OCB into account;
//								 (this gave problems in RGS CTI correction)
//		UL, 2001-04-23: 1.11.0:- XmmCartesianPointList::rowlandDefocusLength()
//								 takes optional DXOFFS from
//									RGS?_LINESPREADFUNC:FIGURE
//								 into account
//		UL, 2001-04-27: 1.12.0:- added
//								   GeometryDataServer::lsfDefocusDistribution()
//		UL, 2001-05-04: 1.13.0:- lsfDefocusDistribution: deltaBeta now depends
//								 on beta
//		UL, 2001-05-07: 1.13.1:- CHIPCOORD1->CC1 also works for negative
//								 pixel numbers
//		UL, 2001-06-01: 1.13.2:- bug fix in ROWCOORD->SACCOORD conversion
//								 (implements SPR 2054: CAL_rowCoordToSacCoord
//								  produces wrong result)
//		UL, 2001-09-20: 1.14.0:- distinguish between ocbX/ocbY in case of
//								 RGS HTR mode (see SPR 1996 for details)
//		UL, 2001-10-21: 1.15.0:- bug fix in XmmPixelList::toCamCoord1():
//								 _excessPixelX/Y was wrongly calculated - this
//								 lead to wrong RGS pixel sizes in OCB 3x3/5x5
//								 for the outermost columns (see email
//								 Doug Reynolds 2001-11-02)
//		UL, 2001-12-20: 1.16.0:- the size of the outermost RFC chip pixel
//								 columns was STILL not correctly calculated;
//								 a number of (tricky) fixes to make this work
//								 now (for all OCBs); see also new CalViewable
//		UL, 2001-12-21: 1.17.1:- another mod to XmmPixelList::yDoubleCoord()
//								 to fix problem with last pixel row in RGS 3x3
//								 (problem was inadvertently introduced with
//								 1.16.0)
//
//	AUTHOR:
//		U. Lammers (UL), Mon Jul  5 15:52:04 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/GeometryDataServer.cc,v 1.30 2002/02/15 15:22:33 ulammers Exp $
//

#include <algorithm>
#include <cmath>
#include <strstream>
#include "CalOalUtility.h"
#include "Random.h"
#include "CalServer.h"

using namespace CalOalUtility;
using namespace std;

GeometryDataServer *CalAtomCollection<GeometryDataServer>::select(
					Ccd ccd,
					Node node,
					unsigned) const
{
	_ccd = ccd;
	_node = node;
	unsigned n = index(ccd, 0);

	if (!(*_collection)[n])
		errstr<<"The CCF does not contain a GeometryDataServer data for CCD "
			  <<ccd
			  <<Sas::error("NoData");

	return (*_collection)[n];
} // CalAtomCollection<CtiData>::select()


template<class IntType>
void PixelList<IntType>::shiftX(IntType dx)
{
	unsigned N = _x.size();
	for (unsigned n=0; n<N; _x[n++] += dx);
} // PixelList::shiftX()
template void PixelList<int16>::shiftX(int16);

template<class IntType>
void PixelList<IntType>::shiftY(IntType dy)
{
	unsigned N = _y.size();
	for (unsigned n=0; n<N; _y[n++] += dy);
} // PixelList::shiftY()
template void PixelList<int16>::shiftY(int16);


void CartesianPointList::translate(const CartesianVector &offset)
{
	real64 dx = offset.x(),
		   dy = offset.y(),
		   dz = offset.z();

	unsigned N = _x.size();
	for (unsigned n=0; n<N; ++n) {
		_x[n] += dx;
		_y[n] += dy;
		_z[n] += dz;
	}
} // CartesianPointList::translate()


void CartesianPointList::rotate(const AttitudeMatrix &matrix)
{
	unsigned N = _x.size();
	for (unsigned n=0; n<N; ++n) {
		CartesianVector v0(_x[n], _y[n], _z[n]),
						v = matrix * v0;
		_x[n] = v.x();
		_y[n] = v.y();
		_z[n] = v.z();
	}
} // CartesianPointList::rotate()
						

XmmPixelList::XmmPixelList(CalInt16Vector &x,
						   CalInt16Vector &y,
						   CoordFrame frame,
						   GeometryDataServer *parent)
	: PixelList<int16>(x, y), CalChild<GeometryDataServer>(parent),
		_frame(frame)
{
	if (parent)
		_cfq = parent->coordFrameQuantities();
	else
		errstr<<"XmmPixelList must be instantiated with parent"
			  <<Sas::error("NoParent");
} // XmmPixelList::XmmPixelList()


XmmRealPixelList XmmPixelList::toReal()
{
	return XmmRealPixelList(_x, _y, _frame, this->parent());
} // XmmRealPixelList::toReal()


XmmPixelList &XmmPixelList::toPixCoord1()
{
	if (_frame == PIXCOORD1) {
		errstr<<"Data already in PIXCOORD1 frame"
			  <<Sas::warning("NothingToDo");
		return *this;
	}

	if (_frame != PIXCOORD0)
		errstr<<"Only conversion from PIXCOORD0 -> PIXCOORD1 is supported"
			  <<Sas::error("UnsupportedConversion");

	errstr<<"PIXCOORD0->PIXCOORD1 conversion not yet implemented"
		  <<Sas::warning("NotYetImplemented");

	_frame = PIXCOORD0;

	return *this;
} // XmmPixelList::toPixCoord1()


XmmPixelList &XmmPixelList::toChipCoord(const CalInt8Vector *nodes)
{
	if (_frame == CHIPCOORD) {
		errstr<<"Data already in CHIPCOORD frame"
			  <<Sas::warning("NothingToDo");
		return *this;
	}

	if (_frame != PIXCOORD1)
		errstr<<"Only conversion from PIXCOORD1 -> CHIPCOORD is supported"
			  <<Sas::error("UnsupportedConversion");

	_ocbY = calServer.state()->ocb()->isActive()
								? calServer.state()->ocb()->get() : 1;
	//	in HTR mode the X coordinate is always unbinned
	const Mode mode = calServer.state()->mode()->get();
	_ocbX = (mode != Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD &&
			 mode != Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD) ? _ocbY : 1;
	unsigned maxXocb = 2*UNT(ceil(DBL(_cfq.maxX/2)/DBL(_ocbX)));
	unsigned N = _x.size();
	if (!nodes) {
		if (calServer.state()->node()->isActive()) {
			int node = calServer.state()->node()->get();
			for (unsigned n=0; n<N; ++n)
				_x[n] = (maxXocb - _x[n] +1) * node + _x[n] * (1-node);
		}
	} else {
		const CalInt8Vector &nodesVec = *nodes;
		for (unsigned n=0; n<N; ++n)
			_x[n] = (maxXocb -_x[n] +1) * INT(nodesVec[n]) +
						_x[n] * (1-INT(nodesVec[n]));
	}

	_frame = CHIPCOORD;

	return *this;
} // XmmPixelList::toChipCoord()


XmmPixelList &XmmPixelList::toPixCoord1(CalInt8Vector &nodes)
{
	if (_frame == PIXCOORD1) {
		errstr<<"Data already in PIXCOORD1 frame"
			  <<Sas::warning("NothingToDo");
		return *this;
	}

	if (_frame != CHIPCOORD)
		errstr<<"Only conversion from CHIPCOORD -> PIXCOORD1 is supported"
			  <<Sas::error("UnsupportedConversion");

	unsigned N = _x.size();
	switch (calServer.instrument()->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2: {
			unsigned node = calServer.state()->node()->get(),
					 maxX = xmmPayLoad.emos1.maxX();
			assert(node <= 1);
			for (unsigned n=0; n<N; ++n)
				_x[n] = (maxX-_x[n]+1) * node + _x[n] * (1-node);
			break;
		}

		case Xmm::EPN:
			break;	//	nothing to be done

		case Xmm::RGS1:
		case Xmm::RGS2: {
			//	binodal readout
			//	EMOS also has a binodal mode; to be added later
			_ocbY = calServer.state()->ocb()->isActive()
									? calServer.state()->ocb()->get() : 1;
			//	in HTR mode the X coordinate is always unbinned
			const Mode mode = calServer.state()->mode()->get();
			_ocbX = (mode != Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD &&
				 mode != Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD) ? _ocbY : 1;

			unsigned maxXocb = 2*UNT(ceil(DBL(_cfq.maxX/2)/DBL(_ocbX))),
					 maxXocbHalf = maxXocb/2;

			for (unsigned n=0; n<N; ++n) {
				nodes[n] = (_x[n]-1)/maxXocbHalf;
				_x[n] = (maxXocb - _x[n] +1) * nodes[n] + _x[n] * (1-nodes[n]);
			}
			break;
		}
		default:
			errstr<<"CHIPCOORD->PIXCOORD1 conversion for instrument "
				  <<calServer.instrument()->id()<<" not supported"
				  <<Sas::error("UnsupportedConversion");
			break;
	}

	_frame = PIXCOORD1;

	return *this;
} // XmmPixelList::toPixCoord1()


XmmCartesianPointList XmmPixelList::toCamCoord1(real64 *xr,
												real64 *yr,
												real64 *zr)
{
	if (_frame != CHIPCOORD)
		errstr<<"Only conversion from CHIPCOORD -> CAMCOORD1 is supported"
			  <<Sas::error("UnsupportedConversion");

	CalReal64Vector xv(xr, x().size()),
					yv(yr, y().size()),
					zv(zr, x().size());

	XmmCartesianPointList cc1(xv, yv, zv, XmmCartesianPointList::CAMCOORD1,
							  parent());

	//
	//	calculate some quantities needed in conversion
	//
	_randomize = calServer.state()->randomize()->get() ? 1. : 0.;
	_ocbY = calServer.state()->ocb()->isActive()
								? calServer.state()->ocb()->get() : 1;
	//	in HTR mode the X coordinate is always unbinned
	const Mode mode = calServer.state()->mode()->get();
	_ocbX = (mode != Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD &&
			 mode != Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD) ? _ocbY : 1;

	_ocbrX = DBL(_ocbX);
	_ocbrY = DBL(_ocbY);
	_excessPixelsX = 0.;
	_excessPixelsY = 0.;
	if (_ocbY == 3 || _ocbY == 5) {
		int nx = INT(.5*DBL(_cfq.maxX)/_ocbrX),
			ny = INT(_cfq.maxY/_ocbrY);
		_excessPixelsX = .5*DBL(_cfq.maxX) - _ocbrX * DBL(nx);
		_excessPixelsY = _cfq.maxY - _ocbrY * DBL(ny);
	}

	_halfCentralOcbCeilX = ceil(DBL(_cfq.centralPixel.first)/_ocbrX);
	_halfCentralOcbCeilY = ceil(DBL(_cfq.centralPixel.second)/_ocbrY);

	//
	//	temperatures + expansion coefficients
	//
	double chipT = calServer.state()->temperature()->get(),
		   refChipT = _cfq.refChipT;
	if (chipT == 0.) {
		errstr<<"The actual chip temperature is set to 0 in the CAL "
				"state - will take reference chip temperature "
			  <<refChipT<<"K for the computation of thermal expansion"
			  <<Sas::warning("ZeroTemperature");
		chipT = refChipT;
	}
	_chipExpansion = 1. + (chipT - refChipT) * _cfq.expansionChip;

	double camT = calServer.state()->temperatureCamera()->get(),
		   refCamT = _cfq.refCamT;
	if (camT == 0.) {
		errstr<<"The actual camera temperature is set to 0 in the CAL "
				"state - will take reference camera temperature "
			  <<refCamT<<"K for the computation of thermal expansion"
			  <<Sas::warning("ZeroTemperature");
		camT = refCamT;
	}
	_camExpansion = 1. + (camT - refCamT) * _cfq.expansionCam;

	_r0 = _cfq.chipToCamCoord1Trans * _camExpansion;

	//	loop over all pixels
	unsigned N = x().size();
	for (unsigned n=0; n<N; ++n) {
		CartesianVector v0(xDoubleCoord(n), yDoubleCoord(n), 0.);

		CartesianVector v = _cfq.chipToCamCoord1Matrix * v0 +
						_r0;	//	roto-translation

		xv[n] = v.x();
		yv[n] = v.y();
		zv[n] = v.z();
	}

	return cc1;
} // XmmPixelList::toCamCoord1()


double XmmPixelList::toDoubleX(unsigned index) const
{
	return DBL(_x[index]);
} // XmmPixelList::toDoubleX

double XmmPixelList::toDoubleY(unsigned index) const
{
	return DBL(_y[index]);
} // XmmPixelList::toDoubleY


//
//	the following two methods are tricky - see CAL implementation notes
//	for details
//
inline double XmmPixelList::xDoubleCoord(unsigned index) const
{
	double xr = this->toDoubleX(index),
		   corrPixel = -.5*(_ocbrX-1.),
		   plusMinPixel = _randomize * _ocbrX * randGen.rand();

	if (_excessPixelsX > 0.) {
		if (xr <= 1.5)
			corrPixel += (xr-1.5)*(_excessPixelsX-_ocbrX);
		else if (xr >= 2.*_halfCentralOcbCeilX-.5)
			corrPixel += (xr-(2.*_halfCentralOcbCeilX-.5))*
													(_excessPixelsX-_ocbrX);
		plusMinPixel = _randomize * (_ocbrX - _excessPixelsX) * randGen.rand();
	}

	return (((xr - _halfCentralOcbCeilX) * _ocbrX) +
			corrPixel + plusMinPixel) * _chipExpansion * _cfq.plateScaleX;
} // XmmPixelList::xDoubleCoord()


inline double XmmPixelList::yDoubleCoord(unsigned index) const
{
	double yr = this->toDoubleY(index),
		   corrPixel = -.5*(_ocbrY-1.),
		   plusMinPixel = _randomize * _ocbrY * randGen.rand();

	double d = yr - (2.*_halfCentralOcbCeilY-.5);
	if (d <= 0.)
		return (((yr - _halfCentralOcbCeilY) * _ocbrY) + corrPixel +
				plusMinPixel) * _chipExpansion * _cfq.plateScaleY;
	else
		//	last pixel may have a different size
		return (((_halfCentralOcbCeilY-.5) * _ocbrY +
											corrPixel) * _cfq.plateScaleY +
				(d + _randomize * randGen.rand()) * max(_excessPixelsY,_ocbrY)*
				_cfq.plateScaleLastY) *	_chipExpansion;

	return 0.;
} // XmmCartesianPointList::yDoubleCoord()


XmmCartesianPointList XmmPixelList::toCamCoord1(CalReal64Vector &x,
												CalReal64Vector &y,
												CalReal64Vector &z)
{
	return toCamCoord1(x.ptr(), y.ptr(), z.ptr());
} // XmmPixelList::toCamCoord1()


vector<bool> XmmPixelList::insideFOV()
{
	unsigned N = _x.size();
	vector<bool> res(N);

	if (_frame == PIXCOORD0)
		toPixCoord1();
	if (_frame == PIXCOORD1)
		toChipCoord();

	const Proxy<FieldEvaluator> &fe = parent()->fieldEvaluator();
	for (unsigned n=0; n<N; ++n)
		res[n] = fe->insideRegion(DBL(_x[n]), DBL(_y[n]));

	return res;
} // XmmPixelList::insideFOV()


XmmRealPixelList::XmmRealPixelList(CalInt16Vector &x,
								   CalInt16Vector &y,
								   CoordFrame frame,
								   GeometryDataServer *parent)
	: CalChild<GeometryDataServer>(parent),
	  XmmPixelList(x, y, frame, parent)
{
	unsigned N = x.size();
	_xr = CalReal64Vector(0, N);
	_yr = CalReal64Vector(0, N);
	for (unsigned n=0; n<N; ++n) {
		_xr[n] = _yr[n] = 0.f;
	}
} // XmmRealPixelList::XmmRealPixelList()

namespace {
	CalInt16Vector _dumx, _dumy;
};

XmmRealPixelList::XmmRealPixelList(CalReal64Vector &xr,
								   CalReal64Vector &yr,
								   CoordFrame frame,
								   GeometryDataServer *parent)
	: CalChild<GeometryDataServer>(parent),
	  XmmPixelList(_dumx, _dumy, frame, parent),
	  _xr(xr), _yr(yr)
{
	_x = CalInt16Vector(0, xr.size());
	_y = CalInt16Vector(0, yr.size());
	unsigned N = xr.size();
	for (unsigned n=0; n<N; ++n)
		_x[n] = _y[n] = 0;

	init();
} // XmmRealPixelList::XmmRealPixelList()


real64 XmmRealPixelList::xr(unsigned index) const
{
	if (index >= _x.size())
		errstr<<"Index for XmmRealPixelList-x ("<<index<<") out of bounds - "
			  <<"must be in [0, "<<_x.size()-1<<"]"
			  <<Sas::error("IndexOutOfBounds");

	return _x[index] + _xr[index];
} // XmmRealPixelList::xr()


real64 XmmRealPixelList::yr(unsigned index) const
{
	if (index >= _y.size())
		errstr<<"Index for XmmRealPixelList-y ("<<index<<") out of bounds - "
			    "must be in [0, "<<_y.size()-1<<"]"
			  <<Sas::error("IndexOutOfBounds");

	return _y[index] + _yr[index];
} // XmmRealPixelList::yr()


void XmmRealPixelList::add(const CalReal64Vector &dx,
						   const CalReal64Vector &dy)
{
	unsigned N=_x.size();
	for (unsigned n=0; n<N; ++n) {
		double idum;
		_xr[n] = modf(xr(n)+dx[n], &idum); _x[n] = SHORT(idum);
		_yr[n] = modf(yr(n)+dy[n], &idum); _y[n] = SHORT(idum);
	}
} // XmmRealPixelList::add()


void XmmRealPixelList::init()
{
	unsigned N=_x.size();
	for (unsigned n=0; n<N; ++n) {
		double idum;
		_xr[n] = modf(xr(n), &idum); _x[n] = SHORT(idum);
		_yr[n] = modf(yr(n), &idum); _y[n] = SHORT(idum);
	}
} // XmmRealPixelList::init()


void XmmRealPixelList::normalize()
{
	unsigned N=_x.size();
	for (unsigned n=0; n<N; ++n) {
		double idum;
		_xr[n] = modf(_xr[n], &idum); _x[n] = SHORT(idum);
		_yr[n] = modf(_yr[n], &idum); _y[n] = SHORT(idum);
	}
} // XmmRealPixelList::normalize()


double XmmRealPixelList::toDoubleX(unsigned index) const
{
	return DBL(_x[index]) + _xr[index];
} // XmmRealPixelList::toDoubleX()

double XmmRealPixelList::toDoubleY(unsigned index) const
{
	return DBL(_y[index]) + _yr[index];
} // XmmRealPixelList::toDoubleY()


XmmCartesianPointList::XmmCartesianPointList(CalReal64Vector &x,
											 CalReal64Vector &y,
											 CalReal64Vector &z,
											 CoordFrame frame,
											 GeometryDataServer *parent)
	: CartesianPointList(x, y, z), CalChild<GeometryDataServer>(parent),
	  _frame(frame)
{
	if (parent)
		_cfq = parent->coordFrameQuantities();
	else
		errstr<<"XmmCartesianPointList must be instantiated with parent"
			  <<Sas::error("NoParent");

	double chipT = calServer.state()->temperature()->get(),
		   refChipT = _cfq.refChipT;
	if (chipT == 0.) {
		errstr<<"The actual chip temperature is set to 0 in the CAL "
				"state - will take reference chip temperature "
			  <<refChipT<<"K for the computation of thermal expansion"
			  <<Sas::warning("ZeroTemperature");
		chipT = refChipT;
	}

	_cc1ToCc2 = _cfq.cc1ToCc2 * (1. + (chipT - refChipT) * _cfq.expansionChip);
} // XmmCartesianPointList::XmmCartesianPointList()


XmmPixelList XmmCartesianPointList::toChipCoordSingleCcd(int16 *x,
														 int16 *y)
{
	if (_frame != CAMCOORD1)
		errstr<<"Only conversion from CAMCOORD1 -> CHIPCOORD is supported"
			  <<Sas::error("UnsupportedConversion");

	CalInt16Vector xv(x, _x.size()),
				   yv(y, _y.size());

	XmmPixelList cl(xv, yv, XmmPixelList::CHIPCOORD, parent());

	//
	//	calculate some quantities needed in conversion
	//
	bool oldRand = calServer.state()->randomize()->get();
    calServer.state()->randomize()->set(false);
	_ocbY = calServer.state()->ocb()->isActive()
								? calServer.state()->ocb()->get() : 1;
	//	in HTR mode the X coordinate is always unbinned
	const Mode mode = calServer.state()->mode()->get();
	_ocbX = (mode != Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD &&
			 mode != Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD) ? _ocbY : 1;

	_ocbrX = DBL(_ocbX);
	_ocbrY = DBL(_ocbY);
	_halfCentralOcbCeilX = ceil(DBL(_cfq.centralPixel.first)/_ocbrX);
	_halfCentralOcbCeilY = ceil(DBL(_cfq.centralPixel.second)/_ocbrY);

	//
	//	temperatures + expansion coefficients
	//
	double chipT = calServer.state()->temperature()->get(),
		   refChipT = _cfq.refChipT;
	if (chipT == 0.) {
		errstr<<"The actual chip temperature is set to 0 in the CAL "
				"state - will take reference chip temperature "
			  <<refChipT<<"K for the computation of thermal expansion"
			  <<Sas::warning("ZeroTemperature");
		chipT = refChipT;
	}
	_chipExpansion = 1. + (chipT - refChipT) * _cfq.expansionChip;

	double camT = calServer.state()->temperatureCamera()->get(),
		   refCamT = _cfq.refCamT;
	if (camT == 0.) {
		errstr<<"The actual camera temperature is set to 0 in the CAL "
				"state - will take reference camera temperature "
			  <<refCamT<<"K for the computation of thermal expansion"
			  <<Sas::warning("ZeroTemperature");
		camT = refCamT;
	}
	_camExpansion = 1. + (camT - refCamT) * _cfq.expansionCam;

	_r0 = _cfq.chipToCamCoord1Trans * _camExpansion;

	//	loop over all pixels
	unsigned N = _x.size();
	AttitudeMatrix A = _cfq.chipToCamCoord1Matrix.inverse();
	for (unsigned n=0; n<N; ++n) {
		CartesianVector v0 = CartesianVector(_x[n], _y[n], _z[n]) - _r0;
		CartesianVector v = A * v0;

		xv[n] = xIntCoord(v.x());
		yv[n] = yIntCoord(v.y());
	}
    calServer.state()->randomize()->set(oldRand);

	return cl;
} // XmmCartesianPointList::toChipCoord()


XmmPixelList XmmCartesianPointList::toChipCoord(int16 *x, int16 *y)
{
	vector<Ccd> ccds = onCcd();

	if (_frame != CAMCOORD1)
		toCamCoord1();

	unsigned N = _x.size();
	assert(N == ccds.size());
	CalInt16Vector xv(x, N),
				   yv(y, N);
	XmmPixelList cl(xv, yv, XmmPixelList::CHIPCOORD, parent());

	if (!N)
		return cl;

	unsigned n=0, k;
	Ccd ccd = ccds[0];
	while (n<N) {
		k = 0;
		while (k+n<N && ccds[n+k] == ccd)
			++k;
		if (ccd == Xmm::CCD_NONE) {
			errstr<<"Camcoord1 position ("
				  <<_x[n]<<", "<<_y[n]<<", "<<_z[n]
			      <<") does not lie on a CCD - will assign coordinates (0, 0)"
				  <<Sas::message(Msg::Lib, Msg::Verbose);
			for (unsigned i=n; i<n+k; ++i)
				xv[i] = yv[i] = 0;
			if (k+n<N)
				ccd = ccds[k+n];
			n += k;
			continue;
		}

		GeometryDataServer *gds = parent()->parent()->select(ccd,
															 Xmm::toNode(0),0);
		ccd = ccds[k+n];
		CalReal64Vector x(_x.ptr()+n, k),
						y(_y.ptr()+n, k),
						z(_z.ptr()+n, k);
		XmmCartesianPointList cc1 = gds->camCoord1List(x, y, z);
		cc1.toChipCoordSingleCcd(xv.ptr()+n, yv.ptr()+n);
		n += k;
	}

	return cl;
} // XmmCartesianPointList::toChipCoord()



inline int16 XmmCartesianPointList::xIntCoord(double xr)
{
	//	inverse of xDoubleCoord()
	return INT(xr/_cfq.plateScaleX/_chipExpansion/_ocbrX +
			   _halfCentralOcbCeilX +.5);
} // XmmCartesianPointList::xIntCoord()


inline int16 XmmCartesianPointList::yIntCoord(double yr)
{
	//	inverse of yDoubleCoord()
	return INT(yr/_cfq.plateScaleY/_chipExpansion/_ocbrY +
			   _halfCentralOcbCeilY + .5);
} // XmmCartesianPointList::yIntCoord()



XmmPixelList XmmCartesianPointList::toChipCoord(CalInt16Vector &x,
												CalInt16Vector &y)
{
	return toChipCoord(x.ptr(), y.ptr());
} // XmmCartesianPointList::toChipCoord()


XmmCartesianPointList &XmmCartesianPointList::toCamCoord1()
{
	if (_frame == CAMCOORD1) {
		errstr<<"Data already in CAMCOORD1 frame"
			  <<Sas::warning("NothingToDo");
		return *this;
	}

	if (_frame != CAMCOORD2)
		errstr<<"Only conversion from CAMCOORD2 -> CAMCOORD1 is supported"
			  <<Sas::error("UnsupportedConversion");

	this->translate(_cc1ToCc2);

	_frame = CAMCOORD1;

	return *this;
} // XmmCartesianPointList::toCamCoord1()


XmmCartesianPointList &XmmCartesianPointList::toCamCoord2()
{
	if (_frame == CAMCOORD2) {
		errstr<<"Data already in CAMCOORD2 frame"
			  <<Sas::warning("NothingToDo");
		return *this;
	}

	if (_frame != CAMCOORD1 && _frame != SACCOORD)
		errstr<<"Only conversion from CAMCOORD1/SACCOORD -> CAMCOORD2 is "
				"supported"
			  <<Sas::error("UnsupportedConversion");

	if (_frame == SACCOORD)
		toCamCoord2FromSacCoord();
	else		
		this->translate(-_cc1ToCc2);

	_frame = CAMCOORD2;

	return *this;
} // XmmCartesianPointList::toCamCoord2()


XmmCartesianPointList &XmmCartesianPointList::toSacCoord()
{
	if (_frame == SACCOORD) {
		errstr<<"Data already in SACOORD frame"
			  <<Sas::warning("NothingToDo");
		return *this;
	}

	if (_frame != CAMCOORD2)
		errstr<<"Only conversion from CAMCOORD2 -> SACCOORD is supported"
			  <<Sas::error("UnsupportedConversion");

	//	loop over all locations
	unsigned N = x().size();
	for (unsigned n=0; n<N; ++n) {
		CartesianVector v0(_x[n], _y[n], _z[n]);

		CartesianVector v = _cfq.cc2ToSacCoordMatrix * v0 +
							_cfq.cc2ToSacCoordTrans;	//	roto-translation

		_x[n] = v.x();
		_y[n] = v.y();
		_z[n] = v.z();
	}

	_frame = SACCOORD;

	return *this;
} // XmmCartesianPointList::toSacCoord()


XmmCartesianPointList &XmmCartesianPointList::toCamCoord2FromSacCoord()
{
	//	loop over all locations
	unsigned N = x().size();
	AttitudeMatrix A = _cfq.cc2ToSacCoordMatrix.inverse();
	for (unsigned n=0; n<N; ++n) {
		CartesianVector v0(_x[n], _y[n], _z[n]);

		v0 -= _cfq.cc2ToSacCoordTrans;
		CartesianVector v = A * v0;

		_x[n] = v.x();
		_y[n] = v.y();
		_z[n] = v.z();
	}

	_frame = CAMCOORD2;

	return *this;
} // XmmCartesianPointList::toCamCoord2FromSacCoord()


XmmRowlandPointList XmmCartesianPointList::toRowCoord(real64 *beta,
													  real64 *chi)
{
	if (_frame != SACCOORD)
		errstr<<"Only conversion from SACCOORD -> ROWCOORD is supported"
			  <<Sas::error("UnsupportedConversion");

	CalReal64Vector betav(beta, _x.size()),
					chiv(chi, _x.size());

	XmmRowlandPointList rl(betav, chiv, XmmRowlandPointList::ROWCOORD,
						   parent());

	unsigned N = _x.size();
	double GminusMMx = _cfq.G.x() - _cfq.MM0.x();
	for (unsigned n=0; n<N; ++n) {
		CartesianVector v0(_x[n] - _cfq.G.x(),
						   _y[n] - _cfq.G.y(),
						   _z[n] - _cfq.G.z()),
			v = parent()->coordFrameQuantities().sacCoordToRowCoordMatrix * v0;

		betav[n] = atan(-v.z()/v.x()) - _cfq.alpha;
		chiv[n] = atan(v.y()/(v.x() + GminusMMx));
	}

	return rl;
} // XmmCartesianPointList::toRowCoord()


XmmRowlandPointList XmmCartesianPointList::toRowCoord(CalReal64Vector &beta,
													  CalReal64Vector &chi)
{
	return toRowCoord(beta.ptr(), chi.ptr());
} // XmmCartesianPointList::toRowCoord()


XmmTelescopePointList XmmCartesianPointList::toTelCoord(real64 *theta,
													    real64 *phi)
{
	if (_frame != CAMCOORD2)
		errstr<<"Only conversion from CAMCOORD2 -> TELCOORD is supported"
			  <<Sas::error("UnsupportedConversion");

	CalReal64Vector thetav(theta, _x.size());
	CalReal64Vector phiv(phi, _x.size());

	XmmTelescopePointList tl(thetav, phiv, XmmTelescopePointList::TELCOORD,
							 parent());

	CalReal64Vector x(0, _x.size()), y(0, _x.size()), z(0, _x.size());
	XmmCartesianPointList pl(x, y, z, XmmCartesianPointList::CAMCOORD2,
							 parent());
	for (unsigned n=0; n<_x.size(); ++n) {
		pl.x()[n] = _x[n];
		pl.y()[n] = _y[n];
		pl.z()[n] = _z[n];
	}
	pl.toSacCoord();
	pl.translate(-_cfq.opticsToSacCoordTrans);
	//
	//	now origin is optical axis/focal plane intersection point and
	//	orientation is SACOORD;	calculate angular telescope coordinates
	//
	for (unsigned n=0; n<_x.size(); ++n) {
		double zmm = pl.z()[n], ymm = pl.y()[n];
		phiv[n] = atan2(ymm, zmm) + M_PI;	//	+M_PI: mirror inversion
		thetav[n] = sqrt(CalOalUtility::SQR(zmm * _cfq.arcsecsPerMMx) +
						 CalOalUtility::SQR(ymm * _cfq.arcsecsPerMMy));
	}

	return tl;
} // XmmCartesianPointList::toTelCoord()


XmmTelescopePointList XmmCartesianPointList::toTelCoord(CalReal64Vector &theta,
														CalReal64Vector &phi)
{
	return toTelCoord(theta.ptr(), phi.ptr());
} // XmmCartesianPointList::toTelCoord()


vector<Ccd> XmmCartesianPointList::onCcd()
{
	unsigned N = _x.size();
	vector<Ccd> res(N);

	if (_frame != CAMCOORD2)
		toCamCoord2();

	const vector<FBox> &boxlist = *parent()->ccdBoundingBoxes();
	for (unsigned n=0; n<N; ++n) {
		res[n] = Xmm::CCD_NONE;
		unsigned nCCDs = boxlist.size();
		for (unsigned i=0; i<nCCDs; ++i)
			if (boxlist[i].contains(FPoint(_x[n], _y[n]))) {
				res[n] = i+1;
				break;
			}
	}

	return res;
} // XmmCartesianPointList::onCcd()


CalReal64Vector XmmCartesianPointList::rowlandDefocusLength(double theta,
															double phi)
{
	if (_frame == CAMCOORD1)
		toCamCoord2();
	if (_frame == CAMCOORD2)
		toSacCoord();

	GratingDataServer *gds = calServer.getAtom(gds);
	double alpha0Prime = gds->alpha0();
	const CartesianVector &G = parent()->coordFrameQuantities().G;

	unsigned N = _x.size();
	double betaV[N], chiV[N];
	CalReal64Vector beta(betaV, N),
					chi(chiV, N);
	for (unsigned n=0; n<N; ++n) {
		CartesianVector r = CartesianVector(_x[n], _y[n], _z[n]) - G;

		betaV[n] = atan2(-r.z(), r.x()) - alpha0Prime;
		chiV[n] = 0.;
	}

	return parent()->rowCoordList(beta, chi).defocusLength(theta, phi);
} // XmmCartesianPointList::rowlandDefocusLength()


XmmTelescopePointList::XmmTelescopePointList(CalReal64Vector &theta,
					  CalReal64Vector &phi,
					  CoordFrame frame,
					  GeometryDataServer *parent)
	: AngularPointList<CalReal64Vector>(theta, phi),
	  CalChild<GeometryDataServer>(parent),
	  _frame(frame)
{
	if (!parent)
		errstr<<"XmmCartesianPointList must be instantiated with parent"
			  <<Sas::error("NoParent");
} // XmmTelescopePointList::XmmTelescopePointList()


XmmCartesianPointList XmmTelescopePointList::toCamCoord2(CalReal64Vector &x,
														 CalReal64Vector &y,
														 CalReal64Vector &z)
{
	return toCamCoord2(x.ptr(), y.ptr(), z.ptr());
} // XmmTelscopePointList::toCamCoord2()


XmmCartesianPointList XmmTelescopePointList::toCamCoord2(real64 *x,
														 real64 *y,
														 real64 *z)
{
	CalReal64Vector xv(x, phi().size()),
					yv(y, phi().size()),
					zv(z, phi().size());

	XmmCartesianPointList cc2(xv, yv, zv, XmmCartesianPointList::SACCOORD,
							  parent());

	unsigned N = phi().size();
	for (unsigned n=0; n<N; ++n) {
		double arcsecsZ = -theta()[n] * cos(phi()[n]),
               arcsecsY = -theta()[n] * sin(phi()[n]);
		//	minus signs: mirror inversion: cos(phi+M_PI) = -cos(phi)
		//								   sin(phi+M_PI) = -sin(phi)
		zv[n] = arcsecsZ/parent()->coordFrameQuantities().arcsecsPerMMx;
		yv[n] = arcsecsY/parent()->coordFrameQuantities().arcsecsPerMMy;
		xv[n] = 0.;
	}

	cc2.translate(parent()->coordFrameQuantities().opticsToSacCoordTrans);
	cc2.toCamCoord2();

	return cc2;
} // XmmTelescopePointList::toCamCoord2()


XmmRowlandPointList::XmmRowlandPointList(CalReal64Vector &beta,
										 CalReal64Vector &chi,
										 CoordFrame frame,
										 GeometryDataServer *parent)
	: AngularPointList<CalReal64Vector>(beta, chi),
	  CalChild<GeometryDataServer>(parent),
	  _frame(frame)
{
	if (!parent)
		errstr<<"XmmRowlandPointList must be instantiated with parent"
			  <<Sas::error("NoParent");
} // XmmRowlandPointList::XmmRowlandPointList()


XmmCartesianPointList XmmRowlandPointList::toSacCoord(CalReal64Vector &x,
													  CalReal64Vector &y,
													  CalReal64Vector &z)
{
	return toSacCoord(x.ptr(), y.ptr(), z.ptr());
} // XmmTelscopePointList::toCamCoord2()


XmmCartesianPointList XmmRowlandPointList::toSacCoord(real64 *x,
													  real64 *y,
													  real64 *z)
{
	CalReal64Vector xv(x, phi().size()),
					yv(y, phi().size()),
					zv(z, phi().size());



	XmmCartesianPointList sc(xv, yv, zv, XmmCartesianPointList::SACCOORD,
							 parent());

	unsigned N = beta().size();
	CartesianVector G = parent()->coordFrameQuantities().G,
					M = parent()->coordFrameQuantities().M;
	double			alpha = parent()->coordFrameQuantities().alpha,
					GMdist = CartesianVector(G - M).norm();

	errstr<<"The ROWCOORD->SACCOORD transformation yields merely "
			 "approximate S/C coordinate values (all points on Rowland circle)"
		  <<Sas::warning("ApproximateResults");

	for (unsigned n=0; n<N; ++n) {
		CartesianVector v0(cos(beta()[n] + alpha) * GMdist,
						   sin(chi()[n]) * GMdist,
						   -sin(beta()[n] + alpha) * GMdist),

						v = parent()->coordFrameQuantities().
								sacCoordToRowCoordMatrix.inverse() * v0 + G;

		xv[n] = v.x();
		yv[n] = v.y();
		zv[n] = v.z();
	}

	return sc;
} // XmmRowlandPointList::toSacCoord()


CalReal64Vector XmmRowlandPointList::defocusLength(double theta, double phi)
{
	GratingDataServer *gds = calServer.getAtom(gds);
	double alpha0Prime = gds->alpha0();
	MiscDataServer *mds = calServer.getAtom(mds);
	double gamma0 = (mds->get("INCIDENCE_ANGLE") +
					 mds->get("BLAZE_ANGLE")) * DEG_TO_RAD,
		   F = mds->get("FOCAL_LENGTH"), // [mm]
		   R = mds->get("ROWLAND"),	//	[mm];
		   dxOffs = 0.;
	if (mds->has("LINESPREADFUNC:FIGURE%DXOFFS"))
		dxOffs = mds->get("LINESPREADFUNC:FIGURE%DXOFFS");

	double alphaCorr = -F/(2.*R*cos(gamma0)) * theta/3600.*DEG_TO_RAD*cos(phi),
		   alpha = alpha0Prime + alphaCorr,
		   L = 2.*R*cos(gamma0 + alphaCorr),
		   FPrimeMinusL = F*cos(-theta/3600.*DEG_TO_RAD*cos(phi))-L;

	const CartesianVector &G = parent()->coordFrameQuantities().G;

	unsigned N = beta().size();
	CalReal64Vector l(0, N);
	for (unsigned n=0; n<N; ++n) {
		double x = RowlandBeta(beta()[n]).x() - G.x();
		l[n] = x - 2.*R*cos(alpha0Prime+beta()[n]-gamma0) * 
												cos(alpha0Prime+beta()[n]) -
				(G.x()-FPrimeMinusL) * sin(alpha)/sin(beta()[n]) + dxOffs;
	}

	return l;
} // XmmRowlandPointList::defocusLength()


//
//	---( RowlandBeta )---------------------------------------------------------
//
InterpolatedCurve *RowlandBeta::_betaToX = 0;
Xmm::Instrument RowlandBeta::_ins;

void RowlandBeta::init() const
{
	delete _betaToX;
	const XmmInstrument *ins = calServer.instrument();
	_ins = ins->id();
	calServer.pushState();
	calServer.setInstrument(_ins);
	calServer.state()->ocb()->set(1);
	calServer.state()->randomize()->set(false);

	unsigned N = ins->maxX();
	int16 x[N], y[N];
	for (unsigned n=0; n<N; ++n) {
		x[n] = N-n;
		y[n] = 0;
	}
	CalInt16Vector xv(x, N),
				   yv(y, N);
	unsigned nCcds = ins->nCcds();
	vector<double> xV, betaV;
	for (Ccd ccd=nCcds; ccd>=1; --ccd) {
		calServer.state()->ccd()->set(ccd);
		GeometryDataServer *gds = calServer.getAtom(gds);
		XmmPixelList pl = gds->chipCoordList(xv, yv);
		double xmm[N], ymm[N], zmm[N];
		XmmCartesianPointList cpl = pl.toCamCoord1(xmm, ymm, zmm).
										toCamCoord2().toSacCoord();
		double beta[N], chi[N];
		XmmRowlandPointList rpl = cpl.toRowCoord(beta, chi);

		for (unsigned n=0; n<N; ++n) {
			xV.push_back(xmm[n]);
			betaV.push_back(beta[n]);
		}
	}
	calServer.popState();

	_betaToX = new InterpolatedCurve(betaV, xV);
} // RowlandBeta::init()


double RowlandBeta::x() const
{
	if (!_betaToX || calServer.instrument()->id() != _ins)
		init();

	return (*_betaToX)(_beta);
} // RowlandBeta::x()


//
//	---( GeometryDataServer )--------------------------------------------------
//
Xmm::Ccd GeometryDataServer::onCcd(double x, double y)
{
	double z = 0.;
	CalReal64Vector xv(&x, 1), yv(&y, 1), zv(&z, 1);

	return camCoord2List(xv, yv, zv).onCcd()[0];
} // GeometryDataServer::onCcd()


//
//	The following implementation is based on code received
//	from Andy Rasmussen, 2001-04-25
//
Spectrum GeometryDataServer::lsfDefocusDistribution(double nBetaHWHM,
													double betaChannelWidth,
													double theta,
													double phi,
													int order,
													double E)
{
	MiscDataServer *mds = calServer.getAtom(mds);
	double rMirrorInner = mds->get("R_INNER"),	//	[mm]
		   rMirrorOuter = mds->get("R_OUTER"),	//	[mm]
		   F = mds->get("FOCAL_LENGTH"),		// [mm]
		   F2 = F*F;
	GratingDataServer *gds = calServer.getAtom(gds);
	double betaM = gds->beta(gds->alpha(theta, phi), order, E);

	//	compute defocus distance at beta=betaM
	double betav[] = { betaM },
		   chiv[] = { 0. };
	CalReal64Vector betaV(betav, 1),
					chiV(chiv, 1);
	XmmRowlandPointList rpl = rowCoordList(betaV, chiV);
	CalReal64Vector defocusLength = rpl.defocusLength(theta, phi);
	double defocus = defocusLength[0];	//	[mm]
	//	defocus is how many mm the focal plane is from the position of best
	//	focus.

	double norm1=1./(M_PI *(1.-CalOalUtility::SQR(rMirrorInner/rMirrorOuter))),
		   norm2=1./(M_PI*(-1.+CalOalUtility::SQR(rMirrorOuter/rMirrorInner)));

	// outer & inner radii of the mirror assembly projected onto the beta scale
	double outerDbeta = defocus * rMirrorOuter/F2,
		   innerDbeta = defocus * rMirrorInner/F2;

	Spectrum spec;
	//
	//	chosen crossover point - if the following condition is true then
	//	the convolution kernel is constructed 'by hand' rather than by
	//	using the formula of arcsines & angles etc.
	//
	if (defocus <= 1.5 * F2/rMirrorOuter * betaChannelWidth) {
		//	construct the 3 nonzero elements of the spectrum using
		//	the expected variance directly.

		//	this is the formula for the variance of the defocus contribution.
		double variance = M_PI*(norm1*CalOalUtility::SQR(outerDbeta)/4. -
								norm2*CalOalUtility::SQR(innerDbeta)/4.),
			   p1 = variance/(2.* betaChannelWidth*betaChannelWidth);

		SAS_DEBUG(cout<<"defocus_shortcut: p1 = "<<p1<<endl;)
		spec.resize(3);
		spec[0] = spec[2] = p1;
		spec[1] = 1. - 2.*p1;
	} else {
		//
		//	non-shortcut way to evaluate the defocus contribution
		//	the spectrum is a 1-dimensional projection of a doughnut
		//
		double betaSigma = 2.*nBetaHWHM/M_2SQ2LN2;
		const double SIGMA	= 5.e-4;	//	typical width in [rad]
		unsigned nChanHalf = UNT(SIGMA/betaChannelWidth * betaSigma + .5);
		spec.resize(2*nChanHalf+1);

		double ratMin = 0.,
			   ratMax = min(.5 * fabs(betaChannelWidth)/outerDbeta, 1.);

		spec[nChanHalf] = 2.*norm1*(asin(ratMax)-asin(ratMin) +
								ratMax * sqrt(1.-ratMax*ratMax) -
								ratMin * sqrt(1.-ratMin*ratMin));
		double addContrib = spec[nChanHalf];
		ratMin = 0.,
		ratMax = min(.5 * fabs(betaChannelWidth)/innerDbeta, 1.);
		spec[nChanHalf] -= 2.*norm2*(asin(ratMax)-asin(ratMin) +
								ratMax * sqrt(1.-ratMax*ratMax) -
								ratMin * sqrt(1.-ratMin*ratMin));
		double subContrib = addContrib - spec[nChanHalf],
			   totalContrib = spec[nChanHalf];
		unsigned addnContrib = 0,
				 subnContrib = 0;
		for (unsigned i=1; i<nChanHalf; ++i) {
			spec[nChanHalf+i] = spec[nChanHalf-i] = 0.;
			double deltaBeta = i * betaChannelWidth;

			ratMin = min(fabs((DBL(i)-.5)*deltaBeta/outerDbeta), 1.);
			ratMax = min(fabs((DBL(i)+.5)*deltaBeta/outerDbeta), 1.);
			if (ratMin != ratMax) {
				//	some contribution to make
				double contrib = norm1*(asin(ratMax) - asin(ratMin) +
									ratMax*sqrt(1.-ratMax*ratMax) -
									ratMin*sqrt(1.-ratMin*ratMin));
				spec[nChanHalf+i] = spec[nChanHalf-i] = contrib;
				addContrib += 2.*contrib;
				totalContrib += 2.*contrib;
				addnContrib += 2;
			}
			ratMin = min(fabs((DBL(i)-.5)*deltaBeta/innerDbeta), 1.);
			ratMax = min(fabs((DBL(i)+.5)*deltaBeta/innerDbeta), 1.);
			if (ratMin != ratMax) {
				//	some contribution to make
				double contrib = norm2*(asin(ratMax) - asin(ratMin) +
									ratMax*sqrt(1.-ratMax*ratMax) -
									ratMin*sqrt(1.-ratMin*ratMin));
				spec[nChanHalf+i] -= contrib;
				spec[nChanHalf-i] -= contrib;
				subContrib += 2.*contrib;
				totalContrib -= 2.*contrib;
				subnContrib += 2;
			}
		}
	}

	return spec;
} // GeometryDataServer::lsfDefocusDistribution()



CalAtomCollection<GeometryDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs * _nNodes;
	GeometryDataServer *null = 0;
	_collection = new vector<GeometryDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::LinCoord);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet,
							e->queryBlockName(Ccf::toString(Ccf::LinCoord)));
	const Table *table = ccftab.block();
	ColumnData	*cdccdId	= table->column("CCD_ID")->data(),
				*cdnodeId	= table->column("NODE_ID")->data(),
				*cdrawX0	= table->column("RAWX0")->data(),
				*cdrawY0	= table->column("RAWY0")->data(),
				*cdx0		= table->column("X0")->data(),
				*cdy0		= table->column("Y0")->data(),
				*cdz0		= table->column("Z0")->data(),
				*cddx0		= table->column("D_X0")->data(),
				*cddy0		= table->column("D_Y0")->data(),
				*cddz0		= table->column("D_Z0")->data(),
				*cdeulx		= table->column("EULER_PHI")->data(),
				*cdeuly		= table->column("EULER_THETA")->data(),
				*cdeulz		= table->column("EULER_PSI")->data(),
				*cddeulx	= table->column("D_EULER_PHI")->data(),
				*cddeuly	= table->column("D_EULER_THETA")->data(),
				*cddeulz	= table->column("D_EULER_PSI")->data();

	int8	*ccdId	= cdccdId->int8Data(),
			*nodeId	= cdnodeId->int8Data();
	int16	*rawX0	= cdrawX0->int16Data(),
			*rawY0	= cdrawY0->int16Data();
	real32	*x0		= cdx0->real32Data(),
			*y0		= cdy0->real32Data(),
			*z0		= cdz0->real32Data(),
			*eulx	= cdeulx->real32Data(),
			*euly	= cdeuly->real32Data(),
			*eulz	= cdeulz->real32Data();

	CoordFrameQuantities cfq;
	cfq.plateScaleX = table->attribute("MM_PX_X")->asReal();
	cfq.plateScaleY = table->attribute("MM_PX_Y")->asReal();
	cfq.arcsecsPerMMx = table->attribute("ARC_MM_X")->asReal();
	cfq.arcsecsPerMMy = table->attribute("ARC_MM_Y")->asReal();
	cfq.plateScaleLastY = cfq.plateScaleY;
	if (table->hasAttribute("MM_PX_LY"))
		cfq.plateScaleLastY = table->attribute("MM_PX_LY")->asReal();
	cfq.maxX = instrument->maxX();
	cfq.maxY = instrument->maxY();

	//	CC1 to CC2 vector
	double ccdx, ccdy, ccdz;
	ccdx = ccdy = ccdz = 0.;
	if (table->hasAttribute("CC12_TX")) {
		ccdx = table->attribute("CC12_TX")->asReal();
		ccdy = table->attribute("CC12_TY")->asReal();
		ccdz = table->attribute("CC12_TZ")->asReal();
	}
	cfq.cc1ToCc2 = CartesianVector(ccdx, ccdy, ccdz);

	cfq.refChipT = table->attribute("R_CHIP_T")->asReal();
	cfq.refCamT = table->attribute("R_CAM_T")->asReal();
	cfq.expansionChip = table->attribute("EXP_CHIP")->asReal();
	cfq.expansionCam = table->attribute("EXP_CAM")->asReal();

	//	CAMCOORD2 to SACCOORD
	cfq.MM0 = CartesianVector(table->attribute("MM_TX")->asReal(),
							  table->attribute("MM_TY")->asReal(),
							  table->attribute("MM_TZ")->asReal());
	cfq.cc2ToSacCoordTrans = 
			CartesianVector(table->attribute("CC2_TX")->asReal(),
							table->attribute("CC2_TY")->asReal(),
							table->attribute("CC2_TZ")->asReal()) +
						cfq.MM0;
	cfq.sacCoordToCc2.phi(table->attribute("CC2_RPH")->asReal() * DEG_TO_RAD);
	cfq.sacCoordToCc2.theta(
			table->attribute("CC2_RTH")->asReal() * DEG_TO_RAD);
	cfq.sacCoordToCc2.psi(
		   table->attribute("CC2_RPS")->asReal() * DEG_TO_RAD);

	cfq.cc2ToSacCoordMatrix =
		AttitudeMatrix(cfq.sacCoordToCc2.phi(), cfq.sacCoordToCc2.theta(),
					   cfq.sacCoordToCc2.psi()).inverse();

	//	SACCOORD to ROWCOORD
	double gx, gy, gz;
	gx = gy = gz = 0.;
	if (table->hasAttribute("G_TX")) {
		gx = table->attribute("G_TX")->asReal();
		gy = table->attribute("G_TY")->asReal();
		gz = table->attribute("G_TZ")->asReal();
		cfq.gry = table->attribute("G_RY")->asReal() * DEG_TO_RAD;
		cfq.sacCoordToRowCoordMatrix =
			AttitudeMatrix(table->attribute("G_RX")->asReal() * DEG_TO_RAD,
						   table->attribute("G_RY")->asReal() * DEG_TO_RAD,
						   table->attribute("G_RZ")->asReal() * DEG_TO_RAD,
						   EulerAngles::Euler123);
	}
	cfq.G = CartesianVector(gx, gy, gz) + cfq.MM0;
	cfq.alpha = 0.;
	cfq.M = cfq.cc2ToSacCoordTrans;
	if (table->hasAttribute("G_ALPHA"))
		cfq.alpha = table->attribute("G_ALPHA")->asReal() * DEG_TO_RAD;


	//
	//	instaniate default FieldEvaluator
	//
	strstream s;
	s<<"(CHIPX, CHIPY) in rectangle("
	 <<toString(instrument->minX())<<", "
	 <<toString(instrument->minY())<<", "
	 <<toString(instrument->maxX())<<", "
	 <<toString(instrument->maxY())<<", 0.)"<<ends;
	Proxy<FieldEvaluator> defaultFieldEvaluator = new FieldEvaluator(s.str());
	const Table *fovTable = 0;
	if (dataSet->hasBlock("FOV")) {
		CcfConstituentTable ccftab(dataSet, "FOV");
		fovTable = ccftab.block();
	} else
		errstr<<"`"<<dataSet->qualifiedName()<<"' has no FOV table - all "
				"chips are considered to be fully within the FOV"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

	//
	//	chip vertices
	//
	unsigned minX = instrument->minX(),
			 maxX = instrument->maxX(),
			 minY = instrument->minY(),
			 maxY = instrument->maxY();

	Proxy<vector<FBox> > ccdBoundingBoxes = new vector<FBox>;
	bool oldRand = calServer.state()->randomize()->get();
    calServer.state()->randomize()->set(false);
	unsigned oldOcb = 0;
	if (calServer.state()->ocb()->isActive()) {
		oldOcb = calServer.state()->ocb()->get();
		calServer.state()->ocb()->set(1);
	}
	for (unsigned n=0; n<table->rows(); ++n) {
		if (ccdId[n] <= 0 || ccdId[n] > INT(_nCCDs) ||
			nodeId[n] >= INT(_nNodes))
			errstr<<"Illegal chip and/or node number in row #"
				  <<n<<" of CCF file "
				  <<Xmm::toString(instrument->id())<<"/"
				  <<Ccf::toString(Ccf::LinCoord)
				  <<Sas::error("IllegalValues");

		//	integer pixel numbers refer to the center of the pixel;
		//	chipToCamCoord1Trans is therefore a vector from the camera
		//	center to the center of integer pixel rawX0/rawY0
		cfq.centralPixel = PixelLocation(rawX0[n], rawY0[n]);
		cfq.chipToCamCoord1Trans = CartesianVector(x0[n], y0[n], z0[n]);
		cfq.chipToCamCoord1Matrix =
						AttitudeMatrix(eulx[n], euly[n], eulz[n]).inverse();

		unsigned pos = index(ccdId[n], nodeId[n]);
		GeometryDataServer *&geometryDataServer = (*_collection)[pos];
		if (!geometryDataServer) {
			Proxy<FieldEvaluator> fe(defaultFieldEvaluator);
			GeometryDataServer *p = (*_collection)[index(ccdId[n], 0)];
			if (p)
				//	already assigned fieldEvaluator for node 0 - use
				//	the same for other node
				fe = p->fieldEvaluator();
			else if (fovTable) {
				ColumnData *cdccd = fovTable->column("CCD_ID")->data(),
						   *cdreg = fovTable->column("REGION")->data();
				int8 *ccd = cdccd->int8Data();
				unsigned i;
				for (i=0; i<fovTable->rows(); ++i)
					if (ccd[i] == ccdId[n]) {
						fe = Proxy<FieldEvaluator>(new FieldEvaluator(
								stringColumnRow(cdreg, i)));
						break;
					}
				if (i >= fovTable->rows())
					errstr<<"No FOV region definition for CCD "<<INT(ccdId[n])
						  <<" in `"<<fovTable->qualifiedName()<<"' CCD is "
							"considered to be fully within the FOV"
						  <<Sas::warning("FullyVisible");

				delete cdccd;
				delete cdreg;
			}

			geometryDataServer = new GeometryDataServer(cfq, fe, this);

			//
			//	determine bounding boxes of CCDs
			//
			if (ccdId[n] <= ccdBoundingBoxes->size())
				continue;

	        double chipX[] = { minX-.5, maxX+.5, maxX+.5, minX-.5 },
				  chipY[] = { minY-.5, minY-.5, maxY+.5, maxY+.5 };
			if (cfq.plateScaleY != cfq.plateScaleLastY) {
				//	last pixel in column has different size
				//	correct position
				chipY[2] -= cfq.plateScaleLastY/cfq.plateScaleY - 1. -.5;
				chipY[3] -= cfq.plateScaleLastY/cfq.plateScaleY - 1. -.5;
			}
			CalReal64Vector chipXVec(chipX, NUM_OF(chipX)),
							chipYVec(chipY, NUM_OF(chipY));
			double xmm[NUM_OF(chipX)], ymm[NUM_OF(chipX)], zmm[NUM_OF(chipX)];
			XmmRealPixelList pl = geometryDataServer->chipCoordList(chipXVec,
																	chipYVec);
			XmmCartesianPointList cpl = pl.toCamCoord1(xmm, ymm, zmm);
	        cpl.toCamCoord2();
			vector<double> xv(xmm, xmm+NUM_OF(xmm)),
						   yv(ymm, ymm+NUM_OF(ymm));
			sort(xv.begin(), xv.end());
			sort(yv.begin(), yv.end());
			double cenx = .25 * (xmm[0]+xmm[1]+xmm[2]+xmm[3]),
				   ceny = .25 * (ymm[0]+ymm[1]+ymm[2]+ymm[3]),
				   dx = .5*(fabs(xv[3]-xv[0]) + fabs(xv[2]-xv[1])),
				   dy = .5*(fabs(yv[3]-yv[0]) + fabs(yv[2]-yv[1]));

			//	approximate shape of CCD as box(cenx, ceny, xwidth, ywidth)
			//	i.e. disregard chip rotations
			ccdBoundingBoxes->push_back(FBox(FPoint(cenx, ceny), dx, dy));
		}
	}

	//
	//	determine opticsToSacCoordTrans vector: get OPTICS_X/OPTICS_Y
	//	(CHIPCOORD) from CCF and convert to SACCCORD; for this instantiate
	//	temporary geometry data server
	//
	MiscDataServer *mds = calServer.getAtom(mds);
	double opticsX = 0.,
		   opticsY = 0.;
	if (mds->has("OPTICS_X") && mds->has("OPTICS_Y")) {
		opticsX = mds->get("OPTICS_X");
		opticsY = mds->get("OPTICS_Y");
	} else {
		opticsX = .5*DBL(instrument->maxX()-instrument->minX()+1);
		opticsY = .5*DBL(instrument->maxY()-instrument->minY()+1);
	}
	unsigned opticsCcd = mds->has("OPTICS_CCD")
							? UNT(mds->get("OPTICS_CCD")) : 1;
	GeometryDataServer *gds = (*_collection)[index(opticsCcd, 0)];
	CalReal64Vector xv(&opticsX, 1),
					yv(&opticsY, 1);
	double xmm, ymm, zmm;
	XmmCartesianPointList cpl =
		gds->chipCoordList(xv, yv).toCamCoord1(&xmm, &ymm, &zmm).toCamCoord2().
			toSacCoord();
	CartesianVector vec(xmm, ymm, zmm);

	//	register CCD chip boxes and opticsToSacCoordTrans vector
	for (unsigned n=0; n<_nEntries; ++n)
		if ((*_collection)[n]) {
			(*_collection)[n]->registerCcdBoxes(ccdBoundingBoxes);
			(*_collection)[n]->_cfq.opticsToSacCoordTrans = vec;
		}

	//	restore CAL state
    calServer.state()->randomize()->set(oldRand);
	if (oldOcb)
		calServer.state()->ocb()->set(oldOcb);

	//	tidy up
	delete cdccdId;
	delete cdnodeId;
	delete cdrawX0;
	delete cdrawY0;
	delete cdx0;
	delete cdy0;
	delete cdz0;
	delete cddx0;
	delete cddy0;
	delete cddz0;
	delete cdeulx;
	delete cdeuly;
	delete cdeulz;
	delete cddeulx;
	delete cddeuly;
	delete cddeulz;

	dataSetServer->close(dataSet);
} // CalAtomCollection<GeometryDataServer>::CalAtomCollection()



FieldEvaluator::FieldEvaluator(const string &region)
{
	if (!Evaluator::isValidExpression(region))
		errstr<<"The region specification `"<<region<<"' is syntactically "
				"incorrect"<<Sas::error("SyntaxError");

	std::string regStr = replaceall(lowercase(region),
									"polygon(", "polygon2(");
	
	_evaluator = new Evaluator(regStr);
	NumericSymbols syms = _evaluator->numericIdentifiers();
	_pchipX = _pchipY = &_chipDum;
	if (syms.name.size() == 2) {
		if (uppercase(syms.name[0]) == "CHIPX" &&
			uppercase(syms.name[1]) == "CHIPY") {
			_pchipX = syms.valuePtr[0];
			_pchipY = syms.valuePtr[1];
		} else if (uppercase(syms.name[0]) == "CHIPY" &&
				   uppercase(syms.name[1]) == "CHIPX") {
			_pchipX = syms.valuePtr[1];
			_pchipY = syms.valuePtr[0];
	} else
		errstr<<"Unrecognized symbols "<<syms.name[0]<<'/'
			  <<syms.name[1]<<" in FOV region expression - expected "
				"CHIPX + CHIPY"
			  <<Sas::error("UnknownSymbols");
    } else if (syms.name.size() != 0)
        errstr<<"Number of symbols in FOV REGION specification `"
              <<region<<"' must be two: CHIPX + CHIPY"
              <<Sas::error("IllegalNumberOfSymbols");
} // FieldEvaluator::FieldEvaluator()


bool FieldEvaluator::insideRegion(double chipX, double chipY) const
{
	*_pchipX = chipX;
	*_pchipY = chipY;

	return _evaluator->evalLogic();
} // FieldEvaluator::insideRegion()


FieldEvaluator::~FieldEvaluator()
{
	delete _evaluator;
} // FieldEvaluator::~FieldEvaluator()
