//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	MiscDataServer.cc --- implementation of MiscDataServer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-09-08: 1.0.0: - first incarnation (from CalAtomsXmm.cc)
//		UL, 2000-05-05: 1.1.0: - more descriptive error message if data
//								 item is not found (SPR 1264)
//		UL, 2001-04-23: 1.2.0: - MiscDataServer::has() can now handle
//								 set:tab%attr syntax
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/MiscDataServer.cc,v 1.12 2001/10/29 16:57:59 ulammers Exp $
//

#include "ExceptionErrorHandler.h"
#include "CalServer.h"
#include "CalOalUtility.h"
#include "stringutil.h"
#include "MiscDataServer.h"

using namespace CalOalUtility;
using namespace std;

class MiscDataValue {
public:
	MiscDataValue(double val) :
		_val(val), _ref(0), _hasValue(true)	{ PROC ENDPROC }
	MiscDataValue(const string &ref) :
		_val(0.), _ref(new string(ref)), _hasValue(false) { PROC ENDPROC }

	bool hasValue() const { PROC RETURN _hasValue; }
	double value() const
		{ PROC if (!_hasValue)
				   errstr<<"Data item has no value"<<Sas::error("NoValue");
		  RETURN _val; }
	const string &reference() const
		{ PROC if (_hasValue)
				   errstr<<"Data item has value"<<Sas::error("HasValue");
		  RETURN *_ref; }

	~MiscDataValue() { PROC delete _ref; ENDPROC }

private:
	double			_val;
	const string	*_ref;
	bool			_hasValue;
};

ostream &operator<<(ostream &os, const MiscDataValue *p)
{
	PROC

	if (p->hasValue())
		os<<p->value();
	else
		os<<"value to be found in "<<p->reference();

	RETURN os;
} // operator<<


//
//	---( MiscDataServer )---------------------------------------------------
//
CalAtomCollection<MiscDataServer>::CalAtomCollection(
					const XmmInstrument *instrument,
					const Ccf *ccf,
					Xmm::Filter filter,
					AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_collection	= new vector<MiscDataServer *>;

	CifEntry *e = ccf->entry(instrument, Ccf::MiscData);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet,
							  e->queryBlockName(Ccf::toString(Ccf::MiscData)));
	const Table *table = ccftab.block();

	_collection->push_back(new MiscDataServer(table->name()));
	MiscDataServer &mds = *(*_collection)[0];

	ColumnData *cdvalue = table->column("PARM_VAL")->data();
	real64 *value = cdvalue->real64Data();

	errstr<<"Reading miscellaneous data from file "<<e->fileName()
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	//	read all parameters corresponding to the current instrument, the
	//	relevant telescope and the satellite
	string instrumentName(Xmm::toString(instrument->id())),
		   telescopeName(Xmm::toString(instrument->behindTelescope())),
		   xmmName(Xmm::toString(Xmm::SC));

	ColumnData *insCol = table->column("INSTRUMENT_ID")->data(),
			   *parmCol = table->column("PARM_ID")->data(),
			   *findInCol = table->column("FIND_IT_IN")->data();
	for (unsigned n=0; n<table->rows(); ++n) {
		string instId(uppercase(stringColumnRow(insCol, n)));
		if (instId == instrumentName || instId == telescopeName ||
			instId == xmmName) {
			string parmName(uppercase(stringColumnRow(parmCol, n))),
				   findItIn(trim(uppercase(stringColumnRow(findInCol, n))));
			if (findItIn.length())
				//	reference to other data set given
				mds[parmName] = new MiscDataValue(findItIn);
			else
				mds[parmName] = new MiscDataValue(value[n]);

			if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
				errstr<<parmName<<" = "<<value[n]<<endl
					  <<Sas::message(Msg::Lib, Msg::Verbose);
		}
	}

	delete insCol;
	delete parmCol;
	delete findInCol;
	delete cdvalue;

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<MiscData>::CalAtomCollection()


void MiscDataServer::dump(ostream &os)
{
	PROC

	os<<"MiscDataServer{\n";
	for (const_iterator i = begin(); i != end(); ++i)
		os<<(*i).first<<"="<<(*i).second<<endl;
	os<<"}";

	ENDPROC
} // MiscDataServer::dump()


MiscDataServer *CalAtomCollection<MiscDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[0];
} // MiscDataCollection<MiscDataServer>::select()



bool MiscDataServer::has(const string &keyword) const
{
	PROC

	ExceptionErrorHandler local;
	ErrorHandler* oldHandler = errHandler.handler(&local);
	try {
		get(keyword);
	} catch (SasException::Error) {
		errHandler.handler(oldHandler);
		ALTRETURNV(false);
	}

	errHandler.handler(oldHandler);
	RETURN true;
} // MiscDataServer::has()


double MiscDataServer::get(const string &keyname) const
{
	PROC

	if (keyname.rfind(':') != string::npos)
		//	keyname contains ':' - look for it in other constituent
		ALTRETURNV (getValueFromOtherConstituent(keyname));

	const_iterator i = find(uppercase(keyname));
	if (i != end())
		if ((*i).second->hasValue()) {
			ALTRETURNV ((*i).second->value());
		} else {
			ALTRETURNV (getValueFromOtherConstituent(
												(*i).second->reference()));
		}
	else
		errstr<<"There is no miscellaneous data item `"<<keyname<<"' in `"
			  <<_tableName<<"'"<<Sas::error("MiscalleneousDataItemNotFound");

	RETURN 0.;
} // MiscDataServer::get()


double MiscDataServer::getValueFromOtherConstituent(const string &key) const
{
	PROC

	string s(key);

	vector<string> v = breakup(s, ":%");
	if (v.size() != 3)
		errstr<<"`"<<s<<"' is not a valid reference to an attribute in "
				"another CCF constituent"
			  <<Sas::error("InvalidAttributeReference");

	const CifEntry *ce = calServer.ccf()->
							entry(calServer.instrument(), Ccf::toTypeId(v[0]));
	DataSet *set = ce->open();
	double val = set->hasBlock(v[1])
						? set->block(v[1])->attribute(v[2])->asReal()
						: set->attribute(v[2])->asReal();
	dataSetServer->close(set);

	RETURN val;
} // MiscDataServer::getValueFromOtherConstituent()


MiscDataServer::~MiscDataServer()
{
	for (MiscDataMapType::const_iterator i = begin();
		 i != end(); ++i)
		delete (*i).second;
} // MiscDataServer::~MiscDataServer()

