//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmColorTransformator.cc --- implementation of OmColorTransformator class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2000-01-11: 1.2.0: - initial version
//		UL, 2001-10-18: 1.3.0: - fixed bug in
//									OmColorTransformator::toNaturalMagnitudes()
//								 which lead to too small error estimates
//								 in computed magnitude (see also SPR 2215)
//
//	AUTHOR:
//		U. Lammers (UL), Tue Jan 11 16:42:35 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/OmColorTransformator.cc,v 1.12 2001/10/29 16:58:01 ulammers Exp $
//

#include <cmath>

#include "CalServer.h"
#include "OmColorTransformator.h"
#include "CalOalUtility.h"
#include "Xmm.h"
#include "stringutil.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( OmColorTransformator ) ----------------------------------------------
//
OmColorTransformator::OmColorTransformator(
							const CcfConstituentTable &colorTransTab,
							CalAtomCollection<OmColorTransformator> *parent)
	: CalChild<CalAtomCollection<OmColorTransformator> >(parent),
	  _f1(Xmm::FILTER_NONE), _f2(Xmm::FILTER_NONE), _branch(0),
	  _selectedTrafo(0)
{
	PROC

	//	set ALGOID
	_algoid = colorTransTab.algoid();

	//
	//	populate _zeroPointMap + _apertureRadiiMap from corresponding
	//	attributes
	//
	const Table *table = colorTransTab.block();

	const vector<Filter> &omFilters = xmmPayLoad.om.filters();
	vector<Filter> filters;
	vector<double> zeroPoints;
	for (unsigned n=0; n<omFilters.size(); ++n) {
		string fname(Xmm::toString(omFilters[n])),
			   attName = uppercase("ZPT"+ fname).substr(0, 8);
		if (table->hasAttribute(attName))
			_zeroPointMap[omFilters[n]] = table->attribute(attName)->asReal();
		attName = uppercase("APE" + fname).substr(0, 8);
		if (table->hasAttribute(attName))
			_apertureRadiiMap[omFilters[n]] =
										table->attribute(attName)->asReal();
	}

	unsigned N = table->rows();
	vector<string> f1, f2;
	CalOalUtility::fillVector(table, "FILTERID1", f1);
	CalOalUtility::fillVector(table, "FILTERID2", f2);
	vector<unsigned short> branch;
	if (table->hasColumn("BRANCH"))
		CalOalUtility::fillVector(table, "BRANCH", branch);
	else
		//	treat missing BRANCH column as if all trafos are referring to
		//	branch 0
		branch = vector<unsigned short>(N, 0);

	//
	//	loop over all rows - each row defines one color transformation
	//
	for (unsigned n=0; n<N; ++n) {
		vector<double> vid;
		CalOalUtility::fillVector(table, "TRAFOLIMIT", n, vid);
		if (vid.size() != 2)
			errstr<<"Validity limits of color transformation in `"
				  <<table->qualifiedName()<<"' must be in vector of length 2"
				  <<Sas::error("CcfIntegrityError");
		ValidityInterval vi(vid[0], vid[1]);

		ColorTrafo ct;
		CalOalUtility::fillVector(table, "TRAFOP1", n, ct.colorTrafo);
		CalOalUtility::fillVector(table, "TRAFOP1E", n, ct.colorTrafoError);
		CalOalUtility::fillVector(table, "TRAFOP2", n, ct.magniTrafo);
		CalOalUtility::fillVector(table, "TRAFOP2E", n, ct.magniTrafoError);
		if (ct.colorTrafo.size() != ct.colorTrafoError.size() ||
		    ct.colorTrafo.size() != ct.magniTrafo.size() ||
			ct.colorTrafo.size() != ct.magniTrafoError.size())
			errstr<<"Color transformation coefficient vectors in `"
				  <<table->qualifiedName()<<"' have different lengths"
				  <<Sas::error("CcfIntegrityError");

		registerTrafo(Xmm::toFilter(f1[n]),	Xmm::toFilter(f2[n]),
					  branch[n], vi, ct);
	}

	ENDPROC
} // OmColorTransformator::OmColorTransformator()


void OmColorTransformator::registerTrafo(Filter f1, Filter f2,
										 unsigned short branch,
										 const ValidityInterval &vi,
										 const ColorTrafo &colorTrafo)
{
	PROC

	ColorTrafoMap::iterator i = _colorTrafoMap.find(FilterPair(f1, f2));
	if (i != _colorTrafoMap.end()) {
		//	there is already a trafo with this filter-combo
		ColorTrafoBranches &branches = i->second;
		if (branches.size() <= branch)
			//	need to make more space
			branches.resize(branch+1);
		branches[branch][vi] = colorTrafo;
	} else {
		ColorTrafoBranches branches(branch+1);
		branches[branch][vi] = colorTrafo;
		_colorTrafoMap[FilterPair(f1, f2)] = branches;
	}

	ENDPROC
} // OmColorTransformator::registerTrafo()


void OmColorTransformator::dump(ostream &os)
{
	PROC

	os<<"OmColorTransformator{\n";
	os<<"}";

	ENDPROC
} // OmColorTransformator::dump()


const OmColorTransformator &OmColorTransformator::withFilters(Filter f1,
															  Filter f2)
const
{
	PROC

	//	check that on of the filters is either U, B, or V
	if (f1 != Xmm::U && f1 != Xmm::B && f1 != Xmm::V &&
		f2 != Xmm::U && f2 != Xmm::B && f2 != Xmm::V)
		errstr<<"One of the input filters ("<<f1<<'/'<<f2<<") must be "
				"U, B, or V for the color transformation to work"
			  <<Sas::error("IllegalFilter");

	//	check that f1 is at a shorter wavelength than f2 - they are wavelength
	//	ordered in the enum
	if (UNT(f1) >= UNT(f2))
		errstr<<"The color transformation is only defined if the first filter "
				"("<<f1<<") is at shorter wavelength than the second ("<<f2
			  <<")."<<Sas::error("IncorrectFilterSequence");

	//	check that the f1/f2 combo is supported
	_selectedTrafoList = _colorTrafoMap.find(FilterPair(f1, f2));
	if (_selectedTrafoList == _colorTrafoMap.end())
		errstr<<"The color transformation for the filter combination "
			  <<f1<<'/'<<f2<<" is not defined"
			  <<Sas::error("UnsupportedFilterCombination");

	_f1 = f1;
	_f2 = f2;

	RETURN *this;
} // OmColorTransformator::withFilters()


void OmColorTransformator::checkFilters() const
{
	PROC

	if (_f1 == Xmm::FILTER_NONE || _f2 == Xmm::FILTER_NONE)
		errstr<<"Please select first two input filter via "
				"OmColorTransformator::withFilters(f1, f2)"
			  <<Sas::error("FiltersUndefined");

	ENDPROC
} // OmColorTransformator::checkFilters()


unsigned short OmColorTransformator::branches() const
{
	PROC

	checkFilters();
	RETURN _selectedTrafoList->second.size();
} // OmColorTransformator::branches()


const OmColorTransformator &OmColorTransformator::branch(
												unsigned short branchId) const
{
	PROC

	checkFilters();
	if (_selectedTrafoList->second.size() > branchId)
		_branch = branchId;
	else
		errstr<<"There are/is only "<<_selectedTrafoList->second.size()
			  <<" branche(s) in "<<_f1<<'-'<<_f2<<" color transformation - "
				"hence cannot select branch "<<branchId<<'.'
			  <<Sas::error("TooFewBranches");

	RETURN *this;
} // OmColorTransformator::branch()


vector<OmColorTransformator::ValidityInterval>
								OmColorTransformator::validityRanges() const
{
	PROC

	checkFilters();
	const ColorTrafoList &ctl = _selectedTrafoList->second[_branch];
	vector<ValidityInterval> vi;
	for (ColorTrafoList::const_iterator i = ctl.begin(); i!=ctl.end(); ++i)
		vi.push_back(i->first);

	RETURN vi;
} // OmColorTransformator::validityRanges()


ValueError OmColorTransformator::standardColor(double magnitude1,
											   double magnitude2) const
{
	PROC

	checkFilters();

	//	find the applicable tranformation
	double instColor = magnitude1 - magnitude2;
	const ColorTrafoList &ctl = _selectedTrafoList->second[_branch];
	ColorTrafoList::const_iterator i =
					ctl.upper_bound(ValidityInterval(instColor, 1.e100)),
					j = i, k=i;

	//	select the transformation whose validity range covers <instColor>
	if (ctl.begin() != ctl.end() && (
		(i == ctl.end()
			&& (i != ctl.begin() && (--j)->first.contains(instColor))) ||
		(--k)->first.contains(instColor)))
		_selectedTrafo = const_cast<ColorTrafo *>(&(--i)->second);
	else
		errstr<<"The color term in the instrumental system ("
			  <<instColor<<") does not lie within any of the validity ranges "
				"for branch "<<_branch<<" of the color transformation"
			  <<Sas::error("ColorOutsideValidityRange");

	assert(_algoid <= 1);

	if (_algoid == 0)
		ALTRETURNV(standardColor0(instColor));

	RETURN standardColor1(instColor);
} // OmColorTransformator::standardColor()


ValueError OmColorTransformator::standardColor0(double instColor) const
{
	PROC

	const ColorTrafo &ct = selectedTrafo();
	if (ct.colorTrafo.size() < 3)
		errstr<<"Need at least three coefficients for the color transformation"
				" but found only "<<ct.colorTrafo.size()<<" - something is "
				"wrong in CCF::ColorTrans"<<Sas::error("CcfIntegrityError");

	const double &k0 = ct.colorTrafo[0],
				 &k1 = ct.colorTrafo[1],
				 &k2 = ct.colorTrafo[2];

	double sqrtArg = k1*k1 - 4.*k2 * (k0 - instColor);
	if (sqrtArg < 0)
		errstr<<"Square-root argument less than zero - cannot perform "
				"color transformation"
			  <<Sas::error("NegativeSqrtArg");
	if (k2 == 0.)
		errstr<<"Coefficient k2 of color transformation is 0"
			  <<Sas::error("IllegalDataValue");

	ValueError ve;
	double sqrtVal = sqrt(sqrtArg);
	ve.value = (-k1 + sqrtVal)/(2.*k2);
	ve.error = 0.;

	//	compute variance
	if (sqrtVal != 0.) {
		const double &k0e = ct.colorTrafoError[0],
					 &k1e = ct.colorTrafoError[1],
					 &k2e = ct.colorTrafoError[2];
		double oneOverSqrtVal = 1./sqrtVal,
			   dk0 = -oneOverSqrtVal,
			   dk1 = (k1 * oneOverSqrtVal - 1.)/(2.*k2),
			   dk2 = ((k1 - sqrtVal)/k2 - 2.*(k0 - instColor)*oneOverSqrtVal)/
						(2.*k2);

			   ve.error = sqrt(dk0*dk0*k0e*k0e+
							   dk1*dk1*k1e*k1e+
							   dk2*dk2*k2e*k2e);
	} else
		errstr<<"Zero denominator - cannot compute Color transformation error"
			  <<Sas::warning("ZeroDenominator");

	RETURN ve;
} // OmColorTransformator::standardColor0()


ValueError OmColorTransformator::standardColor1(double instColor) const
{
	PROC

	const ColorTrafo &ct = selectedTrafo();

	//	evaluate polynomial
	unsigned N = ct.colorTrafo.size() - 1,
			 j = N;
	double p = ct.colorTrafo[N],
		   variance = CalOalUtility::SQR(ct.colorTrafoError[0]),
		   instColorPow = 1.;
	while (j>0) {
		p = p*instColor + ct.colorTrafo[--j];
		instColorPow *= instColor;
		variance += CalOalUtility::SQR(ct.colorTrafoError[N-j]) *
					CalOalUtility::SQR(instColorPow);
	}

	ValueError ve;
	ve.value = p;
	ve.error = sqrt(variance);

	RETURN ve;
} // OmColorTransformator::standardColor1()


const OmColorTransformator::ColorTrafo &OmColorTransformator::selectedTrafo()
const
{
	PROC

	if (!_selectedTrafo)
		errstr<<"No color transformation has been selected yet - please "
				"do that by invoking OmColorTransformator::standardColor()"
			  <<Sas::error("UndefinedColorTransformation");

	RETURN *_selectedTrafo;
} // OmColorTransformator::selectedTrafo()


ValueError OmColorTransformator::standardMagnitude(ValueError standardColor,
												   double magnitude) const
{
	PROC

	assert(_algoid <= 1);
	if (_algoid == 0)
		ALTRETURNV(standardMagnitude0(standardColor, magnitude));

	RETURN standardMagnitude1(standardColor, magnitude);
} // OmColorTransformator::standardMagnitude()


ValueError OmColorTransformator::standardMagnitude0(ValueError standardColor,
												    double magnitude) const
{
	PROC

	const ColorTrafo &ct = selectedTrafo();

	ValueError ve;
	if (ct.magniTrafo.size() < 3)
		errstr<<"Need at least three coefficients for the transformation "
				"into the standard system but found only "
			  <<ct.magniTrafo.size()
			  <<" - something is "
				"wrong in CCF::ColorTrans"<<Sas::error("CcfIntegrityError");

	const double &l0 = ct.magniTrafo[0],
				 &l1 = ct.magniTrafo[1],
				 &l2 = ct.magniTrafo[2],
				 &l0e = ct.magniTrafoError[0],
				 &l1e = ct.magniTrafoError[1],
				 &l2e = ct.magniTrafoError[2];

	ve.value = -l2 * standardColor.value * standardColor.value -
				l1 * standardColor.value - l0 + magnitude;

	double dl0 = -1.,
		   dl1 = - standardColor.value,
		   dl2 = - standardColor.value * standardColor.value,
		   dc = -2.*l2 * standardColor.value;

	ve.error = sqrt(dl0*dl0*l0e*l0e+
					dl1*dl1*l1e*l1e+
					dl2*dl2*l2e*l2e+
					dc*dc*standardColor.error*standardColor.error);

	RETURN ve;
} // OmColorTransformator::standardMagnitude0()


ValueError OmColorTransformator::standardMagnitude1(ValueError standardColor,
												    double magnitude) const
{
	PROC

	const ColorTrafo &ct = selectedTrafo();

	//	evaluate polynomial
	unsigned N = ct.magniTrafo.size() - 1,
			 j = N;
	double p = ct.magniTrafo[N],
		   variance = CalOalUtility::SQR(ct.colorTrafoError[0]),
		   colorPow = 1.,
		   pv = 0.;
	while (j>0) {
		p = p*standardColor.value + ct.magniTrafo[--j];
		colorPow *= standardColor.value;
		variance += CalOalUtility::SQR(ct.magniTrafoError[N-j]) *
					CalOalUtility::SQR(colorPow);
		pv = pv*standardColor.value + ct.magniTrafo[j]*j;
	}

	ValueError ve;
	ve.value = p + magnitude;
	ve.error = sqrt(variance + pv*pv*standardColor.error*standardColor.error);

	RETURN ve;
} // OmColorTransformator::standardMagnitude1()


ValueAsymmetricError OmColorTransformator::toNaturalMagnitudes(
							const ValueAsymmetricError &naturalRate) const
{
	PROC

	Filter f = calServer.state()->filter()->get();
	FilterPropertyMap::const_iterator i = _zeroPointMap.find(f);
	if (i == _zeroPointMap.end())
		errstr<<"There is no zero point for the count-rate-to-magnitude "
				"conversion corresponding to filter "<<f<<" defined in "
				"CCF::ColorTrans"<<Sas::error("NoSuchFilter");

	if (naturalRate.value < 0.)
		errstr<<"Negative count rate ("<<naturalRate.value<<") cannot be "
				"converted into natural magnitude value"
			  <<Sas::error("NegativeCountRate");

	ValueAsymmetricError vae;
	vae.value = (*i).second - 2.5 * std::log10(naturalRate.value);
	double fact = -2.5/std::log(10.);
	vae.error.errorPlus = std::fabs(fact * naturalRate.error.errorPlus/
										  naturalRate.value);
	vae.error.errorMinus = std::fabs(fact * naturalRate.error.errorMinus/
										  naturalRate.value);

	RETURN vae;
} // OmColorTransformator::toNaturalMagnitudes()


double OmColorTransformator::apertureRadius() const
{
	PROC

	Filter f = calServer.state()->filter()->get();
	FilterPropertyMap::const_iterator i = _apertureRadiiMap.find(f);
	if (i == _apertureRadiiMap.end())
		errstr<<"There is no aperture radius for filter "<<f<<" defined in "
				"CCF::ColorTrans"<<Sas::error("NoSuchFilter");

	RETURN i->second;
} // OmColorTransformator::apertureRadius()


OmColorTransformator::~OmColorTransformator()
{
	PROC
	ENDPROC
} // OmColorTransformator::~OmColorTransformator()


//
//	---( CalAtomCollection<OmColorTransformator> ) ---------------------------
//
CalAtomCollection<OmColorTransformator>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	OmColorTransformator *null = 0;
	_collection = new vector<OmColorTransformator *>(_nEntries, null);

	//
	//	read data from the CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::ColorTrans);
	DataSet *dataSet = e->open();
	CcfConstituentTable colorTransTab(dataSet, "COLORMAG", 1);

	//
	//  construct new OmColorTransformator
	//
    OmColorTransformator *&omColorTransformator = (*_collection)[0];
	omColorTransformator = new OmColorTransformator(colorTransTab, this);


	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<OmColorTransformator>::CalAtomCollection()


OmColorTransformator *CalAtomCollection<OmColorTransformator>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC

	RETURN (*_collection)[0];
} // CalAtomCollection<OmColorTransformator>::select
