//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmGeoLinearizer.cc --- implementation of OmGeoLinearizer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2000-01-11: 1.2.0: - initial version
//		UL, 2000-07-28: 1.3.0: - bug fix in reading of POLYNOM_MAP
//		UL, 2001-05-04: 1.4.0: - access magnifier filter table as
//								 "FILTER-MAGNI" (see also SPR 2002)
//
//	AUTHOR:
//		U. Lammers (UL), Tue Jan 11 16:42:35 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//		[3]: CCF ICD
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/OmGeoLinearizer.cc,v 1.12 2001/10/29 16:58:02 ulammers Exp $
//

#include "OmGeoLinearizer.h"
#include "CalOalUtility.h"

using namespace std;

//
//	---( CalAtomCollection<OmGeoLinearizer> ) --------------------------------
//
CalAtomCollection<OmGeoLinearizer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	OmGeoLinearizer *null = 0;
	_collection = new vector<OmGeoLinearizer *>(_nEntries, null);

	//
	//	read the principal data from the CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::AstroMet);
	DataSet *dataSet = e->open();
	string filterName = uppercase(Xmm::toString(filter));
	if (filter == Xmm::MAGNI)
		//
		//	OM Magnifier filter is referred to as "MAGNI" in [2] + [3]
		//	but we like to continue using the long name in the SAS;
		//	since [3] is frozen this is only possible with this ugly
		//	construct
		//
		filterName.erase(5); // "MAGNIFIER" -> "MAGNI"

	CcfConstituentTable ccftab(dataSet, "FILTER-" +	filterName);
	const Table *table = ccftab.block();

	ColumnData *cdrawx = table->column("RAWX")->data(),
			   *cdrawy = table->column("RAWY")->data(),
			   *cdoffx = table->column("RAWX_OFF")->data(),
			   *cdoffy = table->column("RAWY_OFF")->data();

	int16	*rawX		= cdrawx->int16Data(),
			*rawY		= cdrawy->int16Data();
	real32	*offX		= cdoffx->real32Data(),
			*offY		= cdoffy->real32Data();

	unsigned N = table->rows();
	vector<real32> x(N), y(N), xo(N), yo(N);
	for (unsigned n=0; n<N; ++n) {
		x[n] = FLT(rawX[n]);
		y[n] = FLT(rawY[n]);
		xo[n] = offX[n];
		yo[n] = offY[n];
	}
	double plateScale = table->attribute("PLTSCALE")->asReal();

	//
	//	read polynomial coefficients
	//
	vector<double>	xa, ya;
	if (dataSet->hasBlock("POLYNOM_MAP")) {
		table = dataSet->table("POLYNOM_MAP");
		ColumnData *cdfilterId = table->column("FILTER_ID")->data(),
				   *cdxav = table->column("XPOLYCOEF")->data(),
				   *cdyav = table->column("YPOLYCOEF")->data();

		//	search the right entry
		unsigned N = table->rows(), n;
		string filterName = Xmm::toString(filter);
		for (n=0; n<N; ++n)
			if (filterName == CalOalUtility::stringColumnRow(cdfilterId, n)) {
				unsigned M = cdxav->elements();
				real32 *xav = cdxav->real32Data(),
					   *yav = cdyav->real32Data();
				for (unsigned m=0; m<M; ++m) {
					unsigned pos = cdxav->position(n, m);
					xa.push_back(xav[pos]);
					ya.push_back(yav[pos]);
				}
				break;
			}

		if (n>=N)
			errstr<<"`"<<table->qualifiedName()<<"' does not contain an "
					"entry for filter `"<<Xmm::toString(filter)
				  <<"' - polynomial offset calculation will not work"
				  <<Sas::warning("NoSuchFilterEntry");

		delete cdfilterId;
		delete cdxav;
		delete cdyav;
	} else
		errstr<<"`"<<dataSet->qualifiedName()<<"' does not contain a table "
				"with name `POLYNOM_MAP' - polynomial offset calculation will "
				"not work"<<Sas::warning("NoSuchTable");


	//
	//	construct new OmGeoLinearizer
	//
	OmGeoLinearizer *&omGeoLinearizer = (*_collection)[0];
	omGeoLinearizer = new OmGeoLinearizer(x, y, xo, yo, xa, ya, plateScale,
										  this);

	//	tidy up
	delete cdrawx;
	delete cdrawy;
	delete cdoffx;
	delete cdoffy;

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<OmGeoLinearizer>::CalAtomCollection()


void OmGeoLinearizer::dump(ostream &os)
{
	PROC

	os<<"OmGeoLinearizer{\n";
	os<<"}";

	ENDPROC
} // OmGeoLinearizer::dump()


OmGeoLinearizer *CalAtomCollection<OmGeoLinearizer>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC

	RETURN (*_collection)[0];
} // CalAtomCollection<OmGeoLinearizer>::select


//
//	---( OmGeoLinearizer> ) -----------------------------------------------
//
OmGeoLinearizer::OmGeoLinearizer(const vector<real32> &rawX,
			 					 const vector<real32> &rawY,
								 const vector<real32> &deltaX,
								 const vector<real32> &deltaY,
								 const vector<double>  &xa,
								 const vector<double>  &ya,
								 double plateScale,
								CalAtomCollection<OmGeoLinearizer> *parent)
	: CalChild<CalAtomCollection<OmGeoLinearizer> >(parent), _ax(xa), _ay(ya),
	  _plateScale(plateScale)
{
	PROC

	_xoffInter =
		new Slatec::Bivariate2DInterpolantScattered<float>(rawX, rawY, deltaX);
	_yoffInter =
		new Slatec::Bivariate2DInterpolantScattered<float>(rawX, rawY, deltaY);

	ENDPROC
} // OmGeoLinearizer::OmGeoLinearizer()


void OmGeoLinearizer::additionalPoints(unsigned nPoints)
{
	PROC

	_xoffInter->additionalDataPoints(nPoints);
	_yoffInter->additionalDataPoints(nPoints);

	ENDPROC
} // OmGeoLinearizer::additionalPoints()


void OmGeoLinearizer::interpolatedDeltas(const CalReal32Vector &rawX,
										 const CalReal32Vector &rawY,
										 CalReal32Vector &deltaX,
										 CalReal32Vector &deltaY,
										 CalReal32Vector &sigmaX,
										 CalReal32Vector &sigmaY) const
{
	PROC

	unsigned N = rawX.size();
	for (unsigned n=0; n<N; ++n) {
		deltaX[n] = (*_xoffInter)(FLT(rawX[n]), FLT(rawY[n]));
		deltaY[n] = (*_yoffInter)(FLT(rawX[n]), FLT(rawY[n]));
		//	we do not know uncertainties - set the sigma vectors to 0
		sigmaX[n] = 0.f;
		sigmaY[n] = 0.f;
	}

	ENDPROC
} // OmGeoLinearizer::interpolatedDeltas()


void OmGeoLinearizer::polynomialDeltas(const CalReal32Vector &rawX,
									   const CalReal32Vector &rawY,
									   CalReal32Vector &deltaX,
									   CalReal32Vector &deltaY,
									   CalReal32Vector &sigmaX,
									   CalReal32Vector &sigmaY) const
{
	PROC

	if (_ax.size() != 36 || _ay.size() != 36)
		errstr<<"Vector with polynomial coefficients too small - size "
				"is "<<_ax.size()<<" - cannot evaluate 7th order "
				"polynomial"<<Sas::error("InsufficientData");

	unsigned N = rawX.size();
	for (unsigned n=0; n<N; ++n) {
		double x1 = DBL(rawX[n]),
			   y1 = DBL(rawY[n]),
			   x2 = x1*x1,
			   x3 = x2*x1,
			   x4 = x2*x2,
			   x5 = x4*x1,
			   x6 = x3*x3,
			   x7 = x6*x1,
			   y2 = y1*y1,
			   y3 = y2*y1,
			   y4 = y2*y2,
			   y5 = y4*y1,
			   y6 = y3*y3,
			   y7 = y6*y1;

		deltaX[n] = _ax[0] + _ax[1]*y1 + _ax[2]*y2 * _ax[3]*y3 + _ax[4]*y4 +
					_ax[5]*y5 + _ax[6]*y6 + _ax[7]*y7 +

					_ax[8]*x1 + _ax[9]*x1*y1 + _ax[10]*x1*y2 + _ax[11]*x1*y3+
					_ax[12]*x1*y4 + _ax[13]*x1*y5 + _ax[14]*x1*y6 +

					_ax[15]*x2 + _ax[16]*x2*y1 + _ax[17]*x2*y2 +
					_ax[18]*x2*y3 + _ax[19]*x2*y4 + _ax[20]*x2*y5 +

					_ax[21]*x3 + _ax[22]*x3*y1 + _ax[23]*x3*y2 +
					_ax[24]*x3*y3 + _ax[25]*x3*y4 +

					_ax[26]*x4 + _ax[27]*x4*y1 + _ax[28]*x4*y2 +
					_ax[29]*x4*y3 +

					_ax[30]*x5 + _ax[31]*x5*y1 + _ax[32]*x5*y2 +

					_ax[33]*x6 + _ax[34]*x6*y1 +

					_ax[35]*x7;

		deltaY[n] = _ay[0] + _ay[1]*y1 + _ay[2]*y2 * _ay[3]*y3 + _ay[4]*y4 +
					_ay[5]*y5 + _ay[6]*y6 + _ay[7]*y7 +

					_ay[8]*x1 + _ay[9]*x1*y1 + _ay[10]*x1*y2 + _ay[11]*x1*y3+
					_ay[12]*x1*y4 + _ay[13]*x1*y5 + _ay[14]*x1*y6 +

					_ay[15]*x2 + _ay[16]*x2*y1 + _ay[17]*x2*y2 +
					_ay[18]*x2*y3 + _ay[19]*x2*y4 + _ay[20]*x2*y5 +

					_ay[21]*x3 + _ay[22]*x3*y1 + _ay[23]*x3*y2 +
					_ay[24]*x3*y3 + _ay[25]*x3*y4 +

					_ay[26]*x4 + _ay[27]*x4*y1 + _ay[28]*x4*y2 +
					_ay[29]*x4*y3 +

					_ay[30]*x5 + _ay[31]*x5*y1 + _ay[32]*x5*y2 +

					_ay[33]*x6 + _ay[34]*x6*y1 +

					_ay[35]*x7;
		sigmaX[n] = sigmaY[n] = 0.f;
	}

	ENDPROC
} // OmGeoLinearizer::polynomialDeltas()



OmGeoLinearizer::~OmGeoLinearizer()
{
	PROC

	delete _xoffInter;
	delete _yoffInter;

	ENDPROC
} // OmGeoLinearizer::~OmGeoLinearizer()


