//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmPsfDataServer.cc	--- implementation of OmPsf related classes
//
//	DESCRIPTION:
//		All methods of class OmPsf and related are defined.
//
//	CHANGE HISTORY:
//		UL, 2001-02-13: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Tue Feb 13 15:25:35 MET 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Calibration Data Access Handbook, C. Erd et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/OmPsfDataServer.cc,v 1.5 2001/10/29 16:58:02 ulammers Exp $
//

#include <cmath>

#include "CalOalUtility.h"
#include "OmPsfDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//  ---( OmPsf )--------------------------------------------------------------
//
OmPsf::OmPsf(const Table *psfTab,
			 CalAtomCollection<OmPsf> *parent)
	: CalChild<CalAtomCollection<OmPsf> >(parent),
	  _cenX(0.), _cenY(0.), _selectedRadialProfile(_radialPsfProfile.end()),
	  _psfr(0), _rpsf(0)
{
	//
	//	set aperture radius
	//
	_apertureRadius = CalOalUtility::optionalAttribute(psfTab, "APERTURE", 0.);

	//
	//	read polynomialCoefficients for stretch factor polynomial
	//
	_fStretchCoeff.push_back(psfTab->attribute("PARX0Y0")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX0Y1")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX0Y2")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX0Y3")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX1Y0")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX1Y1")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX1Y2")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX2Y0")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX2Y1")->asReal());
	_fStretchCoeff.push_back(psfTab->attribute("PARX3Y0")->asReal());

	vector<double> cfrr;
	CalOalUtility::fillVector(psfTab, "CFRR", cfrr);
	unsigned N = psfTab->rows();
	for (unsigned n=0; n<N; ++n) {
		RadialPsf rpsf;
		CalOalUtility::fillVector(psfTab, "RMAX", n, rpsf.first);
		CalOalUtility::fillVector(psfTab, "INTENSITY", n, rpsf.second);
		//	add anchor points at r=0, HUGE to avoid extrapolation artefacts
		rpsf.first.insert(rpsf.first.begin(), 0.); rpsf.first.push_back(1000.);
		rpsf.second.insert(rpsf.second.begin(), 0.); rpsf.second.push_back(1.);
		_radialPsfProfile[cfrr[n]] = rpsf;
	}
} // OmPsf::OmPsf()


void OmPsf::dump(ostream &) const
{
	PROC

	errstr<<"Sorry - dumping to file is not yet implemented"
		  <<Sas::warning("NotYetImplemenetd");

	ENDPROC
} // OmPsf::dump()


const OmPsf &OmPsf::forCountFrameRateRatio(double countFrameRateRatio) const
{
	PROC

	if (countFrameRateRatio<0. || countFrameRateRatio>1.)
		errstr<<"Count-frame-rate ratio must be in interval [0, 1] - saw "
			  <<countFrameRateRatio<<Sas::error("ValueOutOfValidBounds");

	//	select the radial profile curve that is closest to the given
	//	count frame rate ratio
	MapType::const_iterator ir =
						_radialPsfProfile.upper_bound(countFrameRateRatio);
	if (ir == _radialPsfProfile.end())
		//	last profile must be closest one
		_selectedRadialProfile = --ir;
	else if (ir == _radialPsfProfile.begin())
		_selectedRadialProfile = ir;
	else {
		MapType::const_iterator il = ir; --il;
		_selectedRadialProfile = fabs(countFrameRateRatio-il->first) <
						 fabs(countFrameRateRatio-ir->first) ? il : ir;
	}

	RETURN *this;
} // OmPsf::forCountFrameRateRatio()


double OmPsf::fStretch() const
{
	PROC

	if (_fStretchCoeff.size() != 10)
		errstr<<"Expected to have 10 coefficients for OM PSF radial stretching"
				" function but only see "<<_fStretchCoeff.size()
			  <<Sas::error("CcfIntegrityError");

	double x = _cenX,
		   y = _cenY,
		   x2 = x*x,
		   x3 = x2*x,
		   y2 = y*y,
		   y3 = y2*y;

	const vector<double> &a = _fStretchCoeff;
	double fs = a[0] + a[1]*y + a[2]*y2 + a[3]*y3 + a[4]*x + a[5]*x*y +
			    a[6]*x*y2 + a[7]*x2 + a[8]*x2*y + a[9]*x3;
	SAS_DEBUG(cout<<"fstretch="<<fs<<endl;)

//	RETURN a[0] + a[1]*y + a[2]*y2 + a[3]*y3 + a[4]*x + a[5]*x*y +
//		   a[6]*x*y2 + a[7]*x2 + a[8]*x2*y + a[9]*x3;
	RETURN fs;
} // OmPsf::fStretch()


const OmPsf &OmPsf::withCenterAt(double cenX, double cenY) const
{
	PROC

	if (_selectedRadialProfile == _radialPsfProfile.end())
		errstr<<"Please select first a count-frame-rate ratio with "
				"OmPsf::forCountFrameRateRatio() before calling "
				"OmPsf::withCenterAt()"
			  <<Sas::error("CountFrameRateRationNotSet");

	_cenX = cenX-1024.;
	_cenY = cenY-1024.;

	//	construct radial PSF curve
	vector<double> r(_selectedRadialProfile->second.first);
	r *= fStretch();
	delete _psfr;
	delete _rpsf;
	_psfr = new InterpolatedCurve(r, _selectedRadialProfile->second.second);
	_rpsf = new InterpolatedCurve(_selectedRadialProfile->second.second, r);
	SAS_DEBUG(unsigned N=r.size();
		for (unsigned n=0; n<N; ++n)
			cout<<r[n]<<' '<<(*_psfr)(r[n])<<endl;)

	RETURN *this;
} // OmPsf::withCenterAt()


const InterpolatedCurve &OmPsf::encircledEnergy() const
{
	PROC

	if (!_psfr)
		errstr<<"Please set first the PSF center with "
				"OmPsf::withCenterAt()"<<Sas::error("OmPsfCenterNotSet");

	RETURN *_psfr;
} // OmPsf::encircledEnergy()


const InterpolatedCurve &OmPsf::circleRadius() const
{
	PROC

	if (!_rpsf)
		errstr<<"Please set first the PSF center with "
				"OmPsf::withCenterAt()"<<Sas::error("OmPsfCenterNotSet");

	RETURN *_rpsf;
} // OmPsf::circleRadius()


SImage OmPsf::image(const OmPsfWindow &psfWindow,
					unsigned resX, unsigned resY) const
{
	PROC

	const InterpolatedCurve &psf = encircledEnergy();
	unsigned NX = psfWindow.nPixNegX + psfWindow.nPixPosX + 1,
			 NY = psfWindow.nPixNegY + psfWindow.nPixPosY + 1;
	SImage map(NX * resX, NY * resY);
	double dxi = 1./DBL(resX),
		   dyi = 1./DBL(resY),
		   minDist = min(dxi, dyi),
		   subPixArea = dxi*dyi/M_TWO_PI;

	SAS_DEBUG(double tot = 0.;)
	//	iterator over all pixels
	for (unsigned nx=0; nx<NX; ++nx) {
		double x = -DBL(psfWindow.nPixNegX) + DBL(nx) - .5,
			   pixX[] = { x, x+1., x+1., x };
		for (unsigned ny=0; ny<NY; ++ny) {
			double y = -DBL(psfWindow.nPixNegY) + DBL(ny) - .5,
				   pixY[] = { y, y, y+1., y+1. };

			//	iterate over subpixels
			for (unsigned nxi = 0; nxi<resX; ++nxi) {
				double x = pixX[0]+DBL(nxi)*dxi,
					   subPixX[] = { x, x+dxi, x+dxi, x };
				for (unsigned nyi = 0; nyi<resY; ++nyi) {
					double y = pixY[0]+DBL(nyi)*dyi,
						   subPixY[] = { y, y, y+dyi, y+dyi };

					double rmin2 = 1.e20,
						   rmax2 = -rmin2;
					for (unsigned k=0; k<4; ++k) {
						double d = CalOalUtility::SQR(subPixX[k]) +
								   CalOalUtility::SQR(subPixY[k]);
						if (d<rmin2)
							rmin2 = d;
						if (d>rmax2)
							rmax2 = d;
					}

					double rmin = sqrt(rmin2),
						   rmax = sqrt(rmax2);

					map[resX*nx+nxi][resY*ny+nyi] =
						rmin >= minDist ?
						subPixArea/(rmin*(rmax-rmin)) * (psf(rmax)-psf(rmin))
						: 2.*subPixArea/(rmax*rmax) * psf(rmax);

				SAS_DEBUG(tot += map[resX*nx+nxi][resY*ny+nyi];)
				}
			}
		}
	}
	SAS_DEBUG(cout<<"tot="<<tot<<endl;)

	RETURN map.normalize();
} // OmPsf::image()


OmPsf::~OmPsf()
{
	PROC

	delete _psfr;

	ENDPROC
} // OmPsf::~OmPsf()


//
//  ---( CalAtomCollection<OmPsf> )--------------------------------------------
//
CalAtomCollection<OmPsf>::CalAtomCollection(
					const XmmInstrument *instrument,
					const Ccf *ccf,
					Xmm::Filter filter,
					AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
    _nCCDs      = instrument->nCcds();
    _nNodes     = instrument->nNodes();
	_nEntries	= 1;
	OmPsf *null = 0;
	_collection	= new vector<OmPsf *>(_nEntries, null);

    CifEntry *e = ccf->entry(instrument, Ccf::PSF1DRB);
	DataSet *set = e->open();

	//	select the right table
	unsigned N = set->blocks();
	string extName = "PSF-" + uppercase(Xmm::toString(filter));
	unsigned n = 0;
	for (; n<N; ++n)
		if (set->block(n)->name() == extName) {
			(*_collection)[0] = new OmPsf(set->table(n), this);
			break;
		}

	if (n>=N)
		errstr<<"Could not find table for filter `"<<filter<<"' in `"
			  <<set->qualifiedName()<<Sas::error("NoSuchTable");

	dataSetServer->close(set);

	ENDPROC
} // CalAtomCollection<OmPsf>::CalAtomCollection


OmPsf *CalAtomCollection<OmPsf>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[0];
} // CalAtomCollection<OmPsf>::select()
